/**********************************************************
 * smbc_lib.h 
 *********************************************************/
#ifndef _SMBC_LIB_H_
#define _SMBC_LIB_H_

#if 1
#define SMBC_USE_PTL
#endif

#ifdef SMBC_USE_PTL
#include <pthread.h>
#endif /* SMBC_USE_PTL */
#include <sys/types.h>
#include <ctype.h>
#include <smberr.h>

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif


#ifdef  NCS_DEBUG
#define DPRINTF(A)              printf A
#else
#define DPRINTF(A)              ((void) 0)
#endif
#define SMB_CONF_PATH_NAME      "./conf/smb.conf" 
#define NBTD_CONF_PATH_NAME     ""
#define FTAPI_ABORT_REQOFF              0       /* fvȂ                         */
#define FTAPI_ABORT_INTERVAL            1       /* 1b                                  */





#define SMB_MAXSRVNAMELEN       15	/* NetBIOS limit */
#define SMB_MAXHOSTNAMELEN      128	/* FQDN limit */
#define SMB_MAXSCOPELEN         128
#define SMB_MAXUSERNAMELEN      128
#define SMB_MAXGROUPNAMELEN     20
#define SMB_MAXPASSWORDLEN      128
#define SMB_MAXSHARENAMELEN     128
#define SMB_MAXFNAMELEN         255	/* Keep in sync with MAXNAMLEN */
#define SMB_MAXCOMMENTLEN       31	/* Keep in sync with MAXNAMLEN */

#define SMBC_CRITMSG   100     /* critical error */
#define SMBC_WARNMSG   50      /* warning */
#define SMBC_INFOMSG   1       /* infomation */
#define SMBC_DEBUGLEVEL SMBC_INFOMSG

#ifdef SMBC_DEBUGLEVEL
#define SMBC_DEBUG(level, body) ((SMBC_DEBUGLEVEL<=(level))?(printf body):0)
#else
#define SMBC_DEBUG(level, body)
#endif

/*
 * file attributes
 */
/* access */
#define	SMB_FILE_SHARE_READ	1
#define	SMB_FILE_SHARE_WRITE	2
#define	SMB_FILE_SHARE_DELETE	3
/* attribute */
/* these define the attribute byte as seen by DOS */
#define	SMB_ATTR_READONLY	0x01L	
#define	SMB_ATTR_HIDDEN		0x02L	
#define	SMB_ATTR_SYSTEM		0x04L
#define	SMB_ATTR_VOLID		0x08L
#define	SMB_ATTR_DIRECTORY	0x10L
#define	SMB_ATTR_NORMAL		0x80L
#define	SMB_ATTR_ARCHIVE	0x20L
#define	SMB_ATTR_COMPRESSED	0x800L
#define	SMB_ATTR_TEMPORARY	0x100L
/* disposition */
#define	SMB_CREATE_NEW		1
#define	SMB_CREATE_ALWAYS	2
#define	SMB_OPEN_EXISTING	3
#define	SMB_OPEN_ALWAYS		4
/* flag */
#define	SMB_CREATE_DIR		0x01
#define	SMB_OPEN_READONLY	0x02
#define	SMB_CREATE_SUBDIR	0x04
#define	SMB_REMOVE_RECURSION	0x08

/*
 * function id
 */
#ifdef SMBC_USE_PTL
#define SMBC_OPEN       1
#define SMBC_CLOSE      2
#define SMBC_CONNECT    3
#define SMBC_DISCONNECT 4
#define SMBC_OPENFILE   5
#define SMBC_CLOSEFILE  6
#define SMBC_READFILE   7
#define SMBC_WRITEFILE  8
#define SMBC_DELETEFILE 9
#define SMBC_MAKEDIR    10
#define SMBC_DELEDIR    11
#define SMBC_CHKDIR     12
#define SMBC_GETGRPLST  13
#define SMBC_GETSRVLST  14
#define SMBC_GETSHRLST  15
#define SMBC_GETDIRLST  16
#define SMBC_GETSRVCAP  17
#define SMBC_USERLOGIN  18
#define SMBC_USERLOGOUT 19
#define SMBC_USERGETGRP 20
#define SMBC_GETDCINFO  21
#endif /* SMBC_USE_PTL */


#define NBT_LENGTH_NETBIOSNAME                        16
#define NBT_LENGTH_SCOPEID              222
#define LIBNBT_ERR_NOERR                   0
/* NBT Service Type */
#define NBT_STYPE_NAME          0
#define NBT_STYPE_SESSION       1
#define NBT_STYPE_DATAGRAM      2
#define NBT_PORT_ANY                   0
#define NBT_ONT_B       0       // B-node
#define NBT_ONT_P       1       // P-node
#define NBT_ONT_M       2       // M-node
#define NBT_ONT_H       3       // H-node




typedef struct tag_NBTNodeStatistics
{
  unsigned char   unit_id[6];
  unsigned char   jumpers;
  unsigned char   test_result;
  unsigned short  version_number;
  unsigned short  period_of_statistics;
  unsigned short  number_of_crcs;
  unsigned short  number_alignment_errors;
  unsigned short  number_of_collisions;
  unsigned short  number_send_aborts;
  unsigned long   number_good_sends;
  unsigned long   number_good_receives;
  unsigned short  number_retransmits;
  unsigned short  number_no_resource_conditions;
  unsigned short  number_free_command_blocks;
  unsigned short  total_number_command_blocks;
  unsigned short  max_total_number_command_blocks;
  unsigned short  number_pending_sessions;
  unsigned short  max_number_pending_sessions;
  unsigned short  max_total_sessions_possible;
  unsigned short  session_data_packet_size;

} NBTNodeStatistics;

typedef struct tag_NBTAddrPack
{
  unsigned long   naddr;
  unsigned char   group;
  unsigned char   ont;

} NBTAddrPack;

typedef struct tag_NBTNodeName
{
  char                    name[NBT_LENGTH_NETBIOSNAME]; // except suffix
  unsigned char   suffix;
  unsigned char   group;
  unsigned char   ont;
  unsigned char   state;

} NBTNodeName;


typedef struct tag_NBTNamePack
{
  char                    name[NBT_LENGTH_NETBIOSNAME]; // except suffix
  unsigned char   suffix;
  unsigned long   naddr;  // network byte order
  unsigned char   group;  // group or unique

} NBTNamePack;


typedef struct {
	u_short		character;			// 
        char            group[MAX(SMB_MAXSRVNAMELEN,SMB_MAXHOSTNAMELEN)*2+1];
	char		server[SMB_MAXHOSTNAMELEN*2+1];	// 
	char		share[SMB_MAXSHARENAMELEN*2+1];	//
	char		username[SMB_MAXUSERNAMELEN*2+1];	//
	char		password[SMB_MAXPASSWORDLEN*2+1];	//
        u_char          vc_hflags;      /* or'ed with flags in the smb header */
        u_short         vc_hflags2;     /* or'ed with flags in the smb header */
} smb_vc;

typedef struct {
	u_short	character;
	char		name[SMB_MAXFNAMELEN+1];
	u_long		access;
	u_long		attribute;
	u_long		disposition;
	u_long		flag;
} smb_file;

typedef struct {
	u_short		fid;			// 
	time_t		creation_time;		// first creation time 
	time_t		last_access_time;	// last access time
	time_t		last_write_time;	// last write time
	time_t		change_time;		// last attribute change time
	u_long		attribute;		// 
	u_long		size;			// 
} smb_finfo;

typedef struct {
	u_short	character;			// 
	char	gp_name[SMB_MAXSRVNAMELEN*2+1];	// 
	char	gp_version_major;		// 
	char	gp_version_minor;		// 
	u_long	gp_type;			// 
	char	gp_master_browser[SMB_MAXSRVNAMELEN*2+1];	// 
} smb_gp_info;

typedef struct {
	u_long		entries;		// 
	u_long		totalnum;		// 
	smb_gp_info*	group_info;		// 
} smb_gp_list;

typedef struct {
	u_short	character;			// 
	char	sv_name[SMB_MAXSRVNAMELEN*2+1];	// 
	char	sv_version_major;			// 
	char	sv_version_minor;			// 
	u_long	sv_type;				// 
	char	sv_comment[SMB_MAXCOMMENTLEN*2+1];	// 
} smb_sv_info;

typedef struct {
	u_long		entries;			// 
	u_long		totalnum;		// 
	smb_sv_info*	server_info;			// 
} smb_sv_list;

typedef struct {
	u_short	character;				// 
	char	sh_name[SMB_MAXSHARENAMELEN*2+1];	// 
	char	sh_comment[SMB_MAXCOMMENTLEN*2+1];	// 
} smb_sh_info;

typedef struct {
	u_long		entries;			// 
	u_long		totalnum;		// 
	smb_sh_info*	share_info;			// 
} smb_sh_list;

typedef struct {
	u_short		character;			// 
	char		dr_name[SMB_MAXFNAMELEN*2+1];	// 
	smb_finfo	finfo;				// 
} smb_dr_info;

typedef struct {
	u_long		entries;			// 
	u_long		totalnum;		// 
	smb_dr_info*	directory_info;			// 
} smb_dr_list;

typedef struct {
        u_short         character;                      //
        char            ugp_name[SMB_MAXGROUPNAMELEN*2+1];      //
} smb_ugp_info;

typedef struct {
        u_long          entries;                        //
	u_long		totalnum;		// 
        smb_ugp_info*   ugp_info;                       //
} smb_ugp_list;

typedef struct dc_info{
	u_long		address;	// DC address
} smb_dc_info;

typedef struct {
	u_long		flag;			// 
	u_long		sec_mode;		// 
	u_char		encryption_key_len;	// 
	u_char		*encryption_key;	// 
} smb_server_cap;

typedef struct {
        NBTNamePack		local;
        NBTNamePack		rmt;
	int			ns_client;
	int			ss_client;
	int			ds_client;
	struct cli_state	*cli;
#ifdef SMBC_USE_PTL
	pthread_mutex_t		mutex;
#endif /* SMBC_USE_PTL */
        int             smbc_cli_fd;
        u_char          message_ID;
} smb_ctx;

typedef int (*SMBC_PROC)(int, int, int);

typedef struct _SMBC_CALLBACK {
	SMBC_PROC	func;
	int			param;
} SMBC_CALLBACK;

#ifdef SMBC_USE_PTL
typedef struct {
        smb_ctx         *ctx;
        smb_vc          *vcp;
        SMBC_CALLBACK   *callback;
} smbc_connect_t;               /* SmbConnect */

typedef struct {
        smb_ctx         *ctx;
        SMBC_CALLBACK   *callback;
} smbc_disconnect_t;            /* SmbDisconnect */

typedef struct {
        smb_ctx         *ctx;
        smb_vc          *vcp;
        SMBC_CALLBACK   *callback;
} smbc_ulogin_t;               /* SmbUserLogin */

typedef struct {
        smb_ctx         *ctx;
        SMBC_CALLBACK   *callback;
} smbc_ulogout_t;            /* SmbUserLogout */

typedef struct {
        smb_ctx         *ctx;
        smb_file        *file;
        smb_finfo       *finfo;
        SMBC_CALLBACK   *callback;
} smbc_openfile_t;              /* SmbOpenFile */

typedef struct {
        smb_ctx         *ctx;
        u_short         fid;
        SMBC_CALLBACK   *callback;
} smbc_closefile_t;             /* SmbCloseFile */

typedef struct {
        smb_ctx         *ctx;
        u_short         fid;
        u_long          offset;
        u_char          *buf;
        int             *size;
        SMBC_CALLBACK   *callback;
} smbc_readfile_t;              /* SmbReadFile */

typedef struct {
        smb_ctx         *ctx;
        u_short         fid;
        u_long          offset;
        u_long          mode;
        u_char          *buf;
        int             *size;
        SMBC_CALLBACK   *callback;
} smbc_writefile_t;

typedef struct {
        smb_ctx         *ctx;
        smb_file        *file;
        SMBC_CALLBACK   *callback;
} smbc_deletefile_t;            /* SmbDeleteFile */

typedef struct {
        smb_ctx*        ctx;
        smb_file*       file;
        smb_finfo*      finfo;
        SMBC_CALLBACK   *callback;
} smbc_makedir_t;                       /* SmbMakeDirectory */

typedef struct {
        smb_ctx*        ctx;
        smb_file*       file;
        SMBC_CALLBACK   *callback;
} smbc_deledir_t;                       /* SmbDeleteDirectory */

typedef struct {
        smb_ctx*        ctx;
        smb_file*       file;
        SMBC_CALLBACK   *callback;
} smbc_chkdir_t;                        /* SmbCheckDirectory */

typedef struct {
        smb_ctx*        ctx;
        u_short         max_list_num;
        smb_gp_list*    group_list;
        SMBC_CALLBACK   *callback;
} smbc_getgrplist_t;            /* SmbGetGroupList */

typedef struct {
        smb_ctx*        ctx;
        smb_vc*         vcp;
        u_short         max_list_num;
        smb_sv_list*    server_list;
        SMBC_CALLBACK   *callback;
} smbc_getsrvlist_t;            /* SmbGetServerList */

typedef struct {
        smb_ctx*        ctx;
        smb_vc*         vcp;
        u_short         max_list_num;
        smb_sh_list*    share_list;
        SMBC_CALLBACK   *callback;
} smbc_getshrlist_t;            /* SmbGetShareList */

typedef struct {
        smb_ctx*        ctx;
        smb_vc*         vcp;
        const char*     directory;
        u_short         max_list_num;
        smb_dr_list*    dr_list;
        SMBC_CALLBACK   *callback;
} smbc_getdirlist_t;            /* SmbGetDirectoryList */

typedef struct {
        smb_ctx*        ctx;
        smb_vc*         vcp;
        smb_server_cap* capability;
        SMBC_CALLBACK   *callback;
} smbc_getsrvcap_t;             /* SmbGetServerCapability */

typedef struct {
        smb_ctx*        ctx;
        smb_vc*         vcp;
        u_short         max_list_num;
        smb_ugp_list*    ugp_list;
        SMBC_CALLBACK   *callback;
} smbc_usrgroup_t;            /* SmbUserGetGroups */

typedef struct {
        smb_ctx         *ctx;
        smb_vc          *vcp;
	smb_dc_info	*dcinfo;
        SMBC_CALLBACK   *callback;
} smbc_dc_info_t;               /* SmbUserLogin */

#endif /* SMBC_USE_PTL */

//////////////////////////////////////////////////////////
extern int NcsSmbOpen( smb_ctx** );
extern int NcsSmbClose( smb_ctx* );
extern int NcsSmbConnect( smb_ctx* , smb_vc* , SMBC_CALLBACK* );
extern int NcsSmbDisconnect( smb_ctx* , SMBC_CALLBACK* );
extern int NcsSmbOpenFile( smb_ctx* , smb_file* , smb_finfo* , SMBC_CALLBACK* );
extern int NcsSmbCloseFile( smb_ctx* , u_short , SMBC_CALLBACK* );
extern int NcsSmbReadFile( smb_ctx* , u_short , u_long , u_char* , int* , SMBC_CALLBACK* );
extern int NcsSmbWriteFile( smb_ctx* , u_short , u_long , u_long , u_char* , int* , SMBC_CALLBACK* );
extern int NcsSmbDeleteFile( smb_ctx* , smb_file* file, SMBC_CALLBACK* );
extern int NcsSmbMakeDirectory( smb_ctx* , smb_file* , smb_finfo* , SMBC_CALLBACK* );
extern int NcsSmbDeleteDirectory( smb_ctx* , smb_file* , SMBC_CALLBACK* );
extern int NcsSmbCheckDirectory( smb_ctx* , smb_file* , SMBC_CALLBACK* );
extern int NcsSmbGetGroupList( smb_ctx* , u_short , smb_gp_list* , SMBC_CALLBACK* );
extern int NcsSmbGetServerList( smb_ctx* , smb_vc* , u_short , smb_sv_list* , SMBC_CALLBACK* );
extern int NcsSmbGetShareList( smb_ctx* , smb_vc* , u_short , smb_sh_list* , SMBC_CALLBACK* );
extern int NcsSmbDirectoryList( smb_ctx* , smb_vc* , const char* , u_short , smb_dr_list* , SMBC_CALLBACK* );
extern int NcsSmbGetServerCapability( smb_ctx* ctx, smb_vc* , smb_server_cap* , SMBC_CALLBACK* );

extern int NcsSmbUserLogin(smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback);
extern int NcsSmbUserLogout( smb_ctx *ctx, SMBC_CALLBACK* callback );
extern int NcsSmbUserGetGroups( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list, SMBC_CALLBACK* callback);
extern int NcsSmbGetDCInfo(smb_ctx* ctx, smb_vc* vcp, smb_dc_info* dcinfo, SMBC_CALLBACK* callback);


#ifdef SMBC_USE_PTL
extern void *NcsSmbConnectThread( void* );
extern void *NcsSmbDisconnectThread( void* );
extern void *NcsSmbOpenFileThread( void* );
extern void *NcsSmbCloseFileThread( void* );
extern void *NcsSmbReadFileThread( void* );
extern void *NcsSmbWriteFileThread( void* );
extern void *NcsSmbDeleteFileThread( void* );
extern void *NcsSmbMakeDirectoryThread( void* );
extern void *NcsSmbDeleteDirectoryThread( void* );
extern void *NcsSmbCheckDirectoryThread( void* );
extern void *NcsSmbGroupListThread( void* );
extern void *NcsSmbGetServerListThread( void* );
extern void *NcsSmbGetShareListThread( void* );
extern void *NcsSmbDirectoryListThread( void* );
extern void *NcsSmbGetServerCapabilityThread( void* );
extern void *NcsSmbUserLoginThread(void *arg);
extern void *NcsSmbUserLogoutThread(void* arg);
extern void *NcsSmbUserGetGroupsThread( void* arg );
extern void *NcsSmbGetDCInfoThread( void* arg );

extern int NcsSmbConnectSub( smb_ctx*, smb_vc* );
extern int NcsSmbDisconnectSub( smb_ctx* );
extern int NcsSmbOpenFileSub( smb_ctx*, smb_file*, smb_finfo* );
extern int NcsSmbCloseFileSub( smb_ctx*, u_short );
extern int NcsSmbReadFileSub( smb_ctx* , u_short , u_long , u_char* , int* );
extern int NcsSmbWriteFileSub( smb_ctx* , u_short , u_long , u_long , u_char* , int* );
extern int NcsSmbDeleteFileSub( smb_ctx* , smb_file* file );
extern int NcsSmbMakeDirectorySub( smb_ctx* , smb_file* , smb_finfo* );
extern int NcsSmbDeleteDirectorySub( smb_ctx* , smb_file* );
extern int NcsSmbCheckDirectorySub( smb_ctx* , smb_file* );
extern int NcsSmbGetGroupListSub( smb_ctx* , u_short , smb_gp_list* );
extern int NcsSmbGetServerListSub( smb_ctx* , smb_vc* , u_short , smb_sv_list* );
extern int NcsSmbGetShareListSub( smb_ctx* , smb_vc* , u_short , smb_sh_list* );
extern int NcsSmbDirectoryListSub( smb_ctx* , smb_vc* , const char* , u_short , smb_dr_list* );
extern int NcsSmbGetServerCapabilitySub( smb_ctx* ctx, smb_vc* , smb_server_cap* );

extern int NcsSmbUserLoginSub( smb_ctx* ctx, smb_vc* vcp);
extern int NcsSmbUserLogoutSub( smb_ctx* ctx );
extern int NcsSmbUserGetGroupsSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list );
extern int NcsSmbGetDCInfoSub(smb_ctx* ctx, smb_vc* vcp, smb_dc_info* dcinfo);

#endif /* SMBC_USE_PTL */

/*
 * Two levels of connection hierarchy
 */
#define	SMBL_SM		0
#define SMBL_VC		1
#define SMBL_SHARE	2
#define SMBL_NUM	3
#define SMBL_NONE	(-1)

#define	SMB_CS_NONE	0x0000
#define	SMB_CS_UPPER	0x0001		/* convert passed string to upper case */
#define	SMB_CS_LOWER	0x0002		/* convert passed string to lower case */

/*
 * Common object flags
 */
#define SMBO_GONE		0x1000000

/*
 * access modes
 */
#define	SMBM_READ		0400	/* read conn attrs.(like list shares) */
#define	SMBM_WRITE		0200	/* modify conn attrs */
#define	SMBM_EXEC		0100	/* can send SMB requests */
#define	SMBM_READGRP		0040
#define	SMBM_WRITEGRP		0020
#define	SMBM_EXECGRP		0010
#define	SMBM_READOTH		0004
#define	SMBM_WRITEOTH		0002
#define	SMBM_EXECOTH		0001
#define	SMBM_MASK		0777
#define	SMBM_EXACT		010000	/* check for specified mode exactly */
#define	SMBM_ALL		(SMBM_READ | SMBM_WRITE | SMBM_EXEC)
#define	SMBM_DEFAULT		(SMBM_READ | SMBM_WRITE | SMBM_EXEC)
#define	SMBM_ANY_OWNER		((uid_t)-1)
#define	SMBM_ANY_GROUP		((gid_t)-1)

/*
 * VC flags
 */
#define SMBV_PERMANENT		0x0002
#define SMBV_LONGNAMES		0x0004	/* connection is configured to use long names */
#define	SMBV_ENCRYPT		0x0008	/* server asked for encrypted password */
#define	SMBV_WIN95		0x0010	/* used to apply bugfixes for this OS */
#define	SMBV_PRIVATE		0x0020	/* connection can be used only by creator */
#define	SMBV_RECONNECTING	0x0040	/* conn is in the process of reconnection */
#define SMBV_SINGLESHARE	0x0080	/* only one share connectin should be allowed */
#define SMBV_CREATE		0x0100	/* lookup for create opeartion */
/*#define SMBV_FAILED		0x0200*/	/* last reconnect attempt has failed */


/*
 * smb_share flags
 */
#define SMBS_PERMANENT		0x0001
#define SMBS_RECONNECTING	0x0002
#define SMBS_CONNECTED		0x0004

/*
 * share types
 */
#define	SMB_ST_DISK		0x0	/* A: */
#define	SMB_ST_PRINTER		0x1	/* LPT: */
#define	SMB_ST_PIPE		0x2	/* IPC */
#define	SMB_ST_COMM		0x3	/* COMM */
#define	SMB_ST_ANY		0x4
#define	SMB_ST_MAX		0x4
#define SMB_ST_NONE		0xff	/* not a part of protocol */

/*
 * Negotiated protocol parameters
 */
struct smb_sopt {
	int		sv_proto;
	int16_t		sv_tz;		/* offset in min relative to UTC */
	u_int32_t	sv_maxtx;	/* maximum transmit buf size */
	u_char		sv_sm;		/* security mode */
	u_int16_t	sv_maxmux;	/* max number of outstanding rq's */
	u_int16_t 	sv_maxvcs;	/* max number of VCs */
	u_int16_t	sv_rawmode;
	u_int32_t	sv_maxraw;	/* maximum raw-buffer size */
	u_int32_t	sv_skey;	/* session key */
	u_int32_t	sv_caps;	/* capabilites SMB_CAP_ */
};

/*
 * network IO daemon states
 */
enum smbiod_state {
	SMBIOD_ST_NOTCONN,	/* no connect request was made */
	SMBIOD_ST_RECONNECT,	/* a [re]connect attempt is in progress */
	SMBIOD_ST_TRANACTIVE,	/* transport level is up */
	SMBIOD_ST_VCACTIVE,	/* session established */
	SMBIOD_ST_DEAD		/* connection broken, transport is down */
};


/*
 * Info structures
 */
#define	SMB_INFO_NONE		0
#define	SMB_INFO_VC		2
#define	SMB_INFO_SHARE		3


///////////////////////////////////
/*
 * Crediantials of user/process being processing in the connection procedures
 */
struct smb_cred {
        struct proc *   scr_p;
        struct ucred *  scr_cred;
};

/*
 * prototype 
 */
extern int NcsSmbMfprintfCreate( key_t key );
extern int NcsSmbMfprintfDelete( void );
extern void NcsSmbSetAllocatorFunc( void *(*_alloc_func)(size_t), void(*_free_func)(void *) );
extern void NcsSmbSetDefaultCharCode( u_char char_code );
extern int NcsSmbOpen( smb_ctx** ctx );
extern int NcsSmbClose( smb_ctx* ctx );
extern int NcsSmbConnect( smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback );
extern int NcsSmbDisconnect( smb_ctx* ctx, SMBC_CALLBACK* callback );
extern int NcsSmbOpenFile( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback );
extern int NcsSmbCloseFile( smb_ctx* ctx, u_short fid, SMBC_CALLBACK* callback );
extern int NcsSmbReadFile( smb_ctx* ctx, u_short fid, u_long offset, u_char* buf, int* size, SMBC_CALLBACK* callback );
extern int NcsSmbWriteFile( smb_ctx* ctx, u_short fid, u_long offset, u_long mode, u_char* buf, int* size, SMBC_CALLBACK* callback );
extern int NcsSmbDeleteFile( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback );
extern int NcsSmbMakeDirectory( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback );
extern int NcsSmbDeleteDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback );
extern int NcsSmbCheckDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback );
extern int NcsSmbGetGroupList( smb_ctx* ctx, u_short max_list_num, smb_gp_list* group_list, SMBC_CALLBACK* callback );
extern int NcsSmbGetServerList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sv_list* server_list, SMBC_CALLBACK* callback );
extern int NcsSmbGetShareList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sh_list* share_list, SMBC_CALLBACK* callback );
extern int NcsSmbDirectoryList( smb_ctx* ctx, smb_vc* vcp, const char* directory, u_short max_list_num, smb_dr_list* dr_list, SMBC_CALLBACK* callback );
extern int NcsSmbGetServerCapability( smb_ctx* ctx, smb_vc* vcp, smb_server_cap* capability, SMBC_CALLBACK* callback );
#ifdef SMBC_USE_PTL
/* PTL */
extern void *NcsSmbConnectThread( void* th_arg );
extern void *NcsSmbDisconnectThread( void* th_arg );
extern void *NcsSmbOpenFileThread( void* th_arg );
extern void *NcsSmbCloseFileThread( void* th_arg );
extern void *NcsSmbReadFileThread( void* th_arg );
extern void *NcsSmbWriteFileThread( void* th_arg );
extern void *NcsSmbDeleteFileThread( void* th_arg );
extern void *NcsSmbMakeDirectoryThread( void* th_arg );
extern void *NcsSmbDeleteDirectoryThread( void* th_arg );
extern void *NcsSmbCheckDirectoryThread( void* th_arg );
extern void *NcsSmbGetGroupListThread( void* th_arg );
extern void *NcsSmbGetServerListThread( void* th_arg );
extern void *NcsSmbGetShareListThread( void* th_arg );
extern void *NcsSmbDirectoryListThread( void* th_arg );
extern void *NcsSmbGetServerCapabilityThread( void* th_arg );
/* Sub */
extern int NcsSmbConnectSub( smb_ctx* ctx, smb_vc* vcp );
extern int NcsSmbDisconnectSub( smb_ctx* ctx );
extern int NcsSmbOpenFileSub( smb_ctx* ctx, smb_file* file, smb_finfo* finfo );
extern int NcsSmbCloseFileSub( smb_ctx* ctx, u_short fid );
extern int NcsSmbReadFileSub( smb_ctx* ctx, u_short fid, u_long offset, u_char* buf, int* size );
extern int NcsSmbWriteFileSub( smb_ctx* ctx, u_short fid, u_long offset, u_long mode, u_char* buf, int* size );
extern int NcsSmbDeleteFileSub( smb_ctx* ctx, smb_file* file );
extern int NcsSmbMakeDirectorySub( smb_ctx* ctx, smb_file* file, smb_finfo* finfo );
extern int NcsSmbDeleteDirectorySub( smb_ctx* ctx, smb_file* file );
extern int NcsSmbCheckDirectorySub( smb_ctx* ctx, smb_file* file );
extern int NcsSmbGetGroupListSub( smb_ctx* ctx, u_short max_list_num, smb_gp_list* group_list );
extern int NcsSmbGetServerListSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sv_list* server_list );
extern int NcsSmbGetShareListSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sh_list* share_list );
extern int NcsSmbDirectoryListSub( smb_ctx* ctx, smb_vc* vcp, const char* directory, u_short max_list_num, smb_dr_list* dr_list );
extern int NcsSmbGetServerCapabilitySub( smb_ctx* ctx, smb_vc* vcp, smb_server_cap* capability );

extern int NcsSmbUserLoginSub( smb_ctx* ctx, smb_vc* vcp);
extern int NcsSmbUserLogoutSub( smb_ctx* ctx );
extern int NcsSmbUserGetGroupsSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list );
#endif /* SMBC_USE_PTL */
#endif	_SMBC_LIB_H_
