/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt.h
 *------------------------------------------------------------------*/

#ifndef _LIBNBT_H_
#define _LIBNBT_H_

#ifdef __cplusplus
extern "C" {
#endif

/////////////// public define

/* NBT Service Type */
#define NBT_STYPE_NAME		0
#define NBT_STYPE_SESSION	1
#define NBT_STYPE_DATAGRAM	2
//#define NBT_STYPE_NAME_CON	3

/* NBT message type */
/* name */

/* session */
#define NBT_MTYPE_SES_MES		100
#define NBT_MTYPE_SES_REQ		101
#define NBT_MTYPE_SES_RES_ACK	102
#define NBT_MTYPE_SES_RES_NACK	103
#define NBT_MTYPE_SES_RES_CHNG	104
#define NBT_MTYPE_SES_KPALIV	105
#define NBT_MTYPE_SES_DISCON	1000

/* datagram */
#define NBT_MTYPE_DGM_UNICAST		200
#define NBT_MTYPE_DGM_GROUP			201
#define NBT_MTYPE_DGM_BRDCAST		202
#define NBT_MTYPE_DGM_ERROR			203
#define NBT_MTYPE_DGM_QUERYREQ		204
#define NBT_MTYPE_DGM_QUERYRES_ACK	205
#define NBT_MTYPE_DGM_QUERYRES_NACK	206


/* etc */
#define NBT_MTYPE_NONE	0x00

/* NBT MSG */
#define NBT_MSG_ONEMES	0x01


// PORT number (gp|[gԍ)
#define NBT_PORT_NAME			137
#define NBT_PORT_DATAGRAM		138
#define NBT_PORT_SESSION		139
#define NBT_PORT_ANY			0

// OwnerNodeType
#define NBT_ONT_B	0	// B-node
#define NBT_ONT_P	1	// P-node
#define NBT_ONT_M	2	// M-node
#define NBT_ONT_H	3	// H-node

#define NBT_ONT_MAX	NBT_ONT_H

// Group
#define NBT_G_NONE		0
#define NBT_G_UNIQUE	10
#define NBT_G_GROUP		11

#define NBT_G_MAX NBT_G_GROUP


// NodeState
#define NBT_NODESTATE_DRG	(1<<0)
#define NBT_NODESTATE_CNF	(1<<1)
#define NBT_NODESTATE_ACT	(1<<2)
#define NBT_NODESTATE_PRM	(1<<3)


// Rcode
#define NBT_RC_SUCCESS	0x0
#define NBT_RC_FMT_ERR	0x1
#define NBT_RC_SRV_ERR	0x2
#define NBT_RC_NAM_ERR	0x3
#define NBT_RC_IMP_ERR	0x4
#define NBT_RC_RFS_ERR	0x5
#define NBT_RC_ACT_ERR	0x6
#define NBT_RC_CFT_ERR	0x7


// NEGATIVE_SESSION_RESPONSE (use nbt_send_nack)
#define NBT_ERROR_SES_SUCCESS			0
#define NBT_ERROR_SES_NL_CALLEDNAME		(-1)
#define NBT_ERROR_SES_NL_CALLINGNAME	(-2)
#define NBT_ERROR_SES_NP_CALLEDNAME		(-3)
#define NBT_ERROR_SES_RES_INSUFFICIENT	(-4)
#define NBT_ERROR_SES_UNSPECIFIED		(-5)

// nbt_look
#define NBT_LOOK_READ	0
#define NBT_LOOK_NOREAD	1


// nbt_senddatagram_error : ecode
#define NBT_ERROR_DGM_DSTNAME_NOT_PRESENT		0
#define NBT_ERROR_DGM_INVALID_SRCNAME_FORMAT	1
#define NBT_ERROR_DGM_INVALID_DSTNAME_FORMAT	2


#define NBT_LENGTH_NETBIOSNAME			16

// NewBEUIgpł낤fd̃io[ɏdȂȂ悤
#define NBT_FD_COUNT_BEGIN		100

// 쐬łSocketDiscriptor̐
#define NBT_NBTCTBL_MAX			64

// o^łAddressPack̍ő吔
#define NBT_ADDRPACK_MAX		20

// WAIT TIME(etF[Yōő剽~b҂)
#define NBT_WAIT_SEC			1000

// SCOPE_ID̍ő咷
#define NBT_LENGTH_SCOPEID		222

// NetBIOS̍ő咷
#define NBT_LENGTH_NETBIOSNAME			16
#define NBT_LENGTH_NETBIOSNAME_2ENC		34

// BROADCAST NAME
#define BROADCAST_NAME	"*"


//////////

#define BCAST_REQ_RETRY_TIMEOUT		250		// 250millisec
#define BCAST_REQ_RETRY_COUNT		3		// 3times
#define UCAST_REQ_RETRY_TIMEOUT		3000	// 5sec
#define UCAST_REQ_RETRY_COUNT		3		// 3times
#define MAX_DATAGRAM_LENGTH			568		// 576bytes(default) but this size is included UDP header size(8 bytes).
#define MAX_DATA_LENGTH				512
#define CONFLICT_TIMER				1000	// 1sec
#define MESSAGE_BUFSIZE				1024



///////////////// structure define
typedef struct tag_NBTNamePack
{
	char			name[NBT_LENGTH_NETBIOSNAME]; // except suffix
	unsigned char	suffix;
	unsigned long	naddr;	// network byte order
	unsigned char	group;	// group or unique

} NBTNamePack;

typedef struct tag_NBTAddrPack
{
	unsigned long	naddr;
	unsigned char	group;
	unsigned char	ont;

} NBTAddrPack;

typedef struct tag_NBTNodeName
{
	char			name[NBT_LENGTH_NETBIOSNAME]; // except suffix
	unsigned char	suffix;
	unsigned char	group;
	unsigned char	ont;
	unsigned char	state;

} NBTNodeName;


typedef struct tag_NBTNodeStatistics
{
	unsigned char	unit_id[6];
	unsigned char	jumpers;
	unsigned char	test_result;
	unsigned short	version_number;
	unsigned short	period_of_statistics;
	unsigned short	number_of_crcs;
	unsigned short	number_alignment_errors;
	unsigned short	number_of_collisions;
	unsigned short	number_send_aborts;
	unsigned long	number_good_sends;
	unsigned long	number_good_receives;
	unsigned short	number_retransmits;
	unsigned short	number_no_resource_conditions;
	unsigned short	number_free_command_blocks;
	unsigned short	total_number_command_blocks;
	unsigned short	max_total_number_command_blocks;
	unsigned short	number_pending_sessions;
	unsigned short	max_number_pending_sessions;
	unsigned short	max_total_sessions_possible;
	unsigned short	session_data_packet_size;
	
} NBTNodeStatistics;


/* ɐݒ肵lp̍\ */
typedef struct tag_NBTForceSettingInfo
{
	int				interruptd;			/* fp\PbgfBXNv^(-1: not used) */
	unsigned long	flag;				/* tO(0: none) */
	unsigned long	bcast_retry_timeo;	/* u[hLXgɂ郊gCԊu(~0: not used) */
	int				bcast_retry_count;	/* u[hLXgɂ郊gC(-1: not used) */
	unsigned long	ucast_retry_timeo;	/* jLXgɂ郊gCԊu(~0: not used) */
	int				ucast_retry_count;	/* jLXgɂ郊gC(-1: not used) */
	unsigned long	conflict_timeo;		/* OdoԊu(~0: not used) */


} NBTForceSettingInfo;

#define NBT_FORCESET_NOBCAST	(1<<0)	/* Oo^Ńu[hLXgsȂ */
#define NBT_FORCESET_NORESP		(1<<1)	/* O폜ŉ҂Ȃ */



///////////////// public function
#include "libnbt_cm.h"
#include "libnbt_ns.h"
#include "libnbt_ds.h"
#include "libnbt_ss.h"


#ifdef __cplusplus
}
#endif

#endif // _LIBNBT_H_
