/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_cm.c
 *------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#ifdef WIN32
#include <WinSock.h>

#else // WIN32
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#endif // WIN32

#include "libnbt.h"
#include "libnbt_cm.h"
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_mes.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_err.h"
#include "libnbt_debug.h"

#ifdef WIN32
#define close closesocket

#endif // WIN32

// private func def
static int inner_bind(int s, unsigned long naddr, unsigned short nport, unsigned long* exactnaddr);
static int _nbt_open_sub1(	int socket_type, int service_type,
							unsigned long srcnaddr, unsigned short srcnport,
							int* index, NBTCTBL** nct);



/*
------------------------------------------------------------------
inner_bind()

\Pbgbinds

߂l:

:	0
s:	-1(bind error)
------------------------------------------------------------------
*/
static int inner_bind(int s, unsigned long naddr, unsigned short nport, unsigned long* exactnaddr)
{
	// bind
	struct sockaddr_in inaddr;
	// char name[256];
	int res;

	*exactnaddr = 0;

	memset(&inaddr, 0, sizeof inaddr);
#ifndef WIN32
	inaddr.sin_len = sizeof inaddr;
#endif // WIN32
	inaddr.sin_family  = AF_INET;
	inaddr.sin_port = nport;
	inaddr.sin_addr = *(struct in_addr* )&naddr;
	// D_PRINTF(("BIND ADDR = %s\n", inet_ntoa(*(struct in_addr*)&naddr)));
	res = bind(s, (struct sockaddr* )&inaddr, sizeof inaddr);
	if(res == -1){
		// D_PRINTF(("BIND ERROR errno = %u\n", *__errno()));
		return -1;
	}

	if(ntohl(naddr) != INADDR_ANY){
		// IPAhXw肳Ă
		*exactnaddr = naddr;
	}
#if 0	/* ݃[`gp */

	else{
		u_long addr;

		// IPAhXIɎw肳ĂȂ
		res = gethostname(name, sizeof name);
		if(res < 0){
			return 0;
		}
		addr = inet_addr(name);
		if(addr != INADDR_NONE){
			*exactnaddr = addr;
		}
		else{
			struct hostent* he;
			he = gethostbyname(name);
			if(he == NULL){
				return 0;
			}
			*exactnaddr = *(u_long* )he->h_addr_list[0];
		}
	}
#endif

	return 0;
}


// public

/*
------------------------------------------------------------------
_nbt_open_sub1()

nbt_open[`

:
IN	socket_type:	\Pbg^Cv(SOCK_DGRAM, SOCK_STREAM)
OUT	index:			NBTCTBLe[ũCfbNX
OUT	nct:			VKɎ擾NBTCTBL\̂̃AhX


߂l:
:
LIBNBT_ERR_NOERR		(0)

s:
LIBNBT_ERR_SOCKET_ERR	(-210)
LIBNBT_ERR_NBTFD_MAX	(-120)
LIBNBT_ERR_BIND_ERR		(-220)
LIBNBT_ERR_FCNTL_ERR	(-250)
------------------------------------------------------------------
*/
static int _nbt_open_sub1(	int socket_type, int service_type,
							unsigned long srcnaddr, unsigned short srcnport,
							int* index, NBTCTBL** nct)
{
	int s;
	int a;
	int on;
	u_long enaddr;
	NBTCTBL* ct = NULL;
	
	*nct = NULL;
	*index = -1;
	
	s = socket(AF_INET, socket_type, 0);
	if(s < 0){
		return LIBNBT_ERR_SOCKET_ERR;
	}
	ct = getUnusedCtbl(index);
	if(ct == NULL){
		return LIBNBT_ERR_NBTFD_MAX;
	}
	
	// add ctbl
	initCtbl(ct);
	ct->isUsed = 1;
	ct->serviceType = service_type;
	ct->state = STATE_OPENED;
	ct->fd = s;

	// set socket option
	on = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char* )&on, sizeof on);
	on = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEPORT, (char* )&on, sizeof on);
	
	/* \PbgubN[hꍇ̓ubN[hɂ */
#ifndef WIN32
	{
		int fmode;
		fmode = fcntl(s, F_GETFL);
		if(fmode == -1){
			return LIBNBT_ERR_FCNTL_ERR;
		}
		
		if(fmode & O_NONBLOCK){
			fmode &= ~O_NONBLOCK;
			fmode = fcntl(s, F_SETFL, fmode);
			if(fmode == -1){
				return LIBNBT_ERR_FCNTL_ERR;
			}
		}
	}
#else // WIN32
	{
		unsigned long off = 0;
		ioctlsocket(ct->fd, FIONBIO, &off);
	}
#endif // WIN32

	// bind
	a = inner_bind(s, srcnaddr, srcnport, &enaddr);
	if((a < 0) /* || (enaddr == 0) */){
		// error
		initCtbl(ct);
		close(s);
		*nct = NULL;
		return LIBNBT_ERR_BIND_ERR;
	}
	ct->srcnaddr = enaddr;
	*nct = ct;
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_open()

socket֐ĂяoAʐM̏I_쐬


߂l:
:
NBTfBXNv^

s:
LIBNBT_ERR_SOCKET_ERR	(-210)
LIBNBT_ERR_NBTFD_MAX	(-120)
LIBNBT_ERR_BIND_ERR		(-220)

------------------------------------------------------------------
*/
int nbt_open(int serviceType, unsigned long srcnaddr, unsigned short srcnport)
{
	int res = -1;
	int index;
	NBTCTBL* ct;
	int on;
	
	switch(serviceType){
	case NBT_STYPE_NAME:
	case NBT_STYPE_DATAGRAM:
		res = _nbt_open_sub1(SOCK_DGRAM, serviceType,
							srcnaddr, srcnport, &index, &ct);
		if(res != LIBNBT_ERR_NOERR){
			return res;
		}
		/* Datagram̏ꍇ̓u[hLXg\ɂ悤 */

		on = 1;
		setsockopt(ct->fd, SOL_SOCKET, SO_BROADCAST, (char* )&on, sizeof on);
					
		res = encListNumber(index);
		
		break;
			
	case NBT_STYPE_SESSION:
		res = _nbt_open_sub1(SOCK_STREAM, serviceType,
							srcnaddr, srcnport, &index, &ct);
		if(res != LIBNBT_ERR_NOERR){
			return res;
		}
		res = encListNumber(index);
		
		break;
			
	default:
		D_PRINTF(("[nbtsub] no such service type\n"));
		break;
	}
	
	return res;
}


/*
------------------------------------------------------------------
nbt_close()

NBTfBXNv^B


߂l:
:
LIBNBT_ERR_NOERR	(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)

------------------------------------------------------------------
*/
int nbt_close(int s)
{
	NBTCTBL* ct;
	
	ct = getCtbl(decListNumber(s));
	if(ct == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	close(ct->fd);
	if(ct->cfd != -1){
		close(ct->cfd);
	}
	
	initCtbl(ct);
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getsockopt()

߂l:
:
LIBNBT_ERR_NOERR		(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_GETSOCKOPT_ERR	(-230)

------------------------------------------------------------------
*/
int nbt_getsockopt(int s, int level, int optname, void *optval, int *optlen)
{
	NBTCTBL* ct;
	int res;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	res = getsockopt(ct->fd, level, optname, optval, optlen);
	if(res == -1){
		return LIBNBT_ERR_GETSOCKOPT_ERR;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_setsockopt()

߂l:
:
LIBNBT_ERR_NOERR			(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_SETSOCKOPT_ERR	(-240)

------------------------------------------------------------------
*/
int nbt_setsockopt(int s, int level, int optname, const void *optval, int optlen)
{
	NBTCTBL* ct;
	int res;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	res = setsockopt(ct->fd, level, optname, optval, optlen);
	if(res == -1){
		return LIBNBT_ERR_SETSOCKOPT_ERR;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getfd()

߂l:
:
\PbgfBXNv^

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
------------------------------------------------------------------
*/
int nbt_getfd(int s)
{
	NBTCTBL* ct;
	
	ct = getCtbl(decListNumber(s));
	if(ct){
		return ct->fd;
	}
	
	return LIBNBT_ERR_INVALID_PARAM;
}


/*
------------------------------------------------------------------
nbt_getnbtfd()

߂l:
:
NBTfBXNv^

s:
LIBNBT_ERR_INVALID_PARAM	(-110)

------------------------------------------------------------------
*/
int nbt_getnbtfd(int s)
{
	int i;
	for(i=0; i<NBT_NBTCTBL_MAX; i++){
		if(NbtCtblList[i].fd == s){
			return encListNumber(i);
		}
	}
	
	return LIBNBT_ERR_INVALID_PARAM;
}


/*
------------------------------------------------------------------
nbt_setnodetype()

߂l:
:
LIBNBT_ERR_NOERR	(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
------------------------------------------------------------------
*/
int nbt_setnodetype(int nodetype)
{	

	if(
		(nodetype < 0) ||
		(nodetype > NBT_ONT_MAX)
	){
		return LIBNBT_ERR_INVALID_PARAM;
	}
		
	GSetting.ont = nodetype;
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getnodetype()

߂l:

m[h^Cv

------------------------------------------------------------------
*/
int nbt_getnodetype(void)
{

	return GSetting.ont;
}



/*
------------------------------------------------------------------
nbt_setwinsaddr()


߂l:
LIBNBT_ERR_NOERR
-----------------------------------------------------------------
*/
int nbt_setwinsaddr(unsigned long naddr)
{

	GSetting.winsnaddr = naddr;
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getwinsaddr()

߂l:
WINST[o[IPAhX
------------------------------------------------------------------
*/
int nbt_getwinsaddr(void)
{

	return GSetting.winsnaddr;
}


/*
------------------------------------------------------------------
nbt_setlocalwinsaddr()

NBTfBXNv^ŗLWINST[o[AhXݒ肷

߂l:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_NOERR

------------------------------------------------------------------
*/
int nbt_setlocalwinsaddr(int s, unsigned long naddr)
{
	NBTCTBL* ct;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct->winsnaddr = naddr;

	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getlocalwinsaddr()

NBTfBXNv^ŗLWINST[o[AhX擾

߂l:
:
	WINST[o[AhX

s:
	LIBNBT_ERR_INVALID_PARAM
	
------------------------------------------------------------------
*/
int nbt_getlocalwinsaddr(int s)
{
	NBTCTBL* ct;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return ct->winsnaddr;
}



/*
------------------------------------------------------------------
nbt_setscopeid()

߂l:
:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_INVALID_SCOPEID

------------------------------------------------------------------
*/
int nbt_setscopeid(const char* scopeid)
{
	int res;
	
	if((scopeid == NULL) || (strlen(scopeid) >= NBT_LENGTH_SCOPEID)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	res = nbt_checkvalidscopeid(scopeid);
	if(res < 0){
		return LIBNBT_ERR_INVALID_SCOPEID;
	}

	strcpy(GSetting.scopeid, scopeid);
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_getscopeid()


------------------------------------------------------------------
*/
const char* nbt_getscopeid(void)
{

	return GSetting.scopeid;
}



/*
------------------------------------------------------------------
nbt_setlocalscopeid()

߂l:
:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_MALLOC_FAILED
LIBNBT_ERR_INVALID_SCOPEID

------------------------------------------------------------------
*/
int nbt_setlocalscopeid(int s, const char* scopeid)
{
	NBTCTBL* ct;
	int slen;
	int res;

	if((scopeid == NULL) || ((slen = strlen(scopeid)) >= NBT_LENGTH_SCOPEID)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	res = nbt_checkvalidscopeid(scopeid);
	if(res < 0){
		return LIBNBT_ERR_INVALID_SCOPEID;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	if(ct->scopeid){
		int len;
		
		len = strlen(ct->scopeid);
		if(len < slen){
			void* v;
			// AP[g̕Kv
			v = realloc(ct->scopeid, len + 1);
			if(v == NULL){
				return LIBNBT_ERR_MALLOC_FAILED;
			}
			ct->scopeid = (char* )v;
		}
		strcpy(ct->scopeid, scopeid);
	}
	else{
		ct->scopeid = (char* )malloc(slen + 1);
		if(ct->scopeid == NULL){
			return LIBNBT_ERR_MALLOC_FAILED;
		}

		strcpy(ct->scopeid, scopeid);
	}

	return LIBNBT_ERR_NOERR;
}



/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
const char* nbt_getlocalscopeid(int s)
{
	NBTCTBL* ct;

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return NULL;
	}

	return ct->scopeid;
}



/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int nbt_setbcaddr(int s, unsigned long bcnaddr)
{
	NBTCTBL* ct;

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct->bcnaddr = bcnaddr;
	return LIBNBT_ERR_NOERR;
}



/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int nbt_getbcaddr(int s)
{
	NBTCTBL* ct;

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return ct->bcnaddr;
}



/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
void nbt_forcesettinginit(NBTForceSettingInfo* fcinfo)
{
	if(fcinfo){
		fcinfo->interruptd = -1;
		fcinfo->flag	= 0;
		fcinfo->bcast_retry_timeo	= ULONG_MAX;
		fcinfo->bcast_retry_count	= ULONG_MAX;
		fcinfo->ucast_retry_timeo	= ULONG_MAX;
		fcinfo->ucast_retry_count	= ULONG_MAX;
		fcinfo->conflict_timeo	= ULONG_MAX;
	}
}

