/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_ds.c
 *------------------------------------------------------------------*/

#include <stdio.h>
#include <limits.h>

#ifdef WIN32
#include <WinSock.h>
#include <errno.h>

#else // WIN32
#include <string.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#endif // WIN32

#include "libnbt.h"
#include "libnbt_ds.h"
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_mes.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_ns_in.h"
#include "libnbt_debug.h"

// private func def
#if 0
static int getDatagramMessageMacroFromValue(u_char value);
static u_char getDatagramMessageValueFromMacro(int macro);
#endif
static int getDGMErrorCodeMacroFromValue(u_char value);

static int senddatagram_prim(NBTCTBL* ct, void* buf, int size,
							 const NBTNamePack* srcnp, const NBTNamePack* dstnp);
static int senddatagram_pmh_node(NBTCTBL* ct, void* buf, int size,
								const NBTNamePack* srcnp, const NBTNamePack* dstnp);
static int senddatagram_b_node(NBTCTBL* ct, void* buf, int size,
							   const NBTNamePack* srcnp, const NBTNamePack* dstnp);

#if 0
static int recvdatagram_prim(int s, void* buf, int size,
							 const NBTNamePack* srcnp, NBTNamePack* dstnp, int* retval);
#endif

#if 0
// private
static int getDatagramMessageMacroFromValue(u_char value)
{
	int macro;
	switch(value){
	case VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM:	macro = NBT_MTYPE_DGM_UNICAST; break;
	case VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM:	macro = NBT_MTYPE_DGM_GROUP; break;
	case VAL_DSF_MTYPE_BROADCAST_DATAGRAM:		macro = NBT_MTYPE_DGM_BRDCAST; break;
	case VAL_DSF_MTYPE_DATAGRAM_ERROR:			macro = NBT_MTYPE_DGM_ERROR; break;
	case VAL_DSF_MTYPE_DATAGRAM_QUERY_REQUEST:	macro = NBT_MTYPE_DGM_QUERYREQ; break;
	case VAL_DSF_MTYPE_DATAGRAM_POSITIVE_QUERY_RESPONSE:	macro = NBT_MTYPE_DGM_QUERYRES_ACK; break;
	case VAL_DSF_MTYPE_DATAGRAM_NEGATIVE_QUERY_RESPONSE:	macro = NBT_MTYPE_DGM_QUERYRES_NACK; break;

	default:
		macro = -1;
		break;
	}
	
	return macro;
}

static u_char getDatagramMessageValueFromMacro(int macro)
{
	u_char value;

	switch(macro){
	case NBT_MTYPE_DGM_UNICAST:		value = VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM; break;
	case NBT_MTYPE_DGM_GROUP:		value = VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM; break;
	case NBT_MTYPE_DGM_BRDCAST:		value = VAL_DSF_MTYPE_BROADCAST_DATAGRAM; break;
	case NBT_MTYPE_DGM_ERROR:		value = VAL_DSF_MTYPE_DATAGRAM_ERROR; break;
	case NBT_MTYPE_DGM_QUERYREQ:	value = VAL_DSF_MTYPE_DATAGRAM_QUERY_REQUEST; break;
	case NBT_MTYPE_DGM_QUERYRES_ACK:	value = VAL_DSF_MTYPE_DATAGRAM_POSITIVE_QUERY_RESPONSE; break;
	case NBT_MTYPE_DGM_QUERYRES_NACK:	value = VAL_DSF_MTYPE_DATAGRAM_NEGATIVE_QUERY_RESPONSE; break;

	default:
		value = -1;
		break;
	}
	
	return value;
}
#endif


static int getDGMErrorCodeMacroFromValue(u_char value)
{
	int macro;

	switch(value){
	case VAL_DSF_ECODE_DSTNAME_NOT_PRESENT:		macro = NBT_ERROR_DGM_DSTNAME_NOT_PRESENT; break;
	case VAL_DSF_ECODE_INVALID_SRCNAME_FORMAT:	macro = NBT_ERROR_DGM_INVALID_SRCNAME_FORMAT; break;
	case VAL_DSF_ECODE_INVALID_DSTNAME_FORMAT:	macro = NBT_ERROR_DGM_INVALID_DSTNAME_FORMAT; break;

	default:
		macro = -1;
		break;
	}
	return macro;
}

// senddatagram֐{`
static int senddatagram_prim(NBTCTBL* ct, void* buf, int size,
							 const NBTNamePack* srcnp, const NBTNamePack* dstnp)
{
	//     --> B
	//     --> P, M
	//     --> H

	int res = -1;

	switch(GSetting.ont){
	case NBT_ONT_B:
		res = senddatagram_b_node(ct, buf, size, srcnp, dstnp);
		break;

	case NBT_ONT_P:
	case NBT_ONT_M:
	case NBT_ONT_H:
		res = senddatagram_pmh_node(ct, buf, size, srcnp, dstnp);
		break;

	default:
		D_PRINTF(("[libnbt] no such ont type\n"));
		return -1;
	}

	return res;
}


// #define BROADCAST_NAME	"*"

static int senddatagram_pmh_node(NBTCTBL* ct, void* buf, int size,
								const NBTNamePack* srcnp, const NBTNamePack* dstnp)
{
	// routine --> mestype --> create_DatagramFormat --> sendto
	
	u_char buff[MESSAGE_BUFSIZE];
	stDatagramPrimitive prim;
	u_long srcnaddr;
	u_long dstnaddr;
	u_char group;
	u_char* datap;
	
	NBTAddrPack	ap;
	int num;
	int findnum;

	int totalsend;
	int datasize;
	int headersize;
	int res;
	int i;
	int oncesend;



	/* src ipaddrZbgĂȂ΁AftHgIPAhXgpB */
	srcnaddr =
		(srcnp->naddr == 0)? ct->srcnaddr: srcnp->naddr;

	/* u[hLXgf[^OM? */
	if(strcmp(dstnp->name, BROADCAST_NAME) == 0){
		// broadcast
		dstnaddr = ct->bcnaddr;
		prim.msgtype = VAL_DSF_MTYPE_BROADCAST_DATAGRAM;

	}
	else{
		/* not broadcast */
		/* dst ipaddrZbgĂȂ΁AOsȂB */
		if(dstnp->naddr == 0){
			int ns_client;
			num = 1;
			ns_client = nbt_open(NBT_STYPE_NAME, ct->srcnaddr, htons(NBT_PORT_ANY));
			if(ns_client < 0){
				return -1;
			}
			/* scopeidݒ肷 */
			if(ct->scopeid){
				nbt_setlocalscopeid(ns_client, ct->scopeid);
			}
			/* broadcastAhXݒ肷 */
			nbt_setbcaddr(ns_client, ct->bcnaddr);

			/* winsT[o[AhXݒ肷 */
			if(ct->winsnaddr){
				nbt_setlocalwinsaddr(ns_client, ct->winsnaddr);
			}

			res = nbt_findname(ns_client, dstnp, &ap, 1, &findnum, NULL);
			if(res <= 0){
				D_PRINTF(("[libnbt] senddatagram_pmh_node: in_findname failed\n"));
				nbt_close(ns_client);
				return -1;
			}
			if(findnum != 1){
				D_PRINTF(("[libnbt] senddatagram_pmh_node: name conflict found. findnum = %d\n", findnum));
			}
			nbt_close(ns_client);

			dstnaddr = ap.naddr;

			if(ntohl(dstnaddr) == 0xFFFFFFFFUL){	// !!
				// group
				group = VAL_G_GROUP;
				prim.msgtype = VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM;

			}
			else{
				// unique
				group = VAL_G_UNIQUE;
				prim.msgtype = VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM;
			}
		}
		else{
			dstnaddr = dstnp->naddr;
			// ...
			group = dstnp->group;
			if(group == VAL_G_UNIQUE){
				prim.msgtype = VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM;
			}
			else{
				prim.msgtype = VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM;
			}
		}
	}

	datasize = 0;
	headersize = 0;

	prim.snt = GSetting.ont;
	prim.naddr = srcnaddr;
	prim.nport = nbt_getnport(ct->fd);
	nbt_padname(srcnp->name, prim.DP1_srcname, NBT_LENGTH_NETBIOSNAME, ' ', srcnp->suffix);
	datasize += NBT_LENGTH_NETBIOSNAME_2ENC;

	nbt_padname(dstnp->name, prim.DP1_dstname, NBT_LENGTH_NETBIOSNAME, ' ', dstnp->suffix);
	datasize += NBT_LENGTH_NETBIOSNAME_2ENC;

	strcpy(prim.DP1_scopeid, ct->scopeid? ct->scopeid: GSetting.scopeid);

	if(strlen(prim.DP1_scopeid) != 0){
		datasize += 2*(strlen(prim.DP1_scopeid) + 1);
	}
	
	headersize = datasize;

	datasize+=size;
	datap = (u_char* )buf;

	totalsend = 0;

	for(i=0; datasize > 0; i++){
		// set
		prim.id = (ct->tid == USHRT_MAX)? ct->tid = 0: ct->tid++;
		prim.first_flag = (i == 0)? 1: 0;
		if(datasize > MAX_DATA_LENGTH){	
			prim.more_flag = 1;
			prim.DP1_datalen = MAX_DATA_LENGTH;
		}
		else{
			prim.more_flag = 0;
			prim.DP1_datalen = datasize;
		}
		prim.DP1_data = datap;
		prim.DP1_offset = totalsend;

		memset(buff, 0, sizeof buff);
		res = create_DatagramFormat(buff, &prim);
		if(res < 0){
			D_PRINTF(("[libnbt] senddatagram_pmh_node: create_DatagramFormat failed\n"));
			return -1;
		}

		oncesend = nbt_in_sendto(ct->fd, buff, res, 0, dstnaddr, htons(NBT_PORT_DATAGRAM), NBT_WAIT_SEC, -1);
		if(oncesend != res){
			return -1;
		}
		if(i == 0){
			totalsend += (prim.DP1_datalen - headersize);
		}
		else{
			totalsend += prim.DP1_datalen;
		}

		datap += prim.DP1_datalen;
		datasize -= prim.DP1_datalen;
	}

	return totalsend;
}


static int senddatagram_b_node(NBTCTBL* ct, void* buf, int size,
							   const NBTNamePack* srcnp, const NBTNamePack* dstnp)
{
	// routine --> mestype --> create_DatagramFormat --> sendto
	return senddatagram_pmh_node(ct, buf, size, srcnp, dstnp);
}

#if 0
static int recvdatagram_prim(int s, void* buf, int size,
							 const NBTNamePack* srcnp, NBTNamePack* dstnp, int* retval)
{
	// routine --> get_DatagramFormat
	return -1;
}
#endif




// public

// UNI, MULTI, BROAD
int nbt_senddatagram(int s, void* buf, int size,
					 const NBTNamePack* srcnp, const NBTNamePack* dstnp)
{
	NBTCTBL* ct;
	
	// parameter check
	if((buf == NULL) || (size < 0) || (srcnp == NULL) || (dstnp == NULL)){
		return -1;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return -1;
	}
	if(ct->serviceType != NBT_STYPE_DATAGRAM){
		return -1;
	}
	
	return senddatagram_prim(ct, buf, size, srcnp, dstnp);

}


int nbt_sendbroadcastdatagram(int s, void* buf, int size, const NBTNamePack* srcnp)
{
	NBTCTBL* ct;
	NBTNamePack dstnp;
	
	// parameter check
	if((buf == NULL) || (size < 0) || (srcnp == NULL)){
		return -1;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return -1;
	}
	if(ct->serviceType != NBT_STYPE_DATAGRAM){
		return -1;
	}
	
	// set broadcast
	memset(&dstnp, 0, sizeof dstnp);
	strcpy(dstnp.name, BROADCAST_NAME);
	dstnp.suffix = 0x20;
	// dstnp.naddr = htonl(0xFFFFFFFFUL);
	dstnp.naddr = ct->bcnaddr;

	return senddatagram_prim(ct, buf, size, srcnp, &dstnp);
}


int nbt_recvdatagram(int s, void* buf, int size,
					 NBTNamePack* srcnp, NBTNamePack* dstnp, int* retval, int* ecode)
{
	NBTCTBL* ct;
	stDatagramPrimitive prim;
	u_char temp[MESSAGE_BUFSIZE];	// MpPbgf[^i[p
	int res;
	int packetlen;
	int totalrecv = 0;
	u_long rnaddr;
	u_short	rnport;

	u_short minoffset = USHRT_MAX;

	// parameter check
	if((buf == NULL) || (size <= 0) || (srcnp == NULL) || (dstnp == NULL) ||
		(retval == NULL) || (ecode == NULL)){
		return -2;
	}
	memset(srcnp, 0, sizeof(NBTNamePack));
	memset(dstnp, 0, sizeof(NBTNamePack));
	memset(buf, 0, size);

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return -2;
	}
	if(ct->serviceType != NBT_STYPE_DATAGRAM){
		return -2;
	}


	do{
		packetlen = nbt_in_recvfrom(ct->fd, temp, sizeof temp, 0, &rnaddr, &rnport, NBT_WAIT_SEC, -1);
		if(packetlen <= 0){
			// res == 0̏ꍇ́ANBTpPbgƂ͌ȂȂ
			return -2;
		}
		res = get_DatagramFormat(&prim, temp);
		if(res < 0){
			// ̓G[
			return -2;
		}

		switch(prim.msgtype){
		case VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM:
		case VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM:
		case VAL_DSF_MTYPE_BROADCAST_DATAGRAM:

			if(prim.first_flag){
				// check scopeid same?
				if(strcmp(prim.DP1_scopeid, ct->scopeid? ct->scopeid: GSetting.scopeid) != 0){
					// scopeid not same
					return -1;
				}
				// srcname
				nbt_devidename(srcnp->name, &srcnp->suffix, prim.DP1_srcname);
				srcnp->naddr = rnaddr;
				// dstname
				nbt_devidename(dstnp->name, &dstnp->suffix, prim.DP1_dstname);
			}
			
			if(prim.DP1_offset < minoffset){
				// ŏItZbg
				minoffset = prim.DP1_offset;
				// obt@̃f[^datalenɈړ
				if(prim.DP1_datalen < size){
					MemcpySafe((u_char*)buf + prim.DP1_datalen, buf, size - prim.DP1_datalen);
				}
				// 󂢂ƂɃpPbgf[^
				MemcpySafe(buf, prim.DP1_data, (prim.DP1_datalen < size)? prim.DP1_datalen: size);
			}
			else{
				prim.DP1_offset -= minoffset; // ItZbg␳
				// obt@ɃpPbgf[^ǉ
				if(prim.DP1_offset < size){
					MemcpySafe((u_char* )buf + prim.DP1_offset, prim.DP1_data,
						(int )(prim.DP1_datalen + prim.DP1_offset < size)? prim.DP1_datalen: size - prim.DP1_offset);
				}
			}
			
			totalrecv += prim.DP1_datalen;

			break;

		case VAL_DSF_MTYPE_DATAGRAM_ERROR:
			// 1 packet only
			res = getDGMErrorCodeMacroFromValue(prim.DP2_ecode);
			if(res < 0){
				return -1;
			}
			*ecode = res;
			totalrecv = 0;

			break;

		case VAL_DSF_MTYPE_DATAGRAM_QUERY_REQUEST:
		case VAL_DSF_MTYPE_DATAGRAM_POSITIVE_QUERY_RESPONSE:
		case VAL_DSF_MTYPE_DATAGRAM_NEGATIVE_QUERY_RESPONSE:
			// 1 packet only
			// check scopeid same?
			if(strcmp(prim.DP3_scopeid, ct->scopeid? ct->scopeid: GSetting.scopeid) != 0){
				// scopeid not same
				return -1;
			}
			nbt_devidename(dstnp->name, &dstnp->suffix, prim.DP3_dstname);

			totalrecv = 0;

			break;

		default:

		D_PRINTF(("[libnbt] int nbt_recvdatagram: no support msgtype %d\n", prim.msgtype));
		return -2;
		}

	} while(prim.more_flag);

	*retval = prim.msgtype;

	return totalrecv;
}
