/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_mes.c
 *------------------------------------------------------------------*/
 

#include <stdio.h>

#ifdef WIN32
#include <WinSock.h>
#include <errno.h>

#else // WIN32
#include <string.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#endif // WIN32


#include "libnbt.h"
#include "libnbt_private.h"
#include "libnbt_mes.h"
#include "libnbt_mes_private.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_macros.h"
#include "libnbt_err.h"
#include "libnbt_debug.h"


////////////////////////////////
// private function prototype
static int create_NameServicePacketHeader(u_char* out, const stNameServicePacketHeaderInfo* data, int* retval);
static int get_NameServicePacketHeader(stNameServicePacketHeaderInfo* data, const u_char* in, int* retval);
static int create_NameServiceQuestionSection(u_char* out, const stNameServiceQuestionSectionInfo* data, int* retval);
static int get_NameServiceQuestionSection(stNameServiceQuestionSectionInfo* data, const u_char* in, int* retval);
static int get_NameServiceQuestionSection2(stNameServiceQuestionSectionInfo2* data, const u_char* in, int* retval);
static int create_NameServiceResourceRecode(u_char* out, const stNameServiceResourceRecodeInfo* data, int* retval);
static int get_NameServiceResourceRecode(stNameServiceResourceRecodeInfo* data, const u_char* in, int* retval);
static int get_NameServiceResourceRecode2(stNameServiceResourceRecodeInfo2* data, const u_char* in, int* retval);
static int create_NameServiceAddrEntryArray(u_char* out, const stNameServiceAddrEntry* ent, int num);
static int get_NameServiceAddrEntryArray(stNameServiceAddrEntry* ent, int num, const u_char* in);
static int create_NameServiceNodeNameEntryArray(u_char* out, const stNameServiceNodeNameEntry* ent, int num);
static int create_NameServiceStatistics(u_char* out, const stNameServiceStatistics* sta);

static void get_NameServiceMessage_clean(stNameServiceMessageInfo* info);


////////////////////////////////

/*
------------------------------------------------------------------
create_NameServicePacketHeader()

wb_֐

:
	out:	bZ[Wރobt@
	data:	bZ[W̌ƂȂf[^
	retval:	񂾃obt@̃TCY(sizeof stNameServiceHeader)

߂l:
	0:	

------------------------------------------------------------------
*/
static int create_NameServicePacketHeader(u_char* out, const stNameServicePacketHeaderInfo* data, int* retval)
{
	stNameServiceHeader* h;
	
	*retval = 0;
	h = (stNameServiceHeader* )out;
	memset(h, 0, sizeof(stNameServiceHeader));

	h->id = htons(data->id);
	if(data->nmflags & NS_NSPR_NMFLAGS_AA_ON){ S_NSH_FLAGS_AA(h->flags, 1); }
	if(data->nmflags & NS_NSPR_NMFLAGS_TC_ON){ S_NSH_FLAGS_TC(h->flags, 1); }
	if(data->nmflags & NS_NSPR_NMFLAGS_RD_ON){ S_NSH_FLAGS_RD(h->flags, 1); }
	if(data->nmflags & NS_NSPR_NMFLAGS_RA_ON){ S_NSH_FLAGS_RA(h->flags, 1); }
	if(data->nmflags & NS_NSPR_NMFLAGS_B_ON){ S_NSH_FLAGS_B(h->flags, 1); }

	S_NSH_FLAGS_OC(h->flags, data->oc);
	S_NSH_FLAGS_R(h->flags, data->response);	
	S_NSH_FLAGS_RC(h->flags, data->rcode);

	h->qdc = htons(data->qdc);
	h->anc = htons(data->anc);
	h->nsc = htons(data->nsc);
	h->arc = htons(data->arc);

	*retval = sizeof(stNameServiceHeader);

	return 0;
}


/*
------------------------------------------------------------------
get_NameServicePacketHeader()

wb_ǂݎ֐

:
	data:	ǂݍ񂾃obt@̏i[\
	in:		wb_Ăobt@
	retval:	ǂݍ񂾃obt@TCY(sizeof(stNameServiceHeader))

߂l:
	0:	

------------------------------------------------------------------
*/
static int get_NameServicePacketHeader(stNameServicePacketHeaderInfo* data, const u_char* in, int* retval)
{
	stNameServiceHeader* h;

	*retval = 0;
	h = (stNameServiceHeader* )in;

	memset(data, 0, sizeof(stNameServicePacketHeaderInfo));

	data->id = ntohs(h->id);

	if(G_NSH_FLAGS_AA(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_AA_ON; }
	if(G_NSH_FLAGS_TC(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_TC_ON; }
	if(G_NSH_FLAGS_RD(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_RD_ON; }
	if(G_NSH_FLAGS_RA(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_RA_ON; }
	// if(G_NSH_FLAGS_G(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_G_ON; }
	if(G_NSH_FLAGS_B(h->flags)){ data->nmflags |= NS_NSPR_NMFLAGS_B_ON; }

	data->oc = G_NSH_FLAGS_OC(h->flags);
	data->response = G_NSH_FLAGS_R(h->flags);
	data->rcode = G_NSH_FLAGS_RC(h->flags);

	data->qdc = ntohs(h->qdc);
	data->anc = ntohs(h->anc);
	data->nsc = ntohs(h->nsc);
	data->arc = ntohs(h->arc);

	*retval = sizeof(stNameServiceHeader);

	return 0;
}



/*
------------------------------------------------------------------
create_NameServiceQuestionSection()

QuestionSection֐

:
	out:	ރobt@
	data:	ƂȂf[^
	retval:	outobt@ɏ񂾃TCY

߂l:
	0:	
	-1:	s

------------------------------------------------------------------
*/
static int create_NameServiceQuestionSection(u_char* out, const stNameServiceQuestionSectionInfo* data, int* retval)
{
	stNameServiceQuestionEntries* q;
	int res;

	*retval = 0;

	q = (stNameServiceQuestionEntries* )out;
	if(data->uselabel){
#if 0
		// xg
		u_short* la;
		//printf("[nbtlib] label index = %d\n", data->labelindex);
		la = (u_short* )out;
		*la = htons( (u_short) ((0xC0<<8) | data->labelindex) );
#else
		u_short a;
		// printf("[nbtlib] label index = %d\n", data->labelindex);
		
		a = htons( (u_short) ((0xC0<<8) | data->labelindex) );
		memcpy(q, &a, sizeof a);
#endif

		(u_char* )q += sizeof(u_short);
	}
	else{
		// xgȂ
		res = nbt_enc2nd((u_char* )q, 256, data->qname, data->qscopeid);
		if(res < 0){
			// D_PRINTF(("[libnbt] create_NameServiceQuestionSection: nbt_enc2nd failed\n"));
			return -1;
		}
		(u_char* )q += res;
	}
	
	q->qt = htons(data->qtype);
	q->qc = htons(data->qclass);

	*retval = (((u_char* )&q->qc) - out) + sizeof q->qc;

	return 0;

}


/*
------------------------------------------------------------------
get_NameServiceQuestionSection()

QuestionSectionFieldǂݎ֐

:
	data:
	in:
	retval:

߂l:
	0:	
	-1:	s
	1:	(XR[vIDw肳ꂽ\̂ɓ肫Ȃ)

------------------------------------------------------------------
*/
static int get_NameServiceQuestionSection(stNameServiceQuestionSectionInfo* data, const u_char* in, int* retval)
{
	stNameServiceQuestionEntries* q;
	int res;
	int readsize;
	int ret = 0;
	int lind;
	
	*retval = 0;
	lind = (int)data->labelindex;
	
	q = (stNameServiceQuestionEntries* )in;

	res = nbt_dec2nd(data->qname, data->qscopeid, data->qscopeid_len, in, data->top, &readsize,
		&data->uselabel, &lind);
	if(res < 0){
		if(res != -2){
			return -1;
		}
		ret = 1;
	}

	(u_char* )q += readsize;
	data->qtype = ntohs(q->qt);
	data->qclass = ntohs(q->qc);

	*retval = (((u_char* )&q->qc) - in) + sizeof q->qc;

	return ret;

}


/*
------------------------------------------------------------------
get_NameServiceQuestionSection2()

QuestionSectionFieldǂݎ֐(ver 2)

:
	data:
	in:
	retval:

߂l:
	0:	
	-1:	s
	1:	(XR[vIDw肳ꂽ\̂ɓ肫Ȃ)

------------------------------------------------------------------
*/
static int get_NameServiceQuestionSection2(stNameServiceQuestionSectionInfo2* data, const u_char* in, int* retval)
{
	stNameServiceQuestionEntries* q;
	int res;
	int readsize;
	int ret = 0;
	int lind;
	
	*retval = 0;
	lind = (int)data->labelindex;

	q = (stNameServiceQuestionEntries* )in;

	res = nbt_dec2nd(data->qname, data->qscopeid, sizeof data->qscopeid, in, data->top, &readsize,
		&data->uselabel, &lind);
	if(res < 0){
		if(res != -2){
			return -1;
		}
		ret = 1;
	}

	(u_char* )q += readsize;
	data->qtype = ntohs(q->qt);
	data->qclass = ntohs(q->qc);

	*retval = (((u_char* )&q->qc) - in) + sizeof q->qc;

	return ret;
}




// ResourceRecode֐
// rdata != NULL̏ꍇ́Af[^݂̏s
// retval:	񂾃obt@̃TCYԂ
//	߂l
//	0:	success
//	-1:	error

/*
------------------------------------------------------------------
create_NameServiceResourceRecode()

ResourceRecode֐

:
	out:
	data:
	retval:

߂l:
	0: 
	-1:	s

------------------------------------------------------------------
*/
static int create_NameServiceResourceRecode(u_char* out, const stNameServiceResourceRecodeInfo* data, int* retval)
{
	stNameServiceARREntries* r;
	int res;
	int readsize;

	*retval = 0;
	
	res = create_NameServiceQuestionSection(out, &data->s, &readsize);
	if(res < 0){
		return -1;
	}

	r = (stNameServiceARREntries*)(out + readsize - (sizeof(u_short)*2));
	r->ttl = htonl(data->ttl);
	r->rlen = htons(data->rdatalen);

	if(data->rdata){
		memcpy(r->rdata, data->rdata, data->rdatalen);
		*retval = ((u_char* )&r->rlen - out) + data->rdatalen + sizeof r->rlen;
	}
	else{
		*retval = ((u_char* )&r->rlen - out) + sizeof r->rlen;
	}

	return 0;

}


/*
------------------------------------------------------------------
get_NameServiceResourceRecode()

ResourceRecodeǂݎ֐

:

߂l:
	0:	success
	1:	success(scopeid)
	2:	success(data)
	-1:	error

------------------------------------------------------------------
*/
static int get_NameServiceResourceRecode(stNameServiceResourceRecodeInfo* data, const u_char* in, int* retval)
{
	stNameServiceARREntries* r;
	int res;
	int readsize;
	int rlen;
	int ret;

	*retval = 0;

	res = get_NameServiceQuestionSection(&data->s, in, &readsize);
	if(res < 0){
		return -1;
	}
	ret = res;

	r = (stNameServiceARREntries*)(in + readsize - (sizeof(u_short)*2));
	data->ttl = ntohl(r->ttl);
	rlen = ntohs(r->rlen);

	if((rlen > 0) && data->rdata){
		if(rlen > data->rdatalen){
			ret = 2;
			memcpy(data->rdata, r->rdata, data->rdatalen);
		}
		else{
			memcpy(data->rdata, r->rdata, rlen);
			data->rdatalen = rlen;
		}
		*retval = ((u_char* )&r->rlen - in) + data->rdatalen + sizeof r->rlen;
	}
	else{
		*retval = ((u_char* )&r->rlen - in) + sizeof r->rlen;
	}

	return ret;
}


/*
------------------------------------------------------------------
get_NameServiceResourceRecode2()

ResourceRecodeǂݎ֐(f[^͓ǂݎȂ)

:

߂l:
	0:	success
	1:	success(scopeidobt@ɓ肫Ȃ)
	-1:	error
------------------------------------------------------------------
*/
static int get_NameServiceResourceRecode2(stNameServiceResourceRecodeInfo2* data, const u_char* in, int* retval)
{
	stNameServiceARREntries* r;
	int res;
	int readsize;
	int ret;

	*retval = 0;

	res = get_NameServiceQuestionSection2(&data->s, in, &readsize);
	if(res < 0){
		return -1;
	}
	ret = res;

	r = (stNameServiceARREntries*)(in + readsize - (sizeof(u_short)*2));
	data->ttl = ntohl(r->ttl);
	data->rdatalen = ntohs(r->rlen);
	data->rdata = r->rdata;

	*retval = ((u_char* )&r->rlen - in) + sizeof r->rlen;

	return ret;

}





/*
------------------------------------------------------------------
ADDR_ENTRY ARRAY 쐬֐
------------------------------------------------------------------
*/
static int create_NameServiceAddrEntryArray(u_char* out, const stNameServiceAddrEntry* ent, int num)
{
	stNameServiceRRData* rd;
	int i;

	rd = (stNameServiceRRData* )out;

	for(i=0; i<num; i++){
		rd->flags = 0;
		S_NSR_FLAGS_G(rd->flags, ent[i].G);
		S_NSR_FLAGS_ONT(rd->flags, ent[i].ONT);
		rd->nbaddr = ent[i].nbaddr;

		rd = (stNameServiceRRData* )rd->next;
	}

	return (u_char* )rd - out;
}


/*
------------------------------------------------------------------
ADDR_ENTRY ARRAY 擾֐
------------------------------------------------------------------
*/
static int get_NameServiceAddrEntryArray(stNameServiceAddrEntry* ent, int num, const u_char* in)
{
	stNameServiceRRData* rd;
	int i;

	rd = (stNameServiceRRData* )in;
	
	for(i=0; i<num; i++){
		ent[i].G	= G_NSR_FLAGS_G(rd->flags);
		ent[i].ONT = G_NSR_FLAGS_ONT(rd->flags);
		ent[i].nbaddr = rd->nbaddr;

		rd = (stNameServiceRRData* )rd->next;
	}

	return (u_char* )rd - in;
}



/*
------------------------------------------------------------------
NODE_NAME ARRAY 쐬֐
------------------------------------------------------------------
*/
static int create_NameServiceNodeNameEntryArray(u_char* out, const stNameServiceNodeNameEntry* ent, int num)
{
	stNodeStatusResponseNodeNamePacket* rd;
	int i;
	u_char* p;
	
	// 擪ɖO̐
	p = out;
	*p = (u_char)num;
	p++;
	
	rd = (stNodeStatusResponseNodeNamePacket* )p;

	for(i=0; i<num; i++){
		memcpy(rd->nbname, ent[i].nbname, sizeof rd->nbname);
		rd->flags = 0;
		S_NSR_FLAGS_G(rd->flags, ent[i].G);
		S_NSR_FLAGS_ONT(rd->flags, ent[i].ONT);
		if(ent[i].state & NS_NNE_STATE_DRG_ON){ S_NSR_FLAGS_DRG(rd->flags, 1); }
		if(ent[i].state & NS_NNE_STATE_CNF_ON){ S_NSR_FLAGS_CNF(rd->flags, 1); }
		if(ent[i].state & NS_NNE_STATE_ACT_ON){ S_NSR_FLAGS_ACT(rd->flags, 1); }
		if(ent[i].state & NS_NNE_STATE_PRM_ON){ S_NSR_FLAGS_PRM(rd->flags, 1); }
		S_NSR_FLAGS_RESERVED(rd->flags, 0);

		rd = (stNodeStatusResponseNodeNamePacket* )rd->next;
	}
	
	return (u_char* )rd - out;
}

/*
------------------------------------------------------------------
NODE_NAME ARRAY 擾֐
------------------------------------------------------------------
*/
static int get_NameServiceNodeNameEntryArray(stNameServiceNodeNameEntry* ent, int num, int* findnum, const u_char* in)
{
	const stNodeStatusResponseNodeNamePacket* rd;
	int i;
	u_char	cnum;
	const u_char* p;

	memset(ent, 0, sizeof(stNameServiceNodeNameEntry) * num);
	p = in;
	cnum = *p;
	*findnum = (int)cnum;

	p++;
	rd = (stNodeStatusResponseNodeNamePacket* )p;
	
	for(i=0; i<(int)cnum; i++){
		if(i<num){
			memcpy(ent[i].nbname, rd->nbname, sizeof rd->nbname);
			ent[i].G	= G_NSR_FLAGS_G(rd->flags);
			ent[i].ONT	= G_NSR_FLAGS_ONT(rd->flags);
			if(G_NSR_FLAGS_DRG(rd->flags)){ ent[i].state |= NS_NNE_STATE_DRG_ON; }
			if(G_NSR_FLAGS_CNF(rd->flags)){ ent[i].state |= NS_NNE_STATE_CNF_ON; }
			if(G_NSR_FLAGS_ACT(rd->flags)){ ent[i].state |= NS_NNE_STATE_ACT_ON; }
			if(G_NSR_FLAGS_PRM(rd->flags)){ ent[i].state |= NS_NNE_STATE_PRM_ON; }
		}

		rd = (stNodeStatusResponseNodeNamePacket* )rd->next;
	}

	return (u_char* )rd - in;
}


/*
------------------------------------------------------------------
STATISTICS쐬֐
------------------------------------------------------------------
*/
static int create_NameServiceStatistics(u_char* out, const stNameServiceStatistics* sta)
{
	stNameServiceStatistics* t;
	
	t = (stNameServiceStatistics* )out;
	memset(t, 0, sizeof(stNameServiceStatistics));
	
	memcpy(t->UNIT_ID, sta->UNIT_ID, sizeof t->UNIT_ID); // UNIT_ID
	t->JUMPERS 						= sta->JUMPERS;
	t->TEST_RESULT 					= sta->TEST_RESULT;
	t->VERSION_NUMBER 				= htons(sta->VERSION_NUMBER);
	t->PERIOD_OF_STATISTICS 		= htons(sta->PERIOD_OF_STATISTICS);
	t->NUMBER_OF_CRCs 				= htons(sta->NUMBER_OF_CRCs);
	t->NUMBER_ALIGNMENT_ERRORS 		= htons(sta->NUMBER_ALIGNMENT_ERRORS);
	t->NUMBER_OF_COLLISIONS 		= htons(sta->NUMBER_OF_COLLISIONS);
	t->NUMBER_SEND_ABORTS 			= htons(sta->NUMBER_SEND_ABORTS);
	t->NUMBER_GOOD_SENDS 			= htonl(sta->NUMBER_GOOD_SENDS);
	t->NUMBER_GOOD_RECEIVES 		= htonl(sta->NUMBER_GOOD_RECEIVES);
	t->NUMBER_RETRANSMITS 			= htons(sta->NUMBER_RETRANSMITS);
	t->NUMBER_NO_RESOURCE_CONDITIONS= htons(sta->NUMBER_NO_RESOURCE_CONDITIONS);
	t->NUMBER_FREE_COMMAND_BLOCKS 	= htons(sta->NUMBER_FREE_COMMAND_BLOCKS);
	t->NUMBER_PENDING_SESSIONS 		= htons(sta->NUMBER_PENDING_SESSIONS);
	t->MAX_NUMBER_PENDING_SESSIONS 	= htons(sta->MAX_NUMBER_PENDING_SESSIONS);
	t->MAX_TOTAL_SESSIONS_POSSIBLE 	= htons(sta->MAX_TOTAL_SESSIONS_POSSIBLE);
	t->SESSION_DATA_PACKET_SIZE 	= htons(sta->SESSION_DATA_PACKET_SIZE);

	return sizeof(stNameServiceStatistics);
}

/*
------------------------------------------------------------------
STATISTICS擾֐
------------------------------------------------------------------
*/
static int get_NameServiceStatistics(stNameServiceStatistics* sta, const u_char* in)
{
	stNameServiceStatistics* t;
	
	t = (stNameServiceStatistics* )in;
	
	if(sta){
		memset(sta, 0, sizeof(stNameServiceStatistics));

		memcpy(sta->UNIT_ID, t->UNIT_ID, sizeof sta->UNIT_ID); // UNIT_ID
		sta->JUMPERS 					= t->JUMPERS;
		sta->TEST_RESULT 				= t->TEST_RESULT;
		sta->VERSION_NUMBER 			= ntohs(t->VERSION_NUMBER);
		sta->PERIOD_OF_STATISTICS 		= ntohs(t->PERIOD_OF_STATISTICS);
		sta->NUMBER_OF_CRCs 			= ntohs(t->NUMBER_OF_CRCs);
		sta->NUMBER_ALIGNMENT_ERRORS 	= ntohs(t->NUMBER_ALIGNMENT_ERRORS);
		sta->NUMBER_OF_COLLISIONS 		= ntohs(t->NUMBER_OF_COLLISIONS);
		sta->NUMBER_SEND_ABORTS 		= ntohs(t->NUMBER_SEND_ABORTS);
		sta->NUMBER_GOOD_SENDS 			= ntohl(t->NUMBER_GOOD_SENDS);
		sta->NUMBER_GOOD_RECEIVES 		= ntohl(t->NUMBER_GOOD_RECEIVES);
		sta->NUMBER_RETRANSMITS 		= ntohs(t->NUMBER_RETRANSMITS);
		sta->NUMBER_NO_RESOURCE_CONDITIONS= ntohs(t->NUMBER_NO_RESOURCE_CONDITIONS);
		sta->NUMBER_FREE_COMMAND_BLOCKS = ntohs(t->NUMBER_FREE_COMMAND_BLOCKS);
		sta->NUMBER_PENDING_SESSIONS 	= ntohs(t->NUMBER_PENDING_SESSIONS);
		sta->MAX_NUMBER_PENDING_SESSIONS = ntohs(t->MAX_NUMBER_PENDING_SESSIONS);
		sta->MAX_TOTAL_SESSIONS_POSSIBLE = ntohs(t->MAX_TOTAL_SESSIONS_POSSIBLE);
		sta->SESSION_DATA_PACKET_SIZE 	= ntohs(t->SESSION_DATA_PACKET_SIZE);
	}
	
	return sizeof(stNameServiceStatistics);
}


/*
------------------------------------------------------------------
get_NameServiceMesssage()

ėpbZ[Wǂݎ֐

get_*RequestɂĎgp

:

߂l:
	
------------------------------------------------------------------
*/

/*
------------------------------------------------------------------
Tu[`()
------------------------------------------------------------------
*/
static void get_NameServiceMessage_clean(stNameServiceMessageInfo* info)
{
	if(info){
		if(info->qd){	free(info->qd);	info->qd = NULL;	}
		if(info->an){	free(info->an);	info->an = NULL;	}
		if(info->ns){	free(info->ns);	info->ns = NULL;	}
		if(info->ar){	free(info->ar);	info->ar = NULL;	}
	}
}


int get_NameServiceMessage(stNameServiceMessageInfo* info, const u_char* in, int* retval)
{
	int res;
	int readsize;
	int total = 0;
	int i;
	const u_char* p;

	*retval = 0;
	p = in;

	memset(info, 0, sizeof(stNameServiceMessageInfo));

	/* ŏɃwb_ǂݎ */
	res = get_NameServicePacketHeader(&info->h, p, &readsize);
	if(res < 0){
		get_NameServiceMessage_clean(info);
		return -1;
	}
	p += readsize;
	total += readsize;
	
	/* Question Entriesǂݎ */
	if(info->h.qdc){
		info->qd = (stNameServiceQuestionSectionInfo2*)malloc(sizeof(stNameServiceQuestionSectionInfo2) * info->h.qdc);
		if(info->qd == NULL){
			D_PRINTF(("[libnbt] malloc failed\n"));
			get_NameServiceMessage_clean(info);
			return -1;
		}
		for(i=0; i<info->h.qdc; i++){
			info->qd[i].top = (u_char* )in;	/* NBTwb_擪̃AhXێ */
			res = get_NameServiceQuestionSection2(&info->qd[i], p, &readsize);
			if(res < 0){
				get_NameServiceMessage_clean(info);
				return -1;
			}
			p += readsize;
			total += readsize;
		}
	}

	/* Answer Resource Recordsǂݎ */
	if(info->h.anc){
		info->an = (stNameServiceResourceRecodeInfo2*)malloc(sizeof(stNameServiceResourceRecodeInfo2) * info->h.anc);
		if(info->an == NULL){
			D_PRINTF(("[libnbt] malloc failed\n"));
			get_NameServiceMessage_clean(info);
			return -1;
		}
		for(i=0; i<info->h.anc; i++){
			info->an[i].s.top = (u_char* )in;
			res = get_NameServiceResourceRecode2(&info->an[i], p, &readsize);
			if(res < 0){
				get_NameServiceMessage_clean(info);
				return -1;
			}
			p += readsize;
			total += readsize;
		}
	}

	/* Authority Resource Recordsǂݎ */
	if(info->h.nsc){
		info->ns = (stNameServiceResourceRecodeInfo2*)malloc(sizeof(stNameServiceResourceRecodeInfo2) * info->h.nsc);
		if(info->ns == NULL){
			D_PRINTF(("[libnbt] malloc failed\n"));
			get_NameServiceMessage_clean(info);
			return -1;
		}
		for(i=0; i<info->h.nsc; i++){
			info->ns[i].s.top = (u_char* )in;
			res = get_NameServiceResourceRecode2(&info->ns[i], p, &readsize);
			if(res < 0){
				get_NameServiceMessage_clean(info);
				return -1;
			}
			p += readsize;
			total += readsize;
		}
	}

	/* read Additional Resource Records */
	if(info->h.arc){
		info->ar = (stNameServiceResourceRecodeInfo2*)malloc(sizeof(stNameServiceResourceRecodeInfo2) * info->h.arc);
		if(info->ar == NULL){
			D_PRINTF(("[libnbt] malloc failed\n"));
			get_NameServiceMessage_clean(info);
			return -1;
		}
		for(i=0; i<info->h.arc; i++){
			info->ar[i].s.top = (u_char* )in;
			res = get_NameServiceResourceRecode2(&info->ar[i], p, &readsize);
			if(res < 0){
				get_NameServiceMessage_clean(info);
				return -1;
			}
			p += readsize;
			total += readsize;
		}
	}

	*retval = total;

	return 0;
}




/*
------------------------------------------------------------------
get_NameServiceMessageType()

bZ[W^Cv֐

:
	in:	bZ[Wobt@

߂l:
	bZ[W^Cv

------------------------------------------------------------------
*/
int get_NameServiceMessageType(const char* in)
{
	stNameServicePacketHeaderInfo info;
	int res;
	int readsize;
	int temp;

	int ret = NBT_NS_MT_UNKNOWN;

	res = get_NameServicePacketHeader(&info, in, &readsize);
	if(res < 0){
		return NBT_NS_MT_UNKNOWN;
	}

	// 1. Request or Response
	if(info.response == 0){
		// Request
		// 2. opcode
		switch(info.oc){

		case VAL_OC_REGISTRATION:
			if(info.nmflags & NS_NSPR_NMFLAGS_RD_ON){
				// NameRegReq
				ret = NBT_NS_MT_REQ_NAMEREG;
			}
			else{
				// NameOwreq
				ret = NBT_NS_MT_REQ_NAMEOW;
			}

			break;

		case VAL_OC_REFRESH:
			ret = NBT_NS_MT_REQ_REF;

			break;

		case VAL_OC_QUERY:
			{
				stNameServiceQuestionSectionInfo a;
				
				if(info.qdc != 1){
					return NBT_NS_MT_UNKNOWN;
				}

				memset(&a, 0, sizeof a);
				res = get_NameServiceQuestionSection(&a, in + readsize, &temp);
				if(res < 0){
					return NBT_NS_MT_UNKNOWN;
				}

				if(a.qtype == 0x20){
					ret = NBT_NS_MT_REQ_NAMEQUERY;
				}
				else if(a.qtype == 0x21){
					ret = NBT_NS_MT_REQ_NODESTATUS;
				}

			}

			break;

		case VAL_OC_MULTIHOMED:
			ret = NBT_NS_MT_REQ_MHNAMEREG;

			break;

		case VAL_OC_RELEASE:
			ret = NBT_NS_MT_REQ_REL;

			break;

		default:
			ret = NBT_NS_MT_UNKNOWN;

			break;

		}
	}
	else{
		// Response
		// 2. opcode
		switch(info.oc){

		case VAL_OC_REGISTRATION:
			if(info.nmflags & NS_NSPR_NMFLAGS_RA_ON){
				if(info.rcode == VAL_RC_SUCCESS){
					ret = NBT_NS_MT_RES_POS_NAMEREG;
				}
				else{
					if(info.rcode == VAL_RC_CFT_ERR){
						ret = NBT_NS_MT_RES_NAMECFTDEMAND;
					}
					else{
						ret = NBT_NS_MT_RES_NEG_NAMEREG;
					}
				}
			}
			else{
				ret = NBT_NS_MT_RES_ENCREG;
			}

			break;

		case VAL_OC_QUERY:
			if(info.nmflags & NS_NSPR_NMFLAGS_AA_ON){
				stNameServiceResourceRecodeInfo a;

				if(info.anc != 1){
					return NBT_NS_MT_UNKNOWN;
				}
			
				memset(&a, 0, sizeof a);
				res = get_NameServiceResourceRecode(&a, in + readsize, &temp);
				if(res < 0){
					return NBT_NS_MT_UNKNOWN;
				}

				if(a.s.qtype == 0x20){
					if(info.rcode == VAL_RC_SUCCESS){
						ret = NBT_NS_MT_RES_POS_NAMEQUERY;
					}
					else{
						ret = NBT_NS_MT_RES_NEG_NAMEQUERY;
					}
				}
				else if(a.s.qtype == 0x21){
					ret = NBT_NS_MT_RES_NODESTATUS;
				}

			}
			else{
				// RedNameQueryRes
				ret = NBT_NS_MT_RES_REDNAMEQUERY;
			}

			break;

		case VAL_OC_RELEASE:
			if(info.rcode == VAL_RC_SUCCESS){
				ret = NBT_NS_MT_RES_POS_NAMEREL;
			}
			else{
				ret = NBT_NS_MT_RES_NEG_NAMEREL;
			}

			break;

		case VAL_OC_WACK:
			ret = NBT_NS_MT_RES_WACK;

			break;

		default:
			ret = NBT_NS_MT_UNKNOWN;
		}
	}

	return ret;

}



// public
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameRegistrationRequest(u_char* out, const stNameRegistrationRequest* data)
{
	int res;
	u_char* q;
	u_char* p;
	stNameServiceAddrEntry ent;
	stNameServiceResourceRecodeInfo rinfo;
	stNameServicePacketHeaderInfo hinfo;
	int total = 0;
	int readsize;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= data->TranID;
	hinfo.response	= VAL_R_REQUEST;
	hinfo.qdc		= 1;
	hinfo.arc		= 1;
	hinfo.nmflags	= data->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_REGISTRATION;

	p = out;
	res = create_NameServicePacketHeader(p, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}
	total += readsize;
	p += readsize;
	q = p;

	memset(&rinfo, 0, sizeof rinfo);
	rinfo.s.qtype			= 0x20;
	rinfo.s.qclass			= 0x01;
	rinfo.s.qname			= (u_char* )data->QName;
	rinfo.s.qscopeid		= (u_char* )data->QScopeId;
	rinfo.s.qscopeid_len	= sizeof data->QScopeId;

	res = create_NameServiceQuestionSection(p, &rinfo.s, &readsize);
	if(res < 0){
		return -1;
	}
	total += readsize;
	p += readsize;

	rinfo.s.uselabel	= 1;
	rinfo.s.labelindex	= q - out;
	rinfo.s.top			= out;
	rinfo.ttl			= data->TTL;
	rinfo.rdatalen		= 0x06;

	res = create_NameServiceResourceRecode(p, &rinfo, &readsize);
	if(res < 0){
		return -1;
	}

	ent.nbaddr	= data->NbAddress;
	ent.G		= data->G;
	ent.ONT		= data->ONT;

	total += readsize;
	p += readsize;


	res = create_NameServiceAddrEntryArray(p, &ent, 1);
	if(res < 0){
		return -1;
	}
	total += res;

	return total;
}


// multihomed
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_MultihomedNameRegistrationRequest(u_char* out, const stNameRegistrationRequest* data)
{
	stNameServiceHeader* h;
	int res;
	res = create_NameRegistrationRequest(out, data);
	if(res >= 0){
		h = (stNameServiceHeader* )out;
		C_NSH_FLAGS_OC(h->flags);
		S_NSH_FLAGS_OC(h->flags, VAL_OC_MULTIHOMED);
	}

	return res;
}



// NameOverwriteRequestbZ[W쐬֐
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameOverwriteRequest(u_char* out, const stNameOverwriteRequest* data)
{
	return create_NameRegistrationRequest(out, data);
}


// NameRefreshRequestbZ[W쐬֐
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameRefreshRequest(u_char* out, const stNameRefreshRequest* data)
{
	stNameServiceHeader* h;
	int res;
	
	res = create_NameRegistrationRequest(out, data);
	if(res >= 0){
		h = (stNameServiceHeader* )out;
		C_NSH_FLAGS_OC(h->flags);
		S_NSH_FLAGS_OC(h->flags, VAL_OC_REFRESH);
	}
	
	return res;
}


// NameReleaseRequestbZ[W쐬֐
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameReleaseRequest(u_char* out, const stNameReleaseRequest* data)
{
	stNameServiceHeader* h;
	int res;
	
	res = create_NameRegistrationRequest(out, data);
	if(res >= 0){
		h = (stNameServiceHeader* )out;
		C_NSH_FLAGS_OC(h->flags);
		S_NSH_FLAGS_OC(h->flags, VAL_OC_RELEASE);
	}
	
	return res;	
}


// NameQueryRequestbZ[W쐬֐
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameQueryRequest(u_char* out, const stNameQueryRequest* data)
{
	stNameServicePacketHeaderInfo hinfo;
	stNameServiceQuestionSectionInfo qinfo;
	int res;
	int total = 0;
	int readsize;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= data->TranID;
	hinfo.response	= VAL_R_REQUEST;
	hinfo.qdc		= 1;
	hinfo.nmflags	= data->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_QUERY;

	// create header
	res = create_NameServicePacketHeader(out, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}

	// create question entries
	total += readsize;
	out += readsize;

	memset(&qinfo, 0, sizeof qinfo);
	qinfo.qtype		= 0x20;
	qinfo.qclass	= 0x01;
	qinfo.qname		= (u_char* )data->QName;
	qinfo.qscopeid	= (u_char* )data->QScopeId;
	qinfo.qscopeid_len	= sizeof data->QScopeId;

	res = create_NameServiceQuestionSection(out, &qinfo, &readsize);
	if(res < 0){
		return -1;
	}

	total += readsize;

	return total;
}


//
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NodeStatusRequest(u_char* out, stNodeStatusRequest* data)
{
	stNameServicePacketHeaderInfo hinfo;
	stNameServiceQuestionSectionInfo qinfo;
	int res;
	int readsize;
	int total = 0;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= data->TranID;
	hinfo.response	= VAL_R_REQUEST;
	hinfo.qdc		= 1;
	hinfo.nmflags	= data->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_QUERY;

	// create header
	res = create_NameServicePacketHeader(out, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}
	
	// create question entries
	total += readsize;
	out += readsize;

	memset(&qinfo, 0, sizeof qinfo);
	qinfo.qtype		= 0x21;
	qinfo.qclass	= 0x01;
	qinfo.qname		= data->QName;
	qinfo.qscopeid	= data->QScopeId;
	qinfo.qscopeid_len	= sizeof data->QScopeId;

	res = create_NameServiceQuestionSection(out, &qinfo, &readsize);
	if(res < 0){
		return -1;
	}

	total += readsize;

	return total;
}



// resp
/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_PositiveNameRegistrationResponse(u_char* out, const stNameRegistrationRequest* data)
{
	stNameServicePacketHeaderInfo hinfo;
	stNameServiceResourceRecodeInfo rinfo;
	stNameServiceAddrEntry	ent;
	int res;
	int readsize;
	int total = 0;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= data->TranID;
	hinfo.response	= VAL_R_RESPONSE;
	hinfo.anc		= 1;
	hinfo.nmflags	= data->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_REGISTRATION;

	// create header
	res = create_NameServicePacketHeader(out, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;

	// create anser
	memset(&rinfo, 0, sizeof rinfo);
	rinfo.s.qtype		= 0x20;
	rinfo.s.qclass		= 0x01;
	rinfo.s.qname		= (u_char* )data->RRName;
	rinfo.s.qscopeid	= (u_char* )data->RScopeId;
	rinfo.s.qscopeid_len= sizeof data->RScopeId;
	rinfo.ttl			= data->TTL;
	rinfo.rdatalen		= 0x06;

	res = create_NameServiceResourceRecode(out, &rinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;
	
	ent.nbaddr = data->NbAddress;
	ent.G = data->G;
	ent.ONT = data->ONT;

	res = create_NameServiceAddrEntryArray(out, &ent, 1);
	if(res < 0){
		return -1;
	}

	total += res;

	return total;

}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NegativeNameRegistrationResponse(u_char* out, const stNameRegistrationRequest* data)
{
	int res;
	res = create_PositiveNameRegistrationResponse(out, data);
	if(res >= 0){
		stNameServiceHeader* h;
		h = (stNameServiceHeader* )out;
		S_NSH_FLAGS_RC(h->flags, data->Rcode);

	}

	return res;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_EndNodeChallengeRegistrationResponse(u_char* out, const stNameRegistrationRequest* data)
{
	// this is same to PositiveNameRegistrationResponse

	return create_PositiveNameRegistrationResponse(out, data);
}



/*
------------------------------------------------------------------
NameRegistrationResponsebZ[W쐬֐
------------------------------------------------------------------
*/
int create_NameRegistrationResponse(u_char* out, int type, const stNameRegistrationRequest* data)
{
	int res;

	switch(type){
	case NBT_NS_MT_RES_POS_NAMEREG:
		res = create_PositiveNameRegistrationResponse(out, data);
		break;

	case NBT_NS_MT_RES_NEG_NAMEREG:
		res = create_NegativeNameRegistrationResponse(out, data);
		break;

	case NBT_NS_MT_RES_ENCREG:
		res = create_EndNodeChallengeRegistrationResponse(out, data);
		break;

	default:
		res = -1;
		break;
	}

	return res;
}



/*
------------------------------------------------------------------
create_PositiveNameReleaseResponse
------------------------------------------------------------------
*/
int create_PositiveNameReleaseResponse(u_char* out, const stNameReleaseRequest* data)
{
	int res;
	res = create_PositiveNameRegistrationResponse(out, data);
	if(res >= 0){
		stNameServiceHeader* h;
		h = (stNameServiceHeader* )out;
		S_NSH_FLAGS_OC(h->flags, VAL_OC_RELEASE);
	}

	return res;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NegativeNameReleaseResponse(u_char* out, const stNameReleaseRequest* data)
{
	int res;
	res = create_NegativeNameRegistrationResponse(out, data);
	if(res <= 0){
		stNameServiceHeader* h;
		h = (stNameServiceHeader* )out;
		S_NSH_FLAGS_OC(h->flags, VAL_OC_RELEASE);
	}

	return res;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameReleaseResponse(u_char* out, int type, const stNameReleaseRequest* data)
{
	int res;

	switch(type){
	case NBT_NS_MT_RES_POS_NAMEREL:
		res = create_PositiveNameReleaseResponse(out, data);
		break;

	case NBT_NS_MT_RES_NEG_NAMEREL:
		res = create_NegativeNameReleaseResponse(out, data);
		break;

	default:
		res = -1;
		break;
	}

	return res;
}



/*
------------------------------------------------------------------
OЉimj֐
------------------------------------------------------------------
*/
int create_PositiveNameQueryResponse(u_char* out, const stNameQueryRequest* dh, const stNameServiceAddrEntry* dd, int num)
{
	stNameServicePacketHeaderInfo hinfo;
	stNameServiceResourceRecodeInfo rinfo;

	int res;
	int readsize;
	int total = 0;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= dh->TranID;
	hinfo.response	= VAL_R_RESPONSE;
	hinfo.anc		= 1;
	hinfo.nmflags	= dh->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_QUERY;

	// create header
	res = create_NameServicePacketHeader(out, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;

	// create anser
	memset(&rinfo, 0, sizeof rinfo);
	rinfo.s.qtype		= 0x20;
	rinfo.s.qclass		= 0x01;
	rinfo.s.qname		= (u_char* )dh->RRName;
	rinfo.s.qscopeid	= (u_char* )dh->RScopeId;
	rinfo.s.qscopeid_len= sizeof dh->RScopeId;
	rinfo.ttl			= dh->TTL;
	rinfo.rdatalen		= sizeof(stNameServiceRRData)*num;

	res = create_NameServiceResourceRecode(out, &rinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;

	res = create_NameServiceAddrEntryArray(out, dd, num);
	if(res < 0){
		return -1;
	}

	total += res;

	return total;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NegativeNameQueryResponse(u_char* out, const stNameQueryRequest* dh)
{
	int res;

	res = create_PositiveNameQueryResponse(out, dh, NULL, 0);
	if(res >= 0){
		stNameServiceHeader* h;
		h = (stNameServiceHeader* )out;
		S_NSH_FLAGS_RC(h->flags, dh->Rcode);
	}

	return res;

}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_RedirectNameQueryResponse(u_char* out, const stNameQueryRequest* dh)
{

	// not support
	return -1;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NameQueryResponse(u_char* out, int type, const stNameQueryRequest* dh, const stNameServiceAddrEntry* dd, int num)
{
	int res;

	switch(type){
	case NBT_NS_MT_RES_POS_NAMEQUERY:
		res = create_PositiveNameQueryResponse(out, dh, dd, num);
		break;

	case NBT_NS_MT_RES_NEG_NAMEQUERY:
		res = create_NegativeNameQueryResponse(out, dh);
		break;

	case NBT_NS_MT_RES_REDNAMEQUERY:
		res = create_RedirectNameQueryResponse(out, dh);
		break;

	default:
		res = -1;
	}

	return res;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int create_NodeStatusResponse(u_char* out, const stNodeStatusRequest* dh,
							  const stNameServiceNodeNameEntry* dnn, int num, const stNameServiceStatistics* dns)
{
	stNameServicePacketHeaderInfo hinfo;
	stNameServiceResourceRecodeInfo rinfo;


	int res;
	int readsize;
	int total = 0;

	memset(&hinfo, 0, sizeof hinfo);
	hinfo.id		= dh->TranID;
	hinfo.response	= VAL_R_RESPONSE;
	hinfo.anc		= 1;
	hinfo.nmflags	= dh->nmflags;
	hinfo.rcode		= 0;
	hinfo.oc		= VAL_OC_QUERY;

	// create header
	res = create_NameServicePacketHeader(out, &hinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;

	// create anser
	memset(&rinfo, 0, sizeof rinfo);
	rinfo.s.qtype		= 0x21;
	rinfo.s.qclass		= 0x01;
	rinfo.s.qname		= (u_char* )dh->RRName;
	rinfo.s.qscopeid	= (u_char* )dh->RScopeId;
	rinfo.s.qscopeid_len= sizeof dh->RScopeId;
	rinfo.ttl			= 0;
	rinfo.rdatalen		= 1 + sizeof(stNameServiceNodeNameEntry)*num + sizeof(stNameServiceStatistics);

	res = create_NameServiceResourceRecode(out, &rinfo, &readsize);
	if(res < 0){
		return -1;
	}

	out += readsize;
	total += readsize;

	// create node name array
	res = create_NameServiceNodeNameEntryArray(out, dnn, num);
	if(res < 0){
		return -1;
	}
	total += res;
	out += res;

	// create node statistics
	res = create_NameServiceStatistics(out, dns);
	if(res < 0){
		return -1;
	}
	total += res;

	return total;
}

/*
------------------------------------------------------------------
bZ[Wǂݎ

------------------------------------------------------------------
*/
int get_xReadMessage(stNameServicePrimitiveRequest* data, stNameServiceMessageInfo* info, const u_char* in)
{
	int res;
	int readsize;

	memset(info, 0, sizeof(stNameServiceMessageInfo));
	memset(data, 0, sizeof(stNameServicePrimitiveRequest));

	/* tB[hinfoɑ */
	res = get_NameServiceMessage(info, in, &readsize);
	if(res < 0){
		return -1;
	}

	data->TranID	= info->h.id;
	data->nmflags	= info->h.nmflags;
	data->Rcode		= info->h.rcode;

	/* Question Entries͖{֐̊Odataɑ */
	return 0;
}



/* get_*Request֐ŚA\߃bZ[W^Cv𒲂ׂĂǍʂɉ֐ĂԕKvB */


int get_NameRegistrationRequest(stNameRegistrationRequest* data, const u_char* in)
{
	stNameServiceMessageInfo info;
	stNameServiceAddrEntry	ent;

	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* NameRegistration Request̏ꍇ́Aqdcarc1ƂȂ */
	if(!(	(info.h.qdc == 1)	||
			(info.h.anc == 0)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 1)	)){
		get_NameServiceMessage_clean(&info);
		return -1;
	}

	/* Question Entries */
	memcpy(data->QName, info.qd[0].qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->QScopeId, info.qd[0].qscopeid, NBT_LENGTH_SCOPEID);
	
	/* Additional Resource Record */
	memcpy(data->RRName, info.ar[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.ar[0].s.qscopeid, NBT_LENGTH_SCOPEID);
	data->TTL = info.ar[0].ttl;
	
	/* Additional Resource Record̃f[^ǂݎ܂ */
	res = get_NameServiceAddrEntryArray(&ent, 1, info.ar[0].rdata);
	get_NameServiceMessage_clean(&info);
	if(res < 0){
		return -1;
	}
	data->G = ent.G;
	data->ONT = ent.ONT;
	data->NbAddress = ent.nbaddr;

	return 0;
}


int get_NameOverwriteRequest(stNameOverwriteRequest* data, const u_char* in)
{
	/* almost same as get_NameRegistrationRequest */
	return get_NameRegistrationRequest(data, in);
}

int get_NameRefreshRequest(stNameRefreshRequest* data, const u_char* in)
{
	/* almost same as get_NameRegistrationRequest */
	return get_NameRegistrationRequest(data, in);
}


int get_NameReleaseRequest(stNameReleaseRequest* data, const u_char* in)
{
	/* almost same as get_NameRegistrationRequest */
	return get_NameRegistrationRequest(data, in);
}


int get_NameQueryRequest(stNameQueryRequest* data, const u_char* in)
{
	stNameServiceMessageInfo info;
	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* NameQuery Request̏ꍇ́Aqdc1ƂȂ */
	if(!(	(info.h.qdc == 1)	||
			(info.h.anc == 0)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}

	/* Question Entries */
	memcpy(data->QName, info.qd[0].qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->QScopeId, info.qd[0].qscopeid, NBT_LENGTH_SCOPEID);

	get_NameServiceMessage_clean(&info);

	return 0;
}


int get_NodeStatusRequest(stNodeStatusRequest* data, const u_char* in)
{
	/* almost same as get_NameRegistrationRequest */
	return get_NameQueryRequest(data, in);
}



// -- get_*Response Section
int get_WackResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	stNameServiceMessageInfo info;
	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* WackResponsȅꍇ́Aanc1ƂȂ */
	if(!(	(info.h.qdc == 0)	||
			(info.h.anc == 1)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}

	/* Anwser Record̃f[^dataɑ */
	memcpy(data->RRName, info.an[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.an[0].s.qname, NBT_LENGTH_SCOPEID);
	data->TTL = info.an[0].ttl;

	get_NameServiceMessage_clean(&info);

	return 0;
}



int get_PositiveNameRegistrationResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	stNameServiceMessageInfo info;
	stNameServiceAddrEntry ent;
	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* PositiveNameRegistration Responsȅꍇ́Aanc1ƂȂ */
	if(!(	(info.h.qdc == 0)	||
			(info.h.anc == 1)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}
	/* Answer Record̃f[^dataɑ */
	memcpy(data->RRName, info.an[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.an[0].s.qname, NBT_LENGTH_SCOPEID);
	data->TTL = info.an[0].ttl;

	/* Answer Record̃f[^͂ */
	get_NameServiceAddrEntryArray(&ent, 1, info.an[0].rdata);
	get_NameServiceMessage_clean(&info);

	/* ͂ʂdataɑ */
	data->G = ent.G;
	data->NbAddress = ent.nbaddr;
	data->ONT = ent.ONT;

	return 0;
}


int get_NegativeNameRegistrationResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	/* almost same as get_PositiveNameRegistrationResponse */
	return get_PositiveNameRegistrationResponse(data, in);
}


int get_EndNodeChallengeRegistrationResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	/* almost same as get_PositiveNameRegistrationResponse */
	return get_PositiveNameRegistrationResponse(data, in);
}


int get_PositiveNameReleaseResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	/* almost same as get_PositiveNameRegistrationResponse */
	return get_PositiveNameRegistrationResponse(data, in);
}

int get_NegativeNameReleaseResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	/* almost same as get_PositiveNameRegistrationResponse */
	return get_PositiveNameRegistrationResponse(data, in);
}


int get_PositiveNameQueryResponse(stNameServicePrimitiveRequest* data, stNameServiceAddrEntry* ent, int num,
								  int* findnum, const u_char* in)
{
	stNameServiceMessageInfo info;

	int res;

	*findnum = 0;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* PositiveNameRegistration Responsȅꍇ́Aanc1ƂȂ */
	if(!(	(info.h.qdc == 0)	||
			(info.h.anc == 1)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}
	/* Answer Record̃f[^dataɑ */
	memcpy(data->RRName, info.an[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.an[0].s.qname, NBT_LENGTH_SCOPEID);
	data->TTL = info.an[0].ttl;
	
	*findnum = info.an[0].rdatalen/sizeof(stNameServiceRRData);

	/* Answer Record̃f[^͂ */
	get_NameServiceAddrEntryArray(ent, MIN(*findnum, num), info.an[0].rdata);
	get_NameServiceMessage_clean(&info);

	return 0;
}


int get_NegativeNameQueryResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	stNameServiceMessageInfo info;
	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* anc1ƂȂ */
	if(!(	(info.h.qdc == 0)	||
			(info.h.anc == 1)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}
	/* Answer Record̃f[^dataɑ */
	memcpy(data->RRName, info.an[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.an[0].s.qname, NBT_LENGTH_SCOPEID);
	data->TTL = info.an[0].ttl;

	return 0;
}


int get_RedirectNameQueryResponse(stNameServicePrimitiveRequest* data, const u_char* in)
{
	// not support!!

	return -1;
}


int get_NodeStatusResponse(stNameServicePrimitiveRequest* data, stNameServiceNodeNameEntry* dnn, int num, int* findnum,
						   stNameServiceStatistics* dns, const u_char* in)
{
	stNameServiceMessageInfo info;
	int res;

	res = get_xReadMessage(data, &info, in);
	if(res < 0){
		return -1;
	}

	/* anc1ƂȂ */
	if(!(	(info.h.qdc == 0)	||
			(info.h.anc == 1)	||
			(info.h.nsc == 0)	||
			(info.h.arc == 0)	)){

		get_NameServiceMessage_clean(&info);
		return -1;
	}
	/* Answer Record̃f[^dataɑ */
	memcpy(data->RRName, info.an[0].s.qname, NBT_LENGTH_NETBIOSNAME);
	memcpy(data->RScopeId, info.an[0].s.qname, NBT_LENGTH_SCOPEID);
	data->TTL = info.an[0].ttl;

	/* Answer Record̃f[^͂ */
	/* NodeNameArrayǂ */
	res = get_NameServiceNodeNameEntryArray(dnn, num, findnum, info.an[0].rdata);
	if(res < 0){
		get_NameServiceMessage_clean(&info);
		return -1;
	}
	
	/* NodeStatisticsǂ */
	res = get_NameServiceStatistics(dns, info.an[0].rdata + res);
	get_NameServiceMessage_clean(&info);
	if(res < 0){
		return -1;
	}

	return 0;
}




/*
------------------------------------------------------------------
tidobt@TranIDtB[hɏ݂܂B
ލۂɃoCgI[_[̕ύXsȂ܂B
------------------------------------------------------------------
*/
int set_tranID(u_char* out, u_short tid)
{
	stNameServiceHeader* h;
	h = (stNameServiceHeader* )out;
	
	h->id = htons(tid);
	
	return 0;
}


/*
------------------------------------------------------------------
send_RequestMessageRepeatAnyTimes()

l[T[rX̏xɂĂ܂֐

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_NORESPONSE

		LIBNBT_ERR_SENDTO_ERR		(-160)	:	sendto֐G[(errno)
		LIBNBT_ERR_SELECT_ERR		(-150)	:	select֐G[(errno)
		LIBNBT_ERR_SELECT_TIMEOUT	(-151)	:	^CAEg(Ms)
		LIBNBT_ERR_INTERRUPTED		(-2000)	:	f

		LIBNBT_ERR_RECVFROM_ERR		(-180)	:	recvfrom֐G[(errno)

		LIBNBT_ERR_RECVDATA_INVALID			:	Mf[^s


------------------------------------------------------------------
*/
int send_RequestMessageRepeatAnyTimes(stRequestMessageInfo* info)
{
	u_char buf[MESSAGE_BUFSIZE];
	u_short* tids = NULL;
	int len;
	int res;
	int i;
	int mestype_res;

	memset(buf, 0, sizeof buf);

	// create message
	switch(info->mestype){
	case NBT_NS_MT_REQ_NAMEREG:
		len = create_NameRegistrationRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_NAMEOW:
		len = create_NameOverwriteRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_REF:
		len = create_NameRefreshRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_REL:
		len = create_NameReleaseRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_NAMEQUERY:
		len = create_NameQueryRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_NODESTATUS:
		len = create_NodeStatusRequest(buf, &info->prim);
		break;
	case NBT_NS_MT_REQ_MHNAMEREG:
		len = create_MultihomedNameRegistrationRequest(buf, &info->prim);
		break;

	default:
		D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: no such mes_type %d\n", info->mestype));
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(len < 0){
		D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: create_XXXXRequest mestype = %d: failed\n",
			info->mestype));
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	// malloc tid cache area
	if(info->waitresponse){
		tids = (u_short* )malloc(sizeof(u_short)*info->retry);
		if(tids == NULL){
			return LIBNBT_ERR_MALLOC_FAILED;
		}
		memset(tids, 0, sizeof(u_short)*info->retry);
	}
	
	// send "retry" times if no response
	for(i=0; i<info->retry; i++){
		if(info->waitresponse){ tids[i] = info->prim.TranID; }
		set_tranID(buf, info->prim.TranID);
		info->prim.TranID++;
		D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: dstaddr=%s, port=%d\n",
		net_ntoa(*(struct in_addr* )&dstnaddr), htons(info->dstnport)));
		
		res = nbt_in_sendto(info->fd, buf, len, 0, info->dstnaddr, info->dstnport, NBT_WAIT_SEC, info->interruptd);
		if(res != len){
			D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: nbt_in_sendto send failed size demand=%d actualy=%d\n", len, res));
			if(info->waitresponse){ free(tids); }

			return res;
		}
		
		if(info->waitresponse){
LABEL_CHECK_ARRIVE:
			res = nbt_readable(info->fd, info->timeout, info->interruptd);
			if(res < 0){
				// error
				D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: nbt_readable failed\n"));
				free(tids);
				if(res == -1){
					return LIBNBT_ERR_SELECT_ERR;
				}
				// res == -2
				return LIBNBT_ERR_INTERRUPTED;

			}
			if(res == 1){	/*  */
				// response
				// check tid
				stNameServiceHeader* header;
				u_short tid_resp;
				u_long rnaddr;
				u_short rnport;
				int	ignore;
				
				// read message to buffer
				res = nbt_in_recvfrom(info->fd, buf, sizeof buf, 0, &rnaddr, &rnport, NBT_WAIT_SEC, info->interruptd);
				if(res < 0){
					D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: nbt_in_recvfrom failed.\n"));
					free(tids);
					return res;
				}

				header = (stNameServiceHeader* )buf;

				/* MƉꂩǂ肷(Broadcast̎͏) */
				if((!(info->prim.nmflags & NS_NSPR_NMFLAGS_B_ON)) && (rnaddr != info->dstnaddr)){
					D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: Reponse from other Machine.\n"));
					goto LABEL_CHECK_ARRIVE;
				}
				tid_resp = ntohs(header->id);
				if(hassameval_short(tid_resp, tids, i+1) == -1){
					// tid not found
					// this packet must be ignore.
					int k;
					D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: INFO: tid not found. tid_resp = %d \n", tid_resp));
					for(k=0; k<i; k++){
						D_PRINTF(("cache tid = %d\n", tids[k]));
					}
					goto LABEL_CHECK_ARRIVE;
				}
				
				// tid found
				// check Operation Code.
				/* bZ[W^Cv𔻒肷 */
				mestype_res = get_NameServiceMessageType(buf);
				if(mestype_res == NBT_NS_MT_UNKNOWN){
					D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: response packet messagetype unknown\n"));
					free(tids);
					return LIBNBT_ERR_RECVDATA_INVALID;
				}

				switch(info->mestype){
				case NBT_NS_MT_REQ_NAMEREG:
				case NBT_NS_MT_REQ_MHNAMEREG:
				case NBT_NS_MT_REQ_NAMEOW:
				case NBT_NS_MT_REQ_REF:
					switch(mestype_res){
					case NBT_NS_MT_RES_WACK:
						res = get_WackResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_POS_NAMEREG:
						res = get_PositiveNameRegistrationResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_NEG_NAMEREG:
						res = get_NegativeNameRegistrationResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_ENCREG:
						res = get_EndNodeChallengeRegistrationResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					default:
						/* invalid message type */
						D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: invalid response packet messagetype %d\n",
							mestype_res));
						res = -1;
						break;
					}

					break;
					
				case NBT_NS_MT_REQ_REL:
					switch(mestype_res){
					case NBT_NS_MT_RES_WACK:
						res = get_WackResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);	break;
					case NBT_NS_MT_RES_POS_NAMEREL:
						res = get_PositiveNameReleaseResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);	break;
					case NBT_NS_MT_RES_NEG_NAMEREL:
						res = get_NegativeNameReleaseResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);	break;
					default:
						/* invalid message type */
						D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: invalid response packet messagetype %d\n",
							mestype_res));
						res = -1;
						break;
					}

					break;

				case NBT_NS_MT_REQ_NAMEQUERY:
					switch(mestype_res){
					case NBT_NS_MT_RES_WACK:
						res = get_WackResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_POS_NAMEQUERY:
						res =get_PositiveNameQueryResponse((stNameServicePrimitiveRequest*)info->rinfo[0],
							(stNameServiceAddrEntry* )info->rinfo[1],
							*(int*)info->rinfo[2], (int*)info->rinfo[3],buf);

						break;
					case NBT_NS_MT_RES_NEG_NAMEQUERY:
						res = get_NegativeNameQueryResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_REDNAMEQUERY:
						/* always return error */
						res = get_RedirectNameQueryResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					default:
						/* invalid message type */
						D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: invalid response packet messagetype %d\n",
							mestype_res));
						res = -1;
						break;
					}

					break;
					
				case NBT_NS_MT_REQ_NODESTATUS:
					switch(mestype_res){
					case NBT_NS_MT_RES_WACK:
						res = get_WackResponse((stNameServicePrimitiveRequest*)info->rinfo[0], buf);
						break;
					case NBT_NS_MT_RES_NODESTATUS:
						res = get_NodeStatusResponse((stNameServicePrimitiveRequest*)info->rinfo[0],
							(stNameServiceNodeNameEntry*)info->rinfo[1],
							*(int*)info->rinfo[2], (int*)info->rinfo[3], (stNameServiceStatistics*)info->rinfo[4], buf);
						break;
					default:
						/* invalid message type */
						D_PRINTF(("[libnbt] send_RequestMessageRepeatAnyTimes: invalid response packet messagetype %d\n",
							mestype_res));
						res = -1;
						break;
					}

					break;

				default:
					D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: no such mes_type\n"));
					return LIBNBT_ERR_RECVDATA_INVALID;
				}

				if(res < 0){
					D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: get_XXXXXXResponse failed\n"));
					free(tids);
					return LIBNBT_ERR_RECVDATA_INVALID;
				}
				
				/* MbZ[W^CvWACKƂ̏ */
				if(mestype_res == NBT_NS_MT_RES_WACK){
					stNameServicePrimitiveRequest* p;
					p = (stNameServicePrimitiveRequest*)info->rinfo[0];
					D_PRINTF(("[libnbt] send_ReuestMessageRepeatAnyTimes: INFO: sleep %lu sec\n", p->TTL));
					/* w肳ꂽ~bԑ҂Ă݂ */
					res = nbt_readable(info->fd, p->TTL, info->interruptd);
					if(res < 0){
						D_PRINTF(("[libnbt] WACK Response Response error\n"));
						free(tids);
						if(res == -2){
							return LIBNBT_ERR_INTERRUPTED;
						}
						return LIBNBT_ERR_SELECT_ERR;
					}
					/* 肩烌X|XȂꍇłʉ߂ */

					goto LABEL_CHECK_ARRIVE;
				}
				
				info->retval = mestype_res;
				ignore = 0;
				/* pPbgI */
				if(info->ignoreresponse & NS_RESTYPE_POSITIVE){
					if(	(mestype_res == NBT_NS_MT_RES_POS_NAMEREG) 	||
						(mestype_res == NBT_NS_MT_RES_POS_NAMEREL)	||
						(mestype_res == NBT_NS_MT_RES_POS_NAMEQUERY)){
						ignore = 1;
					}
				}
				if(info->ignoreresponse & NS_RESTYPE_NEGATIVE){
					if(	(mestype_res == NBT_NS_MT_RES_NEG_NAMEREG)	||
						(mestype_res == NBT_NS_MT_RES_NEG_NAMEREL)	||
						(mestype_res == NBT_NS_MT_RES_NEG_NAMEQUERY)){
						ignore = 1;
					}
				}
				if(info->ignoreresponse & NS_RESTYPE_ENODECHA){
					if(	(mestype_res == NBT_NS_MT_RES_ENCREG)){
						ignore = 1;
					}
				}
				if(info->ignoreresponse & NS_RESTYPE_REDIRECT){
					if(	(mestype_res == NBT_NS_MT_RES_REDNAMEQUERY)){
						ignore = 1;
					}
				}
				if(ignore){
					// ̃pPbg҂
					info->retval = 0;
					goto LABEL_CHECK_ARRIVE;
				}
				
				break;
			} // response
			
			// no response
			
		} // waitresponse
		
	} // for
	
	if(!info->waitresponse){
		return LIBNBT_ERR_NOERR;
	}
	
	free(tids);
	if(i == info->retry){
		// no response
		return LIBNBT_ERR_NORESPONSE;
	}
	
	// response
	return LIBNBT_ERR_NOERR;
}	


// UserDatagram implement
// f[^̃TCY512oCg𒴂Ȃ悤ɂ悤
// out: MbZ[W
// data: etB[h
// return value: NBTf[^TCY
int create_DatagramFormat(u_char* out, const stDatagramPrimitive* data)
{
	stDatagramServiceFormat* format;
	int res;
	int totalsize;

	format = (stDatagramServiceFormat* )out;

	// header
	format->msgtype		= data->msgtype;
	S_DSF_FLAGS_SNT(format->flags, data->snt);
	S_DSF_FLAGS_F(format->flags, data->first_flag);
	S_DSF_FLAGS_M(format->flags, data->more_flag);
	format->id			= htons(data->id);
	format->ip			= data->naddr;
	format->port		= data->nport;

	switch(data->msgtype){
	case VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM:
	case VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM:
	case VAL_DSF_MTYPE_BROADCAST_DATAGRAM:

		format->DSF1_len	= htons(data->DP1_datalen);
		format->DSF1_offset = htons(data->DP1_offset);

		if(data->first_flag){
			res = nbt_enc2nd(format->DSF1_srcname, 256, data->DP1_srcname, data->DP1_scopeid);
			if(res < 0){
				return -1;
			}
			(u_char* )format += res;
			res = nbt_enc2nd(format->DSF1_srcname, 256, data->DP1_dstname, data->DP1_scopeid);
			if(res < 0){
				return -1;
			}
			(u_char* )format += res;
		}
		else{
			// format -= (sizeof format->DSF1_srcname + sizeof format->DSF1_dstname);
		}
		memcpy(format->DSF1_body, data->DP1_data, data->DP1_datalen);

		format = (stDatagramServiceFormat* )out;
		totalsize = (int)((u_long)format->DSF1_body - (u_long)out) + (int)data->DP1_datalen;

		break;

	case VAL_DSF_MTYPE_DATAGRAM_ERROR:
		
		format->DSF2_ecode = data->DP2_ecode;
		totalsize = (int)((u_long)format->DSF2_ecode - (u_long)out) + sizeof(format->DSF2_ecode);

		break;

	case VAL_DSF_MTYPE_DATAGRAM_QUERY_REQUEST:
		res = nbt_enc2nd(format->DSF3_dstname, 256, data->DP3_dstname, data->DP3_scopeid);
		if(res < 0){
			return -1;
		}
		totalsize = (int)((u_long)format->DSF3_dstname - (u_long)out) + res;

		break;

	default:
		D_PRINTF(("[libnbt] int create_DatagramFormat: no support msgtype %d\n", data->msgtype));
		return -1;

	}

	return totalsize;
}

// success: 0
// error: -1
int get_DatagramFormat(stDatagramPrimitive* data, const u_char* in)
{
	// inɓĂpPbgf[^dataɓ
	stDatagramServiceFormat* format;
	int res;
	int totalsize;
	int readsize;

	memset(data, 0, sizeof(stDatagramPrimitive));

	format = (stDatagramServiceFormat* )in;
	
	// common
	data->msgtype = format->msgtype;
	data->snt = G_DSF_FLAGS_SNT(format->flags);
	data->first_flag = G_DSF_FLAGS_F(format->flags);
	data->more_flag = G_DSF_FLAGS_M(format->flags);
	data->id = ntohs(format->id);
	data->naddr = format->ip;	// n-byte order
	data->nport = format->port;	// n-byte order

	switch(format->msgtype){
	case VAL_DSF_MTYPE_DIRECT_UNIQUE_DATAGRAM:
	case VAL_DSF_MTYPE_DIRECT_GROUP_DATAGRAM:
	case VAL_DSF_MTYPE_BROADCAST_DATAGRAM:
		// format1
		data->DP1_datalen = ntohs(format->DSF1_len);
		data->DP1_offset = ntohs(format->DSF1_offset);
		
		if(data->first_flag){
			int s = 0;
			res = nbt_dec2nd(data->DP1_srcname, data->DP1_scopeid, sizeof data->DP1_scopeid, format->DSF1_srcname, in, &readsize, NULL, NULL);
			if(res < 0){
				return -1;
			}
			(u_char* )s += readsize;
			(u_char* )format += readsize;
			data->DP1_headerlen += readsize;

			res = nbt_dec2nd(data->DP1_dstname, NULL, 0, format->DSF1_dstname, in, &readsize, NULL, NULL);
			if(res < 0){
				return -1;
			}
			(u_char* )s += readsize;
			(u_char* )format += readsize;
			data->DP1_headerlen += readsize;

			totalsize = (data->DP1_datalen - s);
		}
		
		data->DP1_data = format->DSF1_body;
		totalsize = data->DP1_datalen;

		break;
	
	case VAL_DSF_MTYPE_DATAGRAM_ERROR:
		// format2
		data->DP2_ecode = format->DP2_ecode;
		totalsize = 0;
		break;
	
	case VAL_DSF_MTYPE_DATAGRAM_QUERY_REQUEST:
	case VAL_DSF_MTYPE_DATAGRAM_POSITIVE_QUERY_RESPONSE:
	case VAL_DSF_MTYPE_DATAGRAM_NEGATIVE_QUERY_RESPONSE:
		// format3
		res = nbt_dec2nd(data->DP3_dstname, data->DP3_scopeid, sizeof data->DP3_scopeid, format->DSF3_dstname, in, &readsize, NULL, NULL);
		if(res < 0){
			return -1;
		}
		totalsize = 0;
		break;

	default:
		D_PRINTF(("[libnbt] get_DatagramFormat: no such megtype %d\n", format->msgtype));
		return -1;
	}

	// success
	return totalsize;
}
