/*
   Unix SMB/Netbios implementation.
   Version 1.9.
   Character set conversion Extensions
   Copyright (C) Andrew Tridgell 1992-1998

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include "includes.h"
#define CTRLZ 26

static BOOL mapsinited = 0;

static char unix2dos[256];
static char dos2unix[256];

static void initmaps(void) {
    int k;

    for (k = 0; k < 256; k++) unix2dos[k] = k;
    for (k = 0; k < 256; k++) dos2unix[k] = k;

    mapsinited = True;
}

/*
 * Convert unix to dos
 */
char *unix2dos_format(char *to, const char *from)
{
    pstring cvtbuf;
    char *out;

    if (to == from) {
	from = pstrcpy ((char *) cvtbuf, from);
    }
    
    if (!mapsinited)
	initmaps();
    
    for (out = to; *from && (out - to < sizeof(pstring) - 1);)
	*out++ = unix2dos[(unsigned char)*from++];
    *out = '\0';
    return to;
}

/*
 * Convert dos to unix
 */
char *dos2unix_format(char *to, const char *from)
{
    pstring cvtbuf;
    char *out;

    if (to == from) {
	from = pstrcpy ((char *) cvtbuf, from);
    }
    if (!mapsinited)
	initmaps();
    
    for (out = to; *from && (out - to < sizeof(pstring) - 1);)
	*out++ = dos2unix[(unsigned char)*from++];
    *out = '\0';
    return to;
}

