/*
   Unix SMB/Netbios implementation.
   Version 3.0
   client RAP calls
   Copyright (C) Andrew Tridgell 1994-1998

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define NO_SYSLOG

#include "includes.h"
#include <smbc_lib.h>

/****************************************************************************
call a remote api
****************************************************************************/
BOOL cli_api(struct cli_state *cli,
	     char *param, int prcnt, int mprcnt,
	     char *data, int drcnt, int mdrcnt,
	     char **rparam, int *rprcnt,
	     char **rdata, int *rdrcnt, u_short *char_code)
{
  cli_send_trans(cli,SMBtrans,
                 PIPE_LANMAN,             /* Name */
                 0,0,                     /* fid, flags */
                 NULL,0,0,                /* Setup, length, max */
                 param, prcnt, mprcnt,    /* Params, length, max */
                 data, drcnt, mdrcnt,      /* Data, length, max */ 
		 char_code
                );

  return (cli_receive_trans(cli,SMBtrans,
                            rparam, rprcnt,
                            rdata, rdrcnt));
}

/****************************************************************************
call a NetShareEnum - try and browse available connections on a host
****************************************************************************/
int cli_RNetShareEnum(struct cli_state *cli, int *totalnum, int *listnum, void *buf, void *state, u_short *char_code)
{
	char *rparam = NULL;
	char *rdata = NULL;
	char *p;
	int rdrcnt,rprcnt;
	pstring param;
	int count = -1;
	smb_sh_info *shinfo = (smb_sh_info *)buf;

	/* now send a SMBtrans command with api RNetShareEnum */
	p = param;
	SSVAL(p,0,0); /* api number */
	p += 2;
	pstrcpy(p,"WrLeh");
	p = skip_string(p,1);
	pstrcpy(p,"B13BWz");
	p = skip_string(p,1);
	SSVAL(p,0,1);
	/*
	 * Win2k needs a *smaller* buffer than 0xFFFF here -
	 * it returns "out of server memory" with 0xFFFF !!! JRA.
	 */
	SSVAL(p,2,0xFFE0);
	p += 4;
	
	if (cli_api(cli, 
		    param, PTR_DIFF(p,param), 1024,  /* Param, length, maxlen */
		    NULL, 0, 0xFFE0,            /* data, length, maxlen - Win2k needs a small buffer here too ! */
		    &rparam, &rprcnt,                /* return params, length */
		    &rdata, &rdrcnt, char_code))     /* return data, length */
		{
			int res = rparam? SVAL(rparam,0) : -1;
			smb_sh_info *shp = (smb_sh_info *)shinfo;
			
			if (res == 0 || res == ERRmoredata) {
				int converter=SVAL(rparam,2);
				int i;
				int found_num;

				count=SVAL(rparam,4);
				p = rdata;
				for (i=0, found_num=0;i<count && found_num<(*listnum);i++,p+=20) {
					char *sname = p;
					u_short stype = SVAL(p, 14);
					int comment_offset = IVAL(p,16) & 0xFFFF;
					char *cmnt = comment_offset?(rdata+comment_offset-converter):"";

					if(stype != 0) {	/* fBXNȊO̓XLbv */
						continue;
					}
					shp->character = *char_code;
					strncpy(shp->sh_name, sname, SMB_MAXSHARENAMELEN*2);
					strncpy(shp->sh_comment, cmnt, SMB_MAXCOMMENTLEN*2);
					shp++;
					found_num++;
				}
				(*listnum) = found_num;
				(*totalnum) = count - (i - found_num);
			} else {
				DEBUG_PRINT(4,("NetShareEnum res=%d\n", res));
			}      
		} else {
			DEBUG_PRINT(4,("NetShareEnum failed\n"));
		}
  
	SAFE_FREE(rparam);
	SAFE_FREE(rdata);
	
	return count;
}

/****************************************************************************
call a NetServerEnum for the specified workgroup and servertype mask.  This
function then calls the specified callback function for each name returned.

The callback function takes 4 arguments: the machine name, the server type,
the comment and a state pointer.
****************************************************************************/
int cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32 stype,
		       int *totalnum, int *listnum, void *buf,
		       void *state,  u_short *char_code)
{
	char *rparam = NULL;
	char *rdata = NULL;
	int rdrcnt,rprcnt;
	char *p;
	pstring param;
	int uLevel = 1;
	int count = -1;
	smb_gp_info *gpinfo = 0;
	smb_sv_info *svinfo = 0;

	/* send a SMBtrans command with api NetServerEnum */
	p = param;
	SSVAL(p,0,0x68); /* api number */
	p += 2;
	pstrcpy(p,"WrLehDz");
	p = skip_string(p,1);
  
	pstrcpy(p,"B16BBDz");

	p = skip_string(p,1);
	SSVAL(p,0,uLevel);
	SSVAL(p,2,CLI_BUFFER_SIZE);
	p += 4;
	SIVAL(p,0,stype);
	p += 4;

	if(stype == SV_TYPE_DOMAIN_ENUM) {
		gpinfo = (smb_gp_info *)buf;
	} else {
		svinfo = (smb_sv_info *)buf;
	}

	p += clistr_push(cli, p, workgroup, -1,
		STR_TERMINATE | STR_CONVERT | STR_ASCII, char_code);

	if (cli_api(cli, 
                    param, PTR_DIFF(p,param), 8,    /* params, length, max */
                    NULL, 0, CLI_BUFFER_SIZE,       /* data, length, max */
                    &rparam, &rprcnt,               /* return params, return size */
                    &rdata, &rdrcnt, char_code      /* return data, return size */
                   )) {
		int res = rparam? SVAL(rparam,0) : -1;
			
		if (res == 0 || res == ERRmoredata) {
			int i;
			int converter=SVAL(rparam,2);
			smb_gp_info *gp = (smb_gp_info *)gpinfo;
			smb_sv_info *sp = (smb_sv_info *)svinfo;

			count=SVAL(rparam,4);
			*totalnum = count;
			if(count > *listnum) {
				count = *listnum;
			}
			*listnum = count;
			p = rdata;
			for (i = 0;i < count;i++, p += 26) {
				char *sname = p;

				int comment_offset = (IVAL(p,22) & 0xFFFF)-converter;
				char *cmnt = comment_offset?(rdata+comment_offset):"";

				if (comment_offset < 0 || comment_offset > rdrcnt) continue;

				stype = IVAL(p,18) & ~SV_TYPE_LOCAL_LIST_ONLY;
				
				if(gp) {
					gp->character = *char_code;
					strncpy(gp->gp_name, sname, SMB_MAXSRVNAMELEN*2);
					gp->gp_type = stype;
					strncpy(gp->gp_master_browser, cmnt, SMB_MAXSRVNAMELEN*2);
					gp++;
				} else if(sp) {
					sp->character = *char_code;
					strncpy(sp->sv_name, sname, SMB_MAXSRVNAMELEN*2);
					sp->sv_type = stype;
					strncpy(sp->sv_comment, cmnt, SMB_MAXCOMMENTLEN*2);
					sp++;
				}
			}
		}
	}
  
	SAFE_FREE(rparam);
	SAFE_FREE(rdata);

	return count;
}

#ifdef NCS_SMB_NTAUTH_SUPPORT
/****************************************************************************/
BOOL cli_NetUserGetGroups(struct cli_state *cli, int *totalnum, int *listnum, void *buf, char *username, u_short *char_code)
{
	char *rparam = NULL;
	char *rdata = NULL;
	int rdrcnt,rprcnt;
	char *p;
	pstring param;
	int uLevel = 0;
	int count = -1;
	smb_ugp_info *ugpinfo = (smb_ugp_info *)buf;

	/* send a SMBtrans command with api NetUserGetGroups */
	/* Initialize */
	bzero( param , sizeof(param));
	p = param;
	SSVAL(p,0,0x3b); /* api number = 59 */
	p += 2;
	pstrcpy(p,"zWrLeh");
	p = skip_string(p,1);
	pstrcpy(p,"B21");
	p = skip_string(p,1);

	p += clistr_push(cli, p, username, -1,
		STR_TERMINATE | STR_CONVERT | STR_ASCII, char_code);
	p = skip_string(p,1);

	SSVAL(p,0,uLevel);

	SSVAL(p,2,CLI_BUFFER_SIZE);
	p += 4;

	if (cli_api(cli, 
                    param, PTR_DIFF(p,param), 8,    /* params, length, max */
                    NULL, 0, CLI_BUFFER_SIZE,       /* data, length, max */
                    &rparam, &rprcnt,               /* return params, return size */
                    &rdata, &rdrcnt, char_code      /* return data, return size */
                   )) {
		int res = rparam? SVAL(rparam,0) : -1;

		if (res == 0 || res == ERRmoredata) {
			int converter=SVAL(rparam,2);
			int i;
			count=SVAL(rparam,4);
			*totalnum = count;
			if(count > *listnum) {
				count = *listnum;
			}
			*listnum = count;
			p = rdata;
			for (i = 0;i < count;i++, p += 21) {
				strncpy(ugpinfo[i].ugp_name, p, 20);
			}
		} else {
			DEBUG_PRINT(4,("NetUserGetGroups res=%d\n", res));
			*totalnum = 0;
			*listnum = 0;
		}
	} else {
		DEBUG_PRINT(4,("NetUserGetGroups failed\n"));
	}
  
	SAFE_FREE(rparam);
	SAFE_FREE(rdata);

	return(count > 0);
}

#endif /* NCS_SMB_NTAUTH_SUPPORT */
