#ifdef notdef
#include <stdio.h>

#include <string.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#include <libnbt/libnbt.h>
#include <libnbt/libnbt_ns.h>
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_ns_in.h"
#include <libnbt/libnbt_mes.h>
#include <libnbt/libnbt_util.h>
#include "libnbt_util_private.h"
#include <libnbt/libnbt_err.h>
#include "libnbt_debug.h"

int in_nodestatus(NBTCTBL* ct, const unsigned long nodeaddr, NBTNamePack *np, int num, int* findnum);
int nbt_nodestatus(int s, const unsigned long nodeaddr, NBTNamePack *np, int num, int* findnum);
/*****************************************************/


int in_nodestatus(NBTCTBL* ct, const unsigned long nodeaddr, NBTNamePack *np, int num, int* findnum)
{
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	stNameServiceNodeNameEntry nodename;
	stNameServiceStatistics statistics;
	int res;
	int tnum = 1;

	if((ct == NULL) || (nodeaddr == 0)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->messageType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	// create
	memset(&info, 0, sizeof info);
	memset(&nodename, 0, sizeof nodename);
	memset(&statistics, 0, sizeof statistics);
	info.fd 		= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_NODESTATUS;
	info.dstnaddr 		= nodeaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 		= UCAST_REQ_RETRY_COUNT;
	info.timeout		= UCAST_REQ_RETRY_TIMEOUT;
	info.waitresponse	= 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 	= (void*)&rinfo;
	info.rinfo[1]	= (void*)&nodename;
	info.rinfo[2]	= (void*)&tnum;
	info.rinfo[3]	= (void*)&statistics;

	info.interruptd		= -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;
	nbt_padname("*", info.prim.QName, NBT_LENGTH_NETBIOSNAME, 0, 0x00);
	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.ONT			= GSetting.ont;
	info.prim.G			= 0; // DONT CARE
	info.prim.nmflags		= 0;

	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(res != LIBNBT_ERR_NOERR){
		if(res == LIBNBT_ERR_INTERRUPTED){
			return res;
		}
		D_PRINTF(("[libnbt] in_nodestatus: send_RequestMessageRepeatAnyTimes failed\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}
		
	switch(info.retval){
	case NBT_NS_MT_RES_NODESTATUS:
		if(findnum){
			*findnum = 1;
		}
		strcpy(np->name, nodename.nbname);
		printf("np->name = %s\n", np->name);
		return LIBNBT_ERR_NOERR; // success

	default:
		D_PRINTF(("[libnbt] in_nodestatus: send_RequestMessageRepeatAnyTimes no such retval\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}

	// never come here.
	return LIBNBT_ERR_OTHER_ERR;
}

int nbt_nodestatus(int s, const unsigned long nodeaddr, NBTNamePack *np, int num, int* findnum)
{
	NBTCTBL* ct;
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(num < 1) {
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return in_nodestatus(ct, nodeaddr, np, num, findnum);
}

#endif
