#ifdef NCS_SMB_NTAUTH_SUPPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>


#include <netinet/in.h>
#include "includes.h"
#include <smbc_lib.h>
#include <syslog.h>

extern int smb_getnodestatus(const NBTNamePack* rmt, NBTNodeName* nn, NBTNodeStatistics* ns, u_short suffix);
extern int nbt_connectTimeout(int *s, NBTNamePack* rmt, const NBTNamePack* local, u_long *ip, int num);
extern int cli_SmbOpenPipe( smb_ctx* ctx, smb_vc* vcp, SMBC_PROC func);
extern int cli_SmbClosePipe( smb_ctx* ctx, SMBC_PROC func);
/**********************************************************
 * NcsSmbUserLogin
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbUserLogin(smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback)
{
	int				ret = 0;
	int				status = 0;
	pthread_t 		thread_id;
	smbc_ulogin_t	*th_arg = NULL;

	if(!ctx || !vcp) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(ctx->cli) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}

	if(callback){
		th_arg = (smbc_ulogin_t *)smb_malloc(sizeof(smbc_ulogin_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogin: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbUserLoginThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].\n",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogin: pthread_create() error[%d].\n",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogin: pthread_mutex_trylock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
		
		ret = NcsSmbUserLoginSub(ctx, vcp);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogin: pthread_mutex_unlock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
	}
	return(ret);
}

void *NcsSmbUserLoginThread(void *arg)
{
	int		ret = 0;
	int		status = 0;
	smbc_ulogin_t  *th_arg = (smbc_ulogin_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginThread: pthread_detach() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginThread: pthread_mutex_trylock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbUserLoginSub(th_arg->ctx, th_arg->vcp);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginThread: pthread_mutex_unlock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);

#ifdef DEBUG
	printf("[NcsSmbUserLoginThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbUserLoginSub( smb_ctx* ctx, smb_vc* vcp)
#else
int NcsSmbUserLogin( smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback )
#endif
{
	int	res = 0;

	struct in_addr dest_ip;
	struct nmb_name called;
	NBTAddrPack	rmt;
        char            group[MAX(SMB_MAXSRVNAMELEN,SMB_MAXHOSTNAMELEN)+1];
	char		server[SMB_MAXHOSTNAMELEN+1];
	char		username[SMB_MAXUSERNAMELEN+1];
	char		password[SMB_MAXPASSWORDLEN+1];
	u_short		char_code = vcp->character;
	int		name_type = 0;
	int		NcsSmbUserLogin_negprot_fail = 0;
	u_long		target_ip[SMB_MAX_IP];
	int		num = SMB_MAX_IP;

	if(!ctx || !vcp) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if(ctx->cli) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	ctx->cli = cli_initialise(NULL);
	if(!ctx->cli) {
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	bzero(&dest_ip, sizeof(dest_ip));
	bzero(&called, sizeof(called));
	bzero(&rmt, sizeof(rmt));
	bzero(group, sizeof(group));
	bzero(server, sizeof(server));
	bzero(username, sizeof(username));
	bzero(password, sizeof(password));

	if(vcp->group[0] != 0) {
		strncpy(group, vcp->group, sizeof(group)-1);
	} else {
		smb_get_mygroup(group);
	}
	if(vcp->server[0] != 0) {
		strncpy(server, (char *)vcp->server, SMB_MAXSRVNAMELEN);
		ctx->rmt.suffix = SUFFIX_SERV;
		name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);
		if(!name_type) {
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			syslog( LOG_ERR, "SMBC lib. cannot resolve server name(%s).",  server);
#ifndef SMBCD_USE
	                smbc_mfprintf("NcsSmbUserLoginSub failed: smb_gethostbyname. \n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else if(name_type != 1) {
			int res = 0;
			NBTNodeName nn;
			NBTNodeStatistics ns;
			int findnum = 0;
			int ii = 0;

			for(ii = 0; ii < num; ii++) {
				ctx->rmt.naddr = target_ip[ii];
				memset(&nn, 0, sizeof(nn));
				memset(&ns, 0, sizeof(ns));
				res = smb_getnodestatus(&ctx->rmt, &nn, &ns, SUFFIX_SERV);
				if(res == LIBNBT_ERR_NOERR) {
					printf("smb_getnodestatus: name = %s\n", nn.name);
					strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
					bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
					strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
					if(ii > 0) {
						u_long tmp_ip;
						tmp_ip = target_ip[0];
						target_ip[0] = target_ip[ii];
						target_ip[ii] = tmp_ip;
					}
					break;
				}
			}
		}
		strncpy(ctx->rmt.name, server, SMB_MAXSRVNAMELEN);
	} else {
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;
		int ii = 0;

		/* get domain controller */ 
		name_type = smb_gethostbyname(group, SUFFIX_DMB, target_ip, &num);
		if(!name_type){
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}
		for(ii = 0; ii < num; ii++) {
			ctx->rmt.naddr = target_ip[ii];
			memset(&nn, 0, sizeof(nn));
			memset(&ns, 0, sizeof(ns));
			res = smb_getnodestatus(&ctx->rmt, &nn, &ns, SUFFIX_SERV);
			if(res == LIBNBT_ERR_NOERR) {
				printf("smb_getnodestatus: name = %s\n", nn.name); 
#ifndef SMBCD_USE
				smbc_mfprintf("NcsSmbUserLoginSub: smb_getnodestatus: name = %s\n", nn.name); 
#endif
				strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
				bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
				strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
				if(ii > 0) {
					u_long tmp_ip;
					tmp_ip = target_ip[0];
					target_ip[0] = target_ip[ii];
					target_ip[ii] = tmp_ip;
				}
				break;
			}
		}
	}

	strncpy(server, vcp->server, SMB_MAXHOSTNAMELEN);
	strncpy(username, vcp->username, SMB_MAXUSERNAMELEN);
	strncpy(password, vcp->password, SMB_MAXPASSWORDLEN);

	bzero(&ctx->rmt, sizeof(ctx->rmt));
	ctx->rmt.suffix = SUFFIX_SERV;
	name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);
	if(!name_type) {
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. cannot resolve server name(server=%s).",  server);
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	} else {
		int res = 0;
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;

		memset(&nn, 0, sizeof(nn));
		memset(&ns, 0, sizeof(ns));
		res = nbt_getnodestatus(ctx->ns_client, &ctx->rmt, &nn, 1, &findnum, &ns, NULL);
		if(res >= 0) {
			printf("nbt_getnodestatus: name = %s\n", nn.name); 
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLoginSub: nbt_getnodestatus: name = %s\n", nn.name); 
#endif
			strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
		}
	}
	strncpy(ctx->rmt.name, server, SMB_MAXSRVNAMELEN);

NcsSmbUserLogin_retry:
	ctx->ss_client = nbt_open(NBT_STYPE_SESSION, htonl(INADDR_ANY), htons(NBT_PORT_ANY));
	if(ctx->ss_client < 0){
		printf("NcsSmbUserLogin: failed nbt_open\n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginSub: failed nbt_open\n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	res = nbt_connectTimeout(&ctx->ss_client, &ctx->rmt, &ctx->local, target_ip, num);
	if(res < 0) {
		printf("NcsSmbUserLogin: failed nbt_connect (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginSub failed: nbt_connectTimeout. \n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. failed to connect server(server=%s).", ctx->rmt.name );
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}

	res = nbt_getfd(ctx->ss_client);
	if(res < 0) {
		printf("NcsSmbUserLogin: failed nbt_getfd (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLoginSub: failed nbt_getfd (%d)\n", res);
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	ctx->cli->fd = res;
	ctx->cli->port = 139;
	fstrcpy(ctx->cli->desthost, server);
	fstrcpy(ctx->cli->user_name, username);
	fstrcpy(ctx->cli->domain, group);
	fstrcpy(ctx->cli->called.name, server);
	ctx->cli->called.name_type = SUFFIX_SERV;
	ctx->cli->dest_ip.s_addr = htonl(ctx->rmt.naddr);

	if(!cli_negprot(ctx->cli, &char_code)) {
		if(ctx->cli->fd < 0) {
			if(NcsSmbUserLogin_negprot_fail++ < 5) {
				usleep(smb_random()%1000+100);
				goto NcsSmbUserLogin_retry;
			}
			printf("cli_negprot fail %d\n", ERRNO);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLoginSub: cli_negprot fail %d\n", ERRNO);
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}

		printf("retry NULL domain\n");
		ctx->cli->domain[0] = NULL;
		if(!cli_negprot(ctx->cli, &char_code)) {
			int smberr;
			if(ctx->cli->fd < 0) {
				if(NcsSmbUserLogin_negprot_fail++ < 5) {
					usleep(smb_random()%1000+100);
					goto NcsSmbUserLogin_retry;
				}
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}
			printf("NcsSmbUserLogin: failed negprot\n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLoginSub: failed negprot\n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to negotiate(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		}
	}
#ifdef DEBUG
	printf("server_type [%s]\n", ctx->cli->server_type);
	printf("server_os [%s]\n", ctx->cli->server_os);
	printf("server_domain [%s]\n", ctx->cli->server_domain);
	printf("capabilities [0x%x]\n", ctx->cli->capabilities);
	printf("protocol [%d]\n", ctx->cli->protocol);
#endif	// DEBUG

	/*
	 * 2002/11/21 terao add
	 */
	if(ctx->cli->server_domain) {
		strncpy(vcp->group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
		strncpy(group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
	}

	if (!cli_session_setup(ctx->cli, username,
		password, strlen(password), 
		password, strlen(password), 
		"", &char_code)) {

		/*
		 * if a password was not supplied
		 * then try again with a null username
		 */
		 if (!password[0] || username[0] ||
				!cli_session_setup(ctx->cli, "", "", 0, 
				"", 0, group, &char_code)) {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}

			printf("NcsSmbUserLogin: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLoginSub: cli_session_setup failed: \n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 } else {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}

			printf("NcsSmbUserLogin: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLoginSub: cli_session_setup failed: \n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 }
	}

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbUserLogout
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbUserLogout( smb_ctx *ctx, SMBC_CALLBACK* callback )
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_ulogout_t	*th_arg = NULL;

	if(!ctx) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}

	if(callback){
		th_arg = (smbc_ulogout_t *)smb_malloc(sizeof(smbc_ulogout_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogout: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbUserLogoutThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogout: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogout: pthread_mutex_trylock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
		
		ret = NcsSmbUserLogoutSub(ctx);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserLogout: pthread_mutex_unlock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
	}
	return(ret);
}

void *NcsSmbUserLogoutThread(void* arg)
{
	int			ret = 0;
	int			status = 0;
	smbc_ulogout_t	*th_arg = (smbc_ulogout_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLogoutThread: pthread_detach() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGOUT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLogoutThread: pthread_mutex_trylock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGOUT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbUserLogoutSub(th_arg->ctx);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLogoutThread: pthread_mutex_unlock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGOUT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_USERLOGOUT, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbUserLogoutThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbUserLogoutSub( smb_ctx* ctx )
#else
int NcsSmbUserLogout( smb_ctx* ctx, SMBC_CALLBACK* callback )
#endif
{
	int	smberr = SMBSUCCESS;

	if(!ctx) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(!cli_ulogoff(ctx->cli)) {
		printf("NcsSmbUserLogout: cli_ulogoff failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserLogoutSub: cli_ulogoff failed: \n");
#endif
		syslog( LOG_ERR, "SMBC lib. failed to user logoff(server=%s).", ctx->rmt.name);
		if(ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else {
			smberr = smbc_errno(ctx->cli); 
		}
	}
	cli_shutdown(ctx->cli);
	ctx->cli = NULL;

	if(ctx->ss_client >= 0) {
	        nbt_close(ctx->ss_client);
		ctx->ss_client = -1;
	}

	return( smberr );
}

/**********************************************************
 * NcsSmbUserGetGroups
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbUserGetGroups( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list, SMBC_CALLBACK* callback)

{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_usrgroup_t		*th_arg = NULL;

	if(!ctx || !vcp || !ugp_list) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if(callback){
		/*
		 * set arg for pthread_create
		 */
		th_arg = (smbc_usrgroup_t *)smb_malloc(sizeof(smbc_usrgroup_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserGetGroups: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->max_list_num = max_list_num;
		th_arg->ugp_list= ugp_list;
		th_arg->callback = callback;

		/*
		 * pthread_create
		 */
		status = pthread_create(&thread_id, NULL, NcsSmbUserGetGroupsThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserGetGroups: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserGetGroups: pthread_mutex_trylock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
		
		ret = NcsSmbUserGetGroupsSub(ctx, vcp, max_list_num, ugp_list);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbUserGetGroups: pthread_mutex_unlock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
	}
	return(ret);
}

void *NcsSmbUserGetGroupsThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_usrgroup_t		*th_arg = (smbc_usrgroup_t *)arg;

	/*
	 * pthread_detach
	 */
	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroupsThread: pthread_detach() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERGETGRP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * mutex lock
	 */
	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroupsThread: pthread_mutex_trylock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERGETGRP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec Sub
	 */
	ret = NcsSmbUserGetGroupsSub(th_arg->ctx, th_arg->vcp, th_arg->max_list_num, th_arg->ugp_list);

	/*
	 * mutex unlock
	 */
	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroupsThread: pthread_mutex_unlock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERGETGRP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec callback
	 */
	th_arg->callback->func(SMBC_USERGETGRP, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbUserGetGroupsThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */


#ifdef SMBC_USE_PTL
int NcsSmbUserGetGroupsSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list )
#else
int NcsSmbUserGetGroups( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_ugp_list* ugp_list, SMBC_CALLBACK* callback)
#endif
{
	int	res = 0;
	smb_ugp_info	*ugp_info = 0;
	int	listnum = max_list_num;
	int	totalnum = 0;
	smb_ctx	*browse_ctx = 0;
	int	smberr = SMBSUCCESS;
	u_short	char_code = 0;
	int	i;

	if(!ctx || !vcp || !ugp_list) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	ugp_info = (smb_ugp_info *)smb_malloc(sizeof(smb_ugp_info)*max_list_num);
	if(!ugp_info) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroups: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	bzero(ugp_info, sizeof(smb_ugp_info)*max_list_num);
	res = NcsSmbOpen(&browse_ctx);
	if(res != SMBSUCCESS) {
		SAFE_FREE(ugp_info);
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	res = cli_SmbOpenPipe(browse_ctx, vcp, NULL);
	if(res != SMBSUCCESS) {
		printf("NcsSmbUserGetGroupsSub: cli_SmbOpenPipe failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroupsSub: cli_SmbOpenPipe \n");
#endif
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(ugp_info);
		return smberr;
	}

	if(cli_NetUserGetGroups(browse_ctx->cli, &totalnum, &listnum, ugp_info, vcp->username, &char_code) < 0) {
		printf("NcsSmbUserGetGroupsSub: cli_NetUserGetGroups\n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbUserGetGroupsSub: cli_NetUserGetGroups\n");
#endif
		if(browse_ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			smberr = smbc_errno(browse_ctx->cli);
		}
		NcsSmbClose(browse_ctx);
		syslog( LOG_ERR, "SMBC lib. cannot get share list(%s).", vcp->server);
		SAFE_FREE(ugp_info);
		return smberr;
	}

	ugp_list->entries = listnum;
	ugp_list->totalnum = totalnum;
	for(i = 0; i < listnum; i++) {
		ugp_list->ugp_info[i] = ugp_info[i];
	}
	res = NcsSmbDisconnect(browse_ctx, NULL);
	if(res != SMBSUCCESS) {
		smberr = smbc_errno(browse_ctx->cli); 
		NcsSmbClose(browse_ctx);
		SAFE_FREE(ugp_info);
		return smberr;
	}
	NcsSmbClose(browse_ctx);
	SAFE_FREE(ugp_info);

	return( SMBSUCCESS );
}



/**********************************************************
 * NcsSmbGetDCInfo
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbGetDCInfo(smb_ctx* ctx, smb_vc* vcp, smb_dc_info* dcinfo, SMBC_CALLBACK* callback)
{
	int				ret = 0;
	int				status = 0;
	pthread_t 		thread_id;
	smbc_dc_info_t	*th_arg = NULL;

	if(!ctx || !vcp) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if(callback){
		th_arg = (smbc_dc_info_t *)smb_malloc(sizeof(smbc_dc_info_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetDCInfo: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->dcinfo = dcinfo;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbGetDCInfoThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].\n",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetDCInfo: pthread_create() error[%d].\n",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetDCInfo: pthread_mutex_trylock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
		
		ret = NcsSmbGetDCInfoSub(ctx, vcp, dcinfo);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetDCInfo: pthread_mutex_unlock() error[%d].",status);
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			return(ret);
		}
	}
	return(ret);
}

void *NcsSmbGetDCInfoThread(void *arg)
{
	int		ret = 0;
	int		status = 0;
	smbc_dc_info_t  *th_arg = (smbc_dc_info_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetDCInfoThread: pthread_detach() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetDCInfoThread: pthread_mutex_trylock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbGetDCInfoSub(th_arg->ctx, th_arg->vcp, th_arg->dcinfo);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetDCInfoThread: pthread_mutex_unlock() error[%d].",status);
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_USERLOGIN, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);

#ifdef DEBUG
	printf("[NcsSmbGetDCInfoThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbGetDCInfoSub(smb_ctx* ctx, smb_vc* vcp, smb_dc_info* dcinfo)
#else
int NcsSmbGetDCInfo( smb_ctx* ctx, smb_vc* vcp, smb_dc_info* dcinfo, SMBC_CALLBACK* callback )
#endif
{
	int	res = 0;

	struct in_addr dest_ip;
	struct nmb_name called;
	NBTAddrPack	rmt;
	char            group[MAX(SMB_MAXSRVNAMELEN,SMB_MAXHOSTNAMELEN)+1];
	char		server[SMB_MAXHOSTNAMELEN+1];
	char		username[SMB_MAXUSERNAMELEN+1];
	char		password[SMB_MAXPASSWORDLEN+1];
	u_short		char_code = vcp->character;
	int		name_type = 0;
	int		NcsSmbGetDCInfo_negprot_fail = 0;
	u_long		target_ip[SMB_MAX_IP];
	int		num = SMB_MAX_IP;
	smb_ctx		*local_ctx = 0;	/* @terao20040210 */

	if(!ctx || !vcp) {
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	
	res = NcsSmbOpen(&local_ctx);
	if(res != SMBSUCCESS) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: NcsSmbOpen. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	local_ctx->cli = cli_initialise(NULL);
	if(!local_ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: cli is NULL. \n");
#endif
		NcsSmbClose(local_ctx);
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	bzero(&dest_ip, sizeof(dest_ip));
	bzero(&called, sizeof(called));
	bzero(&rmt, sizeof(rmt));
	bzero(group, sizeof(group));
	bzero(server, sizeof(server));
	bzero(username, sizeof(username));
	bzero(password, sizeof(password));

	if(vcp->group[0] != 0) {
		strncpy(group, vcp->group, sizeof(group)-1);
	} else {
		smb_get_mygroup(group);
	}
	if(vcp->server[0] != 0) {
		strncpy(server, (char *)vcp->server, SMB_MAXSRVNAMELEN);
		local_ctx->rmt.suffix = SUFFIX_SERV;
		name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);
		if(!name_type) {
			syslog( LOG_ERR, "SMBC lib. cannot resolve server name(%s).",  server);
#ifndef SMBCD_USE
	                smbc_mfprintf("NcsSmbGetDCInfoSub failed: smb_gethostbyname. \n");
#endif
			NcsSmbClose(local_ctx);
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else if(name_type != 1) {
			int res = 0;
			NBTNodeName nn;
			NBTNodeStatistics ns;
			int findnum = 0;
			int ii = 0;

			for(ii = 0; ii < num; ii++) {
				local_ctx->rmt.naddr = target_ip[ii];
				memset(&nn, 0, sizeof(nn));
				memset(&ns, 0, sizeof(ns));
				res = smb_getnodestatus(&local_ctx->rmt, &nn, &ns, SUFFIX_SERV);
				if(res == LIBNBT_ERR_NOERR) {
					printf("smb_getnodestatus: name = %s\n", nn.name);
					strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
					bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
					strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
					if(ii > 0) {
						u_long tmp_ip;
						tmp_ip = target_ip[0];
						target_ip[0] = target_ip[ii];
						target_ip[ii] = tmp_ip;
					}
					break;
				}
			}
		}
		strncpy(local_ctx->rmt.name, server, SMB_MAXSRVNAMELEN);
	} else {
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;
		int ii = 0;

		/* get domain controller */ 
		name_type = smb_gethostbyname(group, SUFFIX_DMB, target_ip, &num);
		if(!name_type) {
			NcsSmbClose(local_ctx);
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}
		for(ii = 0; ii < num; ii++) {
			local_ctx->rmt.naddr = target_ip[ii];
			memset(&nn, 0, sizeof(nn));
			memset(&ns, 0, sizeof(ns));
			res = smb_getnodestatus(&local_ctx->rmt, &nn, &ns, SUFFIX_DMB);
			if(res == LIBNBT_ERR_NOERR) {
				printf("smb_getnodestatus: name = %s\n", nn.name); 
#ifndef SMBCD_USE
				smbc_mfprintf("NcsSmbGetDCInfoSub: smb_getnodestatus: name = %s\n", nn.name); 
#endif
				strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
				bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
				strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
				if(ii > 0) {
					u_long tmp_ip;
					tmp_ip = target_ip[0];
					target_ip[0] = target_ip[ii];
					target_ip[ii] = tmp_ip;
				}
				break;
			}
		}
	}

	dcinfo->address = target_ip[0];
	NcsSmbClose(local_ctx);
	return( SMBSUCCESS );
}

#endif /* NCS_SMB_NTAUTH_SUPPORT */
