/*
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba system utilities
   Copyright (C) Andrew Tridgell 1992-1998

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"

#define FTAPI_ABORT_EXTERN	/* Oe[uQ */

#define FTAPI_ABORT_REQOFF              0       /* fvȂ                         */
#define FTAPI_ABORT_INTERVAL            1       /* 1b                                  */


#ifndef ENOSYS
#define ENOSYS	(-1)
#endif

/*
   The idea is that this file will eventually have wrappers around all
   important system calls in samba. The aims are:

   - to enable easier porting by putting OS dependent stuff in here

   - to allow for hooks into other "pseudo-filesystems"

   - to allow easier integration of things like the japanese extensions

   - to support the philosophy of Samba to expose the features of
     the OS within the SMB model. In general whatever file/printer/variable
     expansions/etc make sense to the OS should be acceptable to Samba.
*/

#define	SMB_SEND_TIMEOUT	60	/* 60 sec */
#define	SMB_RECV_TIMEOUT	120	/* 60 sec */

#if 0
/*******************************************************************
 A wrapper for usleep in case we don't have one.
********************************************************************/

int smb_sys_usleep(long usecs)
{
#ifndef HAVE_USLEEP
	struct timeval tval;
#endif

	/*
	 * We need this braindamage as the glibc usleep
	 * is not SPEC1170 complient... grumble... JRA.
	 */

	if(usecs < 0 || usecs > 1000000) {
		ERRNO = EINVAL;
		return -1;
	}

#if HAVE_USLEEP
	usleep(usecs);
	return 0;
#else /* HAVE_USLEEP */
	/*
	 * Fake it with select...
	 */
	tval.tv_sec = 0;
	tval.tv_usec = usecs/1000;
	select(0,NULL,NULL,NULL,&tval);
	return 0;
#endif /* HAVE_USLEEP */
}
#endif

/*******************************************************************
A read wrapper that will deal with EINTR.
********************************************************************/

ssize_t smb_sys_read(int fd, void *buf, size_t count)
{
	ssize_t ret;
	fd_set	readfds;
	struct	timeval	timeout;
	int	retrycnt = 0;			/* ĎgCJE^	*/
	int	retrymax = 0;			/* ĎgC	*/
	int	retryendinterval = 0;		/* ĎŏIC^[o	*/
	int	option = 0;			/* \PbgIvVerrno	*/
	int	optionlen = sizeof(option);	/* \PbgIvVerrno	*/
	int	abort = FTAPI_ABORT_REQOFF;	/* fv		*/

	bzero(&timeout, sizeof(timeout));

	/* select-sendĎ */
	retrymax = (SMB_RECV_TIMEOUT / FTAPI_ABORT_INTERVAL) + 1;	/* gC񐔐ݒ	*/
	retryendinterval = SMB_RECV_TIMEOUT % FTAPI_ABORT_INTERVAL;	/* ŏIC^[oݒ	*/
	if( retryendinterval == 0 )
	{							/* ŏIC^[o0̎ČvZ	*/
		retrymax -= 1;
		retryendinterval = FTAPI_ABORT_INTERVAL;
	}

	for(retrycnt = 0; retrycnt < retrymax; retrycnt++) {
#ifndef SMBCD_USE
		/* fv`FbN */
		abort = FtapiAbortReqCheck();
		if( abort == FTAPI_ABORT_REQON )
		{
			smbc_mfprintf("smb_sys_read: connection aborted\n");
			return -1;
		}
#endif
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		if( retrycnt == (retrymax - 1) )
		{
			timeout.tv_sec = retryendinterval;	/* ŏIC^[o	*/
		} else {
			timeout.tv_sec = FTAPI_ABORT_INTERVAL;
		}
		timeout.tv_usec = 0;
		select(fd+1, &readfds, NULL, NULL, &timeout);
		if(FD_ISSET (fd, &readfds)) {
			break;
		}
	}
	if(retrycnt >= retrymax) {
		printf("can not read data\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_sys_read: can not read data\n");
#endif
		return -1;
	}

	do {
		ret = read(fd, buf, count);

#ifndef SMBCD_USE
		abort = FtapiAbortReqCheck();
		if(abort == FTAPI_ABORT_REQON) {
			smbc_mfprintf("smb_sys_read: abort\n");
			ret = -1;
			break;
		}
#endif
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}

/*******************************************************************
A send wrapper that will deal with EINTR.
********************************************************************/
ssize_t smb_sys_send(int s, const void *msg, size_t len, int flags)
{
	ssize_t ret = 0;
	fd_set	readfds,writefds;
	struct	timeval	timeout;
	u_char	buf[10];
	int	i = 0;
	sig_t	oldintp = NULL;
	int	retrycnt = 0;			/* ĎgCJE^	*/
	int	retrymax = 0;			/* ĎgC	*/
	int	retryendinterval = 0;		/* ĎŏIC^[o	*/
	int	option = 0;			/* \PbgIvVerrno	*/
	int	optionlen = sizeof(option);	/* \PbgIvVerrno	*/
	int	abort = FTAPI_ABORT_REQOFF;	/* fv		*/

	bzero(&timeout, sizeof(timeout));

	/* select-sendĎ */
	retrymax = (SMB_SEND_TIMEOUT / FTAPI_ABORT_INTERVAL) + 1;	/* gC񐔐ݒ	*/
	retryendinterval = SMB_SEND_TIMEOUT % FTAPI_ABORT_INTERVAL;	/* ŏIC^[oݒ	*/
	if( retryendinterval == 0 )
	{							/* ŏIC^[o0̎ČvZ	*/
		retrymax -= 1;
		retryendinterval = FTAPI_ABORT_INTERVAL;
	}

	for(retrycnt = 0; retrycnt < retrymax; retrycnt++) {

#ifndef SMBCD_USE
		/* fv`FbN */
		abort = FtapiAbortReqCheck();
		if( abort == FTAPI_ABORT_REQON )
		{
			smbc_mfprintf("smb_sys_send: connection aborted\n");
			return -1;
		}
#endif
		FD_ZERO(&readfds);
		FD_SET(s, &readfds);
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		select(s+1, &readfds, NULL, NULL, &timeout);
		if(FD_ISSET (s, &readfds)) {
			if(recv(s, buf, sizeof(buf), MSG_PEEK) <= 0) {
				printf("connection refused\n");
#ifndef SMBCD_USE
				smbc_mfprintf("smb_sys_send: connection refused\n");
#endif
				return -1;
			}
		}

		FD_ZERO(&writefds);
		FD_SET(s, &writefds);
		if( retrycnt == (retrymax - 1) )
		{
			timeout.tv_sec = retryendinterval;			/* ŏIC^[o	*/
		} else {
			timeout.tv_sec=FTAPI_ABORT_INTERVAL;
		}
		timeout.tv_usec=0;
		select(s+1, NULL, &writefds, NULL, &timeout);
		if(FD_ISSET (s, &writefds)) {
			break;
		}
	}
	if(retrycnt >= retrymax) {
		printf("can not send data\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_sys_send: can not send data\n");
#endif
		return -1;
	}
	do {
		oldintp = signal( SIGPIPE, SIG_IGN );
		ret = send(s, msg, len, flags);
		if( oldintp ){ (void)signal( SIGPIPE, oldintp ); }

#ifndef SMBCD_USE
		abort = FtapiAbortReqCheck();
		if(abort == FTAPI_ABORT_REQON) {
			smbc_mfprintf("smb_sys_send: abort\n");
			ret = -1;
			break;
		}
#endif
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}

/*
------------------------------------------------------------------
smb_writable()

\Pbg݉\ׂ


	fd:		\PbgfBXNv^
	msec: 	҂(MilliSeconds)
			:-1		҂
			:0		҂

߂l
	:0	ݕsi^CAEgj
	:1	݉
	:-1	select֐G[(errno)
	:-2 f
------------------------------------------------------------------
*/
int smb_writable(int fd, long msec, int interruptd)
{
	fd_set wset;
	fd_set rset;

	struct timeval tv;
	int ret;
	int maxfd;

	if(msec < -1){
		return -1;
	}
	FD_ZERO(&rset);
	FD_ZERO(&wset);
	FD_SET(fd, &wset);

	if(interruptd != -1){
		FD_SET(interruptd, &rset);
		maxfd = MAX(fd, interruptd);
	}
	else{
		maxfd = fd;
	}

	if(msec != -1){
		tv.tv_sec = msec/1000;
		msec %= 1000;
		msec *= 1000;	// millisec --> microsec

		tv.tv_usec = msec;
	}

	ret = select(maxfd+1, &rset, &wset, NULL, (msec == -1)? NULL: &tv);
	if(!ret){
		return 0;
	}
	if (ret < 0){
		return -1;
	}

	if(interruptd != -1){
		if(FD_ISSET(interruptd, &rset)){
			return -2;
		}
	}

	return 1;
}

/*******************************************************************
A sendto wrapper that will deal with EINTR.
********************************************************************/

ssize_t smb_sys_sendto(int s,  const void *msg, size_t len, int flags, const struct sockaddr *to, socklen_t tolen)
{
	ssize_t ret;

	do {
		ret = sendto(s, msg, len, flags, to, tolen);
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}

/*******************************************************************
A recvfrom wrapper that will deal with EINTR.
********************************************************************/

ssize_t smb_sys_recvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from, socklen_t *fromlen)
{
	ssize_t ret;

	do {
		ret = recvfrom(s, buf, len, flags, from, fromlen);
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}

#if 0
/*******************************************************************
A fcntl wrapper that will deal with EINTR.
********************************************************************/

int smb_sys_fcntl_ptr(int fd, int cmd, void *arg)
{
	int ret;

	do {
		ret = fcntl(fd, cmd, arg);
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}
#endif

/*******************************************************************
A fcntl wrapper that will deal with EINTR.
********************************************************************/

int smb_sys_fcntl_long(int fd, int cmd, long arg)
{
	int ret;

	do {
		ret = fcntl(fd, cmd, arg);
	} while (ret == -1 && ERRNO == EINTR);
	return ret;
}

#if 0
/*******************************************************************
A stat() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_stat(const char *fname,SMB_STRUCT_STAT *sbuf)
{
	int ret;
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OFF64_T) && defined(HAVE_STAT64)
	ret = stat64(fname, sbuf);
#else
	ret = stat(fname, sbuf);
#endif
	/* we always want directories to appear zero size */
	if (ret == 0 && S_ISDIR(sbuf->st_mode)) sbuf->st_size = 0;
	return ret;
}
#endif

#if 0
/*******************************************************************
 An fstat() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_fstat(int fd,SMB_STRUCT_STAT *sbuf)
{
	int ret;
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OFF64_T) && defined(HAVE_FSTAT64)
	ret = fstat64(fd, sbuf);
#else
	ret = fstat(fd, sbuf);
#endif
	/* we always want directories to appear zero size */
	if (ret == 0 && S_ISDIR(sbuf->st_mode)) sbuf->st_size = 0;
	return ret;
}
#endif

#if 0
/*******************************************************************
 An lstat() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_lstat(const char *fname,SMB_STRUCT_STAT *sbuf)
{
	int ret;
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OFF64_T) && defined(HAVE_LSTAT64)
	ret = lstat64(fname, sbuf);
#else
	ret = lstat(fname, sbuf);
#endif
	/* we always want directories to appear zero size */
	if (ret == 0 && S_ISDIR(sbuf->st_mode)) sbuf->st_size = 0;
	return ret;
}
#endif

#if 0
/*******************************************************************
 An ftruncate() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_ftruncate(int fd, SMB_OFF_T offset)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OFF64_T) && defined(HAVE_FTRUNCATE64)
	return ftruncate64(fd, offset);
#else
	return ftruncate(fd, offset);
#endif
}
#endif

#if 0
/*******************************************************************
 An lseek() wrapper that will deal with 64 bit filesizes.
********************************************************************/

SMB_OFF_T smb_sys_lseek(int fd, SMB_OFF_T offset, int whence)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OFF64_T) && defined(HAVE_LSEEK64)
	return lseek64(fd, offset, whence);
#else
	return lseek(fd, offset, whence);
#endif
}
#endif

#if 0
/*******************************************************************
 An fseek() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_fseek(FILE *fp, SMB_OFF_T offset, int whence)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(LARGE_SMB_OFF_T) && defined(HAVE_FSEEK64)
	return fseek64(fp, offset, whence);
#elif defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(LARGE_SMB_OFF_T) && defined(HAVE_FSEEKO64)
	return fseeko64(fp, offset, whence);
#else
	return fseek(fp, offset, whence);
#endif
}
#endif

#if 0
/*******************************************************************
 An ftell() wrapper that will deal with 64 bit filesizes.
********************************************************************/

SMB_OFF_T smb_sys_ftell(FILE *fp)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(LARGE_SMB_OFF_T) && defined(HAVE_FTELL64)
	return (SMB_OFF_T)ftell64(fp);
#elif defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(LARGE_SMB_OFF_T) && defined(HAVE_FTELLO64)
	return (SMB_OFF_T)ftello64(fp);
#else
	return (SMB_OFF_T)ftell(fp);
#endif
}
#endif

#if 0
/*******************************************************************
 A creat() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_creat(const char *path, mode_t mode)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_CREAT64)
	return creat64(path, mode);
#else
	/*
	 * If creat64 isn't defined then ensure we call a potential open64.
	 * JRA.
	 */
	return smb_sys_open(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
#endif
}
#endif

#if 0
/*******************************************************************
 An open() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_open(const char *path, int oflag, mode_t mode)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_OPEN64)
	return open64(path, oflag, mode);
#else
	return open(path, oflag, mode);
#endif
}
#endif

#if 0
/*******************************************************************
 An fopen() wrapper that will deal with 64 bit filesizes.
********************************************************************/

FILE *smb_sys_fopen(const char *path, const char *type)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_FOPEN64)
	return fopen64(path, type);
#else
	return fopen(path, type);
#endif
}
#endif

#if 0
/*******************************************************************
 A readdir wrapper that will deal with 64 bit filesizes.
********************************************************************/

SMB_STRUCT_DIRENT *smb_sys_readdir(DIR *dirp)
{
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_READDIR64)
	return readdir64(dirp);
#else
	return readdir(dirp);
#endif
}
#endif

#if 0
/*******************************************************************
 An mknod() wrapper that will deal with 64 bit filesizes.
********************************************************************/

int smb_sys_mknod(const char *path, mode_t mode, SMB_DEV_T dev)
{
#if defined(HAVE_MKNOD) || defined(HAVE_MKNOD64)
#if defined(HAVE_EXPLICIT_LARGEFILE_SUPPORT) && defined(HAVE_MKNOD64) && defined(HAVE_DEV64_T)
	return mknod64(path, mode, dev);
#else
	return mknod(path, mode, dev);
#endif
#else
	/* No mknod system call. */
	ERRNO = ENOSYS;
	return -1;
#endif
}
#endif

#if 0
/*******************************************************************
 Wrapper for realpath.
********************************************************************/

char *smb_sys_realpath(const char *path, char *resolved_path)
{
#if defined(HAVE_REALPATH)
	return (char *)realpath(path, resolved_path);
#else
	/* As realpath is not a system call we can't return ENOSYS. */
	ERRNO = EINVAL;
	return NULL;
#endif
}
#endif

/*******************************************************************
The wait() calls vary between systems
********************************************************************/

int smb_sys_waitpid(pid_t pid,int *status,int options)
{
#ifdef HAVE_WAITPID
	return waitpid(pid,status,options);
#else /* HAVE_WAITPID */
	return wait4(pid, status, options, NULL);
#endif /* HAVE_WAITPID */
}

/*******************************************************************
 System wrapper for getwd
********************************************************************/

char *smb_sys_getwd(char *s)
{
	char *wd;
#ifdef HAVE_GETCWD
	wd = (char *)getcwd(s, sizeof (pstring));
#else
	wd = (char *)getwd(s);
#endif
	return wd;
}

#if 0
/*******************************************************************
system wrapper for symlink
********************************************************************/

int smb_sys_symlink(const char *oldpath, const char *newpath)
{
#ifndef HAVE_SYMLINK
	ERRNO = ENOSYS;
	return -1;
#else
	return symlink(oldpath, newpath);
#endif
}
#endif

#if 0
/*******************************************************************
system wrapper for readlink
********************************************************************/

int smb_sys_readlink(const char *path, char *buf, size_t bufsiz)
{
#ifndef HAVE_READLINK
	ERRNO = ENOSYS;
	return -1;
#else
	return readlink(path, buf, bufsiz);
#endif
}
#endif

#if 0
/*******************************************************************
system wrapper for link
********************************************************************/

int smb_sys_link(const char *oldpath, const char *newpath)
{
#ifndef HAVE_LINK
	ERRNO = ENOSYS;
	return -1;
#else
	return link(oldpath, newpath);
#endif
}
#endif

#if 0
/*******************************************************************
chown isn't used much but OS/2 doesn't have it
********************************************************************/

int smb_sys_chown(const char *fname,uid_t uid,gid_t gid)
{
#ifndef HAVE_CHOWN
	static int done;
	if (!done) {
		DEBUG(1,("WARNING: no chown!\n"));
		done=1;
	}
#else
	return(chown(fname,uid,gid));
#endif
}
#endif

#if 0
/*******************************************************************
os/2 also doesn't have chroot
********************************************************************/
int smb_sys_chroot(const char *dname)
{
#ifndef HAVE_CHROOT
	static int done;
	if (!done) {
		DEBUG(1,("WARNING: no chroot!\n"));
		done=1;
	}
	ERRNO = ENOSYS;
	return -1;
#else
	return(chroot(dname));
#endif
}
#endif

/**************************************************************************
A wrapper for gethostbyname() that tries avoids looking up hostnames 
in the root domain, which can cause dial-on-demand links to come up for no
apparent reason.
****************************************************************************/

struct hostent *smb_sys_gethostbyname(const char *name)
{
#ifdef REDUCE_ROOT_DNS_LOOKUPS
	char query[256], hostname[256];
	char *domain;

	/* Does this name have any dots in it? If so, make no change */

	if (strchr(name, '.'))
		return(gethostbyname(name));

	/* Get my hostname, which should have domain name 
		attached. If not, just do the gethostname on the
		original string. 
	*/

	gethostname(hostname, sizeof(hostname) - 1);
	hostname[sizeof(hostname) - 1] = 0;
	if ((domain = strchr(hostname, '.')) == NULL)
		return(gethostbyname(name));

	/* Attach domain name to query and do modified query.
		If names too large, just do gethostname on the
		original string.
	*/

	if((strlen(name) + strlen(domain)) >= sizeof(query))
		return(gethostbyname(name));

	slprintf(query, sizeof(query)-1, "%s%s", name, domain);
	return(gethostbyname(query));
#else /* REDUCE_ROOT_DNS_LOOKUPS */
	return(gethostbyname(name));
#endif /* REDUCE_ROOT_DNS_LOOKUPS */
}

#if 0
#if defined(HAVE_IRIX_SPECIFIC_CAPABILITIES)
/**************************************************************************
 Try and abstract process capabilities (for systems that have them).
****************************************************************************/
static BOOL smb_set_process_capability( uint32 cap_flag, BOOL enable )
{
	if(cap_flag == KERNEL_OPLOCK_CAPABILITY) {
		cap_t cap = cap_get_proc();

		if (cap == NULL) {
			DEBUG(0,("set_process_capability: cap_get_proc failed. Error was %s\n",
				strerror(ERRNO)));
			return False;
		}

		if(enable)
			cap->cap_effective |= CAP_NETWORK_MGT;
		else
			cap->cap_effective &= ~CAP_NETWORK_MGT;

		if (cap_set_proc(cap) == -1) {
			DEBUG(0,("set_process_capability: cap_set_proc failed. Error was %s\n",
				strerror(ERRNO)));
			cap_free(cap);
			return False;
		}

		cap_free(cap);

		DEBUG(10,("set_process_capability: Set KERNEL_OPLOCK_CAPABILITY.\n"));
	}
	return True;
}

/**************************************************************************
 Try and abstract inherited process capabilities (for systems that have them).
****************************************************************************/

static BOOL smb_set_inherited_process_capability( uint32 cap_flag, BOOL enable )
{
	if(cap_flag == KERNEL_OPLOCK_CAPABILITY) {
		cap_t cap = cap_get_proc();

		if (cap == NULL) {
			DEBUG(0,("set_inherited_process_capability: cap_get_proc failed. Error was %s\n",
				strerror(ERRNO)));
			return False;
		}

		if(enable)
			cap->cap_inheritable |= CAP_NETWORK_MGT;
		else
			cap->cap_inheritable &= ~CAP_NETWORK_MGT;

		if (cap_set_proc(cap) == -1) {
			DEBUG(0,("set_inherited_process_capability: cap_set_proc failed. Error was %s\n", 
				strerror(ERRNO)));
			cap_free(cap);
			return False;
		}

		cap_free(cap);

		DEBUG(10,("set_inherited_process_capability: Set KERNEL_OPLOCK_CAPABILITY.\n"));
	}
	return True;
}
#endif
#endif

#if 0
/****************************************************************************
 Gain the oplock capability from the kernel if possible.
****************************************************************************/

void smb_oplock_set_capability(BOOL this_process, BOOL inherit)
{
#if HAVE_KERNEL_OPLOCKS_IRIX
	set_process_capability(KERNEL_OPLOCK_CAPABILITY,this_process);
	set_inherited_process_capability(KERNEL_OPLOCK_CAPABILITY,inherit);
#endif
}
#endif

#if 0
/**************************************************************************
 Wrapper for random().
****************************************************************************/

long smb_sys_random(void)
{
#if defined(HAVE_RANDOM)
	return (long)random();
#elif defined(HAVE_RAND)
	return (long)rand();
#else
	DEBUG(0,("Error - no random function available !\n"));
	exit(1);
#endif
}
#endif

#if 0
/**************************************************************************
 Wrapper for srandom().
****************************************************************************/

void smb_sys_srandom(unsigned int seed)
{
#if defined(HAVE_SRANDOM)
	srandom(seed);
#elif defined(HAVE_SRAND)
	srand(seed);
#else
	DEBUG(0,("Error - no srandom function available !\n"));
	exit(1);
#endif
}
#endif

#if 0
/**************************************************************************
 Returns equivalent to NGROUPS_MAX - using sysconf if needed.
****************************************************************************/

int smb_groups_max(void)
{
#if defined(SYSCONF_SC_NGROUPS_MAX)
	int ret = sysconf(_SC_NGROUPS_MAX);
	return (ret == -1) ? NGROUPS_MAX : ret;
#else
	return NGROUPS_MAX;
#endif
}
#endif

#if 0
/**************************************************************************
 Wrapper for getgroups. Deals with broken (int) case.
****************************************************************************/

int smb_sys_getgroups(int setlen, gid_t *gidset)
{
#if !defined(HAVE_BROKEN_GETGROUPS)
	return getgroups(setlen, gidset);
#else

	GID_T gid;
	GID_T *group_list;
	int i, ngroups;

	if(setlen == 0) {
		return getgroups(setlen, &gid);
	}

	/*
	 * Broken case. We need to allocate a
	 * GID_T array of size setlen.
	 */

	if(setlen < 0) {
		ERRNO = EINVAL; 
		return -1;
	} 

	if (setlen == 0)
		setlen = groups_max();

	if((group_list = (GID_T *)smb_malloc(setlen * sizeof(GID_T))) == NULL) {
		DEBUG(0,("sys_getgroups: smb_malloc fail.\n"));
		return -1;
	}

	if((ngroups = getgroups(setlen, group_list)) < 0) {
		int saved_errno = ERRNO;
		SAFE_FREE(group_list);
		ERRNO = saved_errno;
		return -1;
	}

	for(i = 0; i < ngroups; i++)
		gidset[i] = (gid_t)group_list[i];

	SAFE_FREE(group_list);
	return ngroups;
#endif /* HAVE_BROKEN_GETGROUPS */
}
#endif

#ifdef HAVE_SETGROUPS

#if 0
/**************************************************************************
 Wrapper for setgroups. Deals with broken (int) case. Automatically used
 if we have broken getgroups.
****************************************************************************/

int smb_sys_setgroups(int setlen, gid_t *gidset)
{
#if !defined(HAVE_BROKEN_GETGROUPS)
	return setgroups(setlen, gidset);
#else

	GID_T *group_list;
	int i ; 

	if (setlen == 0)
		return 0 ;

	if (setlen < 0 || setlen > groups_max()) {
		ERRNO = EINVAL; 
		return -1;   
	}

	/*
	 * Broken case. We need to allocate a
	 * GID_T array of size setlen.
	 */

	if((group_list = (GID_T *)smb_malloc(setlen * sizeof(GID_T))) == NULL) {
		DEBUG(0,("sys_setgroups: smb_malloc fail.\n"));
		return -1;    
	}
 
	for(i = 0; i < setlen; i++) 
		group_list[i] = (GID_T) gidset[i]; 

	if(setgroups(setlen, group_list) != 0) {
		int saved_errno = ERRNO;
		SAFE_FREE(group_list);
		ERRNO = saved_errno;
		return -1;
	}
 
	SAFE_FREE(group_list);
	return 0 ;
#endif /* HAVE_BROKEN_GETGROUPS */
}
#endif

#endif /* HAVE_SETGROUPS */

/*
 * We only wrap pw_name and pw_passwd for now as these
 * are the only potentially modified fields.
 */

/**************************************************************************
 Helper function for getpwnam/getpwuid wrappers.
****************************************************************************/

struct saved_pw {
	fstring 	pw_name;
	fstring 	pw_passwd;
	fstring		pw_gecos;
	pstring		pw_dir;
	pstring		pw_shell;
	struct passwd pass;
};

#if 0
static struct saved_pw pw_mod; /* This is the structure returned - can be modified. */
static struct saved_pw pw_cache; /* This is the structure saved - used to check cache. */
#endif

#if 0
static int num_lookups; /* Counter so we don't always use cache. */
#endif
#ifndef PW_RET_CACHE_MAX_LOOKUPS
#define PW_RET_CACHE_MAX_LOOKUPS 100
#endif

#if 0
static void smb_copy_pwent(struct saved_pw *dst, struct passwd *pass)
{
	memcpy((char *)&dst->pass, pass, sizeof(struct passwd));

	unix_to_dos(dst->pw_name, pass->pw_name);
	dst->pass.pw_name = dst->pw_name;

	fstrcpy(dst->pw_passwd, pass->pw_passwd);
	dst->pass.pw_passwd = dst->pw_passwd;

	unix_to_dos(dst->pw_gecos, pass->pw_gecos);
	dst->pass.pw_gecos = dst->pw_gecos;

	unix_to_dos(dst->pw_dir, pass->pw_dir);
	dst->pass.pw_dir = dst->pw_dir;

	pstrcpy(dst->pw_shell, pass->pw_shell);
	dst->pass.pw_shell = dst->pw_shell;
}
#endif

#if 0
static struct passwd *smb_setup_pwret(struct passwd *pass)
{
	if (pass == NULL) {
		/* Clear the caches. */
		memset(&pw_cache, '\0', sizeof(struct saved_pw));
		memset(&pw_mod, '\0', sizeof(struct saved_pw));
		num_lookups = 0;
		return NULL;
	}

	smb_copy_pwent( &pw_mod, pass);

	if (pass != &pw_cache.pass) {

		/* If it's a cache miss we must also refill the cache. */

		smb_copy_pwent( &pw_cache, pass);
		num_lookups = 1;

	} else {

		/* Cache hit. */

		num_lookups++;
		num_lookups = (num_lookups % PW_RET_CACHE_MAX_LOOKUPS);
	}

	return &pw_mod.pass;
}
#endif

#if 0
/**************************************************************************
 Wrappers for setpwent(), getpwent() and endpwent()
****************************************************************************/

void smb_sys_setpwent(void)
{
	smb_setup_pwret(NULL); /* Clear cache. */
	setpwent();
}

struct passwd *smb_sys_getpwent(void)
{
	return smb_setup_pwret(getpwent());
}

void smb_sys_endpwent(void)
{
	smb_setup_pwret(NULL); /* Clear cache. */
	endpwent();
}
#endif

#if 0
/**************************************************************************
 Wrapper for getpwnam(). Always returns a static that can be modified.
****************************************************************************/

struct passwd *smb_sys_getpwnam(const char *name)
{
	pstring s;

	if (!name || !name[0])
		return NULL;

	dos_to_unix(s,name);
	/* check for a cache hit first */
	if (num_lookups && pw_cache.pass.pw_name && !strcmp(s, pw_cache.pass.pw_name)) {
		return smb_setup_pwret(&pw_cache.pass);
	}

	return smb_setup_pwret(getpwnam(s));
}
#endif

#if 0
/**************************************************************************
 Wrapper for getpwuid(). Always returns a static that can be modified.
****************************************************************************/

struct passwd *smb_sys_getpwuid(uid_t uid)
{
	if (num_lookups && pw_cache.pass.pw_name && (uid == pw_cache.pass.pw_uid)) {
		return smb_setup_pwret(&pw_cache.pass);
	}
	
  	return smb_setup_pwret(getpwuid(uid));
}
#endif

#if 0
/**************************************************************************
 Extract a command into an arg list. Uses a static pstring for storage.
 Caller frees returned arg list (which contains pointers into the static pstring).
****************************************************************************/

static char **smb_extract_args(const char *command)
{
	static pstring trunc_cmd;
	char *ptr;
	int argcl;
	char **argl = NULL;
	int i;

	pstrcpy(trunc_cmd, command);

	if(!(ptr = strtok(trunc_cmd, " \t"))) {
		ERRNO = EINVAL;
		return NULL;
	}

	/*
	 * Count the args.
	 */

	for( argcl = 1; ptr; ptr = strtok(NULL, " \t"))
		argcl++;

	if((argl = (char **)smb_malloc((argcl + 1) * sizeof(char *))) == NULL)
		return NULL;

	/*
	 * Now do the extraction.
	 */

	pstrcpy(trunc_cmd, command);

	ptr = strtok(trunc_cmd, " \t");
	i = 0;
	argl[i++] = ptr;

	while((ptr = strtok(NULL, " \t")) != NULL)
		argl[i++] = ptr;

	argl[i++] = NULL;
	return argl;
}
#endif

/**************************************************************************
 Wrapper for fork. Ensures that mypid is reset. Used so we can write
 a sys_getpid() that only does a system call *once*.
****************************************************************************/

static pid_t mypid = (pid_t)-1;

pid_t smb_sys_fork(void)
{
	pid_t forkret = fork();

	if (forkret == (pid_t)0) /* Child - reset mypid so sys_getpid does a system call. */
		mypid = (pid_t) -1;

	return forkret;
}

/**************************************************************************
 Wrapper for getpid. Ensures we only do a system call *once*.
****************************************************************************/

pid_t smb_sys_getpid(void)
{
	if (mypid == (pid_t)-1)
		mypid = getpid();

	return mypid;
}

#if 0
/**************************************************************************
 Wrapper for popen. Safer as it doesn't search a path.
 Modified from the glibc sources.
 modified by tridge to return a file descriptor. We must kick our FILE* habit
****************************************************************************/

typedef struct _popen_list
{
	int fd;
	pid_t child_pid;
	struct _popen_list *next;
} popen_list;

static popen_list *popen_chain;

int smb_sys_popen(const char *command)
{
	int parent_end, child_end;
	int pipe_fds[2];
	popen_list *entry = NULL;
	char **argl = NULL;

	if (pipe(pipe_fds) < 0)
		return -1;

	parent_end = pipe_fds[0];
	child_end = pipe_fds[1];

	if (!*command) {
		ERRNO = EINVAL;
		goto err_exit;
	}

	if((entry = (popen_list *)smb_malloc(sizeof(popen_list))) == NULL)
		goto err_exit;

	ZERO_STRUCTP(entry);

	/*
	 * Extract the command and args into a NULL terminated array.
	 */

	if(!(argl = smb_extract_args(command)))
		goto err_exit;

	entry->child_pid = smb_sys_fork();

	if (entry->child_pid == -1) {
		goto err_exit;
	}

	if (entry->child_pid == 0) {

		/*
		 * Child !
		 */

		int child_std_end = STDOUT_FILENO;
		popen_list *p;

		close(parent_end);
		if (child_end != child_std_end) {
			dup2 (child_end, child_std_end);
			close (child_end);
		}

		/*
		 * POSIX.2:  "popen() shall ensure that any streams from previous
		 * popen() calls that remain open in the parent process are closed
		 * in the new child process."
		 */

		for (p = popen_chain; p; p = p->next)
			close(p->fd);

		execv(argl[0], argl);
		_exit (127);
	}

	/*
	 * Parent.
	 */

	close (child_end);
	SAFE_FREE(argl);

	/* Link into popen_chain. */
	entry->next = popen_chain;
	popen_chain = entry;
	entry->fd = parent_end;

	return entry->fd;

err_exit:

	SAFE_FREE(entry);
	SAFE_FREE(argl);
	close(pipe_fds[0]);
	close(pipe_fds[1]);
	return -1;
}

/**************************************************************************
 Wrapper for pclose. Modified from the glibc sources.
****************************************************************************/

int smb_sys_pclose(int fd)
{
	int wstatus;
	popen_list **ptr = &popen_chain;
	popen_list *entry = NULL;
	pid_t wait_pid;
	int status = -1;

	/* Unlink from popen_chain. */
	for ( ; *ptr != NULL; ptr = &(*ptr)->next) {
		if ((*ptr)->fd == fd) {
			entry = *ptr;
			*ptr = (*ptr)->next;
			status = 0;
			break;
		}
	}

	if (status < 0 || close(entry->fd) < 0)
		return -1;

	/*
	 * As Samba is catching and eating child process
	 * exits we don't really care about the child exit
	 * code, a -1 with errno = ECHILD will do fine for us.
	 */

	do {
		wait_pid = smb_sys_waitpid (entry->child_pid, &wstatus, 0);
	} while (wait_pid == -1 && ERRNO == EINTR);

	SAFE_FREE(entry);

	if (wait_pid == -1)
		return -1;
	return wstatus;
}
#endif
