/*
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1998

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#ifdef DEBUG_UNICODE
static void bin_dump(char *buf, int size);
static void bin_dump(char *buf, int size)
{
        int     i;

        printf("buf = 0x%02x, size = %d\n", (int)buf, size);
        for ( i = 0; i < size; i++ ) {
                printf("%02x ", buf[i]);
        }
        printf("\n");
        return;
}
#endif

int PutUniCode(void *dst,const char *src, int dst_len, u_short *char_code, int flags)
{
#ifdef SMBCD_TXTENC_USE
	int  len=0;
        u_long  encflag = 0;
        char    *buf1 = 0;
        char    *buf2 = 0;
        void    *buf3 = 0;
        int     buf1siz = strlen(src) + 1;
        int     buf2siz = strlen(src)*4 + 1;
        int     buf3siz = strlen(src)*4 + 2;
        unsigned int endian = 0;
        int     retval = 0;

        buf1 = (char *)smb_malloc(buf1siz);
        buf2 = (char *)smb_malloc(buf2siz);
        buf3 = (char *)smb_malloc(buf3siz);
        strcpy(buf1, src);
        memset(buf2, '\0', buf2siz);
        memset(buf3, '\0', buf3siz);
#ifdef DEBUG_UNICODE
	printf("src, len = %d\n", strlen(buf1));
	bin_dump(buf1, strlen(buf1)+1);
#endif
        if (flags & STR_UPPER)
                strupper(buf1);
#ifdef SMBCD_TXTENC_USE
        retval = txtencConvertToOneSuitableIso(
                        *char_code,
                        buf1,
                        buf2,
                        buf2siz,
                        TXTENC_UTF_8,
                        &encflag);
#endif
        if(retval < 0) {
                printf("txtencConvertToOneSuitableIso(%d)\n", retval);
#ifndef SMBCD_USE
                smbc_mfprintf("PutUniCode: txtencConvertToOneSuitableIso(%d)\n", retval);
#endif
                SAFE_FREE(buf1);
                SAFE_FREE(buf2);
                SAFE_FREE(buf3);
                return 0;
        }
#ifdef DEBUG_UNICODE
	printf("sjis->utf8, len = %d\n", strlen(buf2));
	bin_dump(buf2, strlen(buf2)+1);
#endif

        endian = TXTENC_UTF_16LE;
#ifdef SMBCD_TXTENC_USE
        retval = txtencConvertUtf8toUtf16(buf2, buf3, buf3siz, &endian);
#endif
        if(retval < 0) {
                printf("txtencConvertUtf8toUtf16(%d)\n", retval);
#ifndef SMBCD_USE
                smbc_mfprintf("PutUniCode: txtencConvertUtf8toUtf16(%d)\n", retval);
#endif
                SAFE_FREE(buf1);
                SAFE_FREE(buf2);
                SAFE_FREE(buf3);
                return 0;
        }

        len += (strlen_w(buf3)*2) + 2;
        memcpy(dst, buf3, MIN(dst_len, strlen_w(buf3)*2+2));

#ifdef DEBUG_UNICODE
	printf("utf8->utf16, len = %d\n", len);
	bin_dump(buf3, len+1);
#endif

        SAFE_FREE(buf1);
        SAFE_FREE(buf2);
        SAFE_FREE(buf3);
	return len;
#else 
	int  len=0;
        int     i=0;
        int     j=0;

        bzero(dst , dst_len );
        for( i =0,j=0 ; i<strlen(src); i++)
          {
            memcpy( dst+j , src+i , 1 );
            j+=2;
          }
        len += j;
        len += 2;

	return len;
#endif/* TXTENC NO USE */
}

int PullUniCode(char *dst,const void *src, int dst_len, int src_len, u_short *char_code)
{
#ifdef SMBCD_TXTENC_USE
	int len = 0;
        void	*buf1 = 0;
        u_char  *buf2 = 0;
        int     buf_siz = 0;
        unsigned int endian = 0;
	u_char	new_char_code = (u_char)*char_code;
	u_char  candidates[] = {
				CHAR_CODE_NONE, 
				CHAR_CODE_SJIS,
				CHAR_CODE_EUSTD,
				CHAR_CODE_CE,
				CHAR_CODE_GREEK,
				CHAR_CODE_CYR,
				CHAR_CODE_TUR,
				CHAR_CODE_BALTIC,
				CHAR_CODE_GB,
				CHAR_CODE_BIG,
				CHAR_CODE_HEBREW,
				CHAR_CODE_ARABIC,
				CHAR_CODE_THAI,
				CHAR_CODE_KS
				};
	size_t	n_cand = sizeof(candidates)/sizeof(u_char);

	if(src_len >= 2) {
		buf_siz = src_len*2 + 2;
		buf1 = (u_char *)smb_malloc(buf_siz);
		buf2 = (u_char *)smb_malloc(buf_siz);
		memset(buf1, '\0', buf_siz);
		memcpy(buf1, src, src_len);
		memset(buf2, '\0', buf_siz);
		endian = TXTENC_UTF_16LE;
#ifdef SMBCD_TXTENC_USE
		len = txtencConvertUtf16toUtf8(buf1, buf2, buf_siz, &endian);
#endif
		if(len < 0) {
#ifndef SMBCD_USE
                	smbc_mfprintf("PullUniCode: txtencConvertUtf16toUtf8(%d)\n", len);
#endif
			SAFE_FREE(buf1);
			SAFE_FREE(buf2);
			return len;
		}

		if(new_char_code == CHAR_CODE_NONE) {

#ifdef SMBCD_TXTENC_USE
			txtencPickBestCharcode(TXTENC_UTF_8, n_cand, candidates, buf2, &new_char_code);
#endif
		}

		memset(dst, '\0', dst_len);
#ifdef SMBCD_TXTENC_USE
		len = txtencConvertInbound(
				TXTENC_UTF_8,
				buf2,
				new_char_code,
				dst,
				dst_len);
#endif
		SAFE_FREE(buf1);
		SAFE_FREE(buf2);
	}
	SSVAL(dst,len,0);
	len += 2;

	*char_code = (u_short)new_char_code;

	return len;
#else 
        int len;
        int     i=0;
        int     j=0;

        bzero( dst , dst_len);
        for( i =0,j=0 ; i<src_len; j++)
          {
            memcpy( dst+j , src+i , 1 );
            i+=2;
          }
        if( len == -1 )
          {
            len=0;
          }
        len = strlen(dst);
        len += 2;


	return len;

#endif/* TXTENC NO USE */

}


/*******************************************************************
 Count the number of characters in a smb_ucs2_t string.
********************************************************************/

size_t strlen_w(const smb_ucs2_t *src)
{
  size_t len;

  for(len = 0; *src++; len++)
    ;

  return len;
}


/*******************************************************************
 Compare the two strings s1 and s2.
********************************************************************/

int strcmp_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	smb_ucs2_t c1, c2;

	for (;;) {
		c1 = *s1++;
		c2 = *s2++;

		if (c1 != c2)
			return c1 - c2;

		if (c1 == 0)
			break;
	}
	return 0;
}

/*******************************************************************
 Compare the first n characters of s1 to s2. len is in ucs2 units.
********************************************************************/

int strncmp_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2, size_t len)
{
	smb_ucs2_t c1, c2;

	for (; len != 0; --len) {
		c1 = *s1++;
		c2 = *s2++;

		if (c1 != c2)
			return c1 - c2;

		if (c1 == 0)
			break;

	}
	return 0;
}

/*******************************************************************
 Search string s2 from s1.
********************************************************************/

smb_ucs2_t *strstr_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	size_t len = strlen_w(s2);
	smb_ucs2_t *s0 = NULL;

	if (!*s2) {
		(const smb_ucs2_t *)s0 = s1;
		return s0;
	}

	for(;*s1; s1++) {
		if (*s1 == *s2) {
			if (strncmp_w(s1, s2, len) == 0) {
				(const smb_ucs2_t *)s0 = s1;
				return s0;
			}
		}
	}
	return NULL; 
}

/*******************************************************************
 Search for ucs2 char c from the beginning of s.
********************************************************************/ 

smb_ucs2_t *strchr_w(const smb_ucs2_t *s, smb_ucs2_t c)
{
	smb_ucs2_t *s0 = NULL;

	do {
		if (*s == c) {
			(const smb_ucs2_t *)s0 = s;
			return s0;
		}
	} while (*s++);

	return NULL;
}

/*******************************************************************
 Search for ucs2 char c from the end of s.
********************************************************************/ 

smb_ucs2_t *strrchr_w(const smb_ucs2_t *s, smb_ucs2_t c)
{
	smb_ucs2_t *retval = 0;

	do {
		if (*s == c)
			(const smb_ucs2_t *)retval = s;
	} while (*s++);

	return retval;
}

/*******************************************************************
 Search token from s1 separated by any ucs2 char of s2.
********************************************************************/

smb_ucs2_t *strtok_w(smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	static smb_ucs2_t *s = NULL;
	smb_ucs2_t *q;

	if (!s1) {
		if (!s)
			return NULL;
		s1 = s;
	}

	for (q = s1; *s1; s1++) {
		smb_ucs2_t *p = strchr_w(s2, *s1);
		if (p) {
			if (s1 != q) {
				s = s1 + 1;
				*s1 = '\0';
				return q;
			}
			q = s1 + 1;
		}
	}

	s = NULL;
	if (*q)
		return q;

	return NULL;
}

/*******************************************************************
 Mapping tables for UNICODE character. Allows toupper/tolower and
 isXXX functions to work.

 tridge: split into 2 pieces. This saves us 5/6 of the memory
 with a small speed penalty
 The magic constants are the lower/upper range of the tables two
 parts
********************************************************************/

typedef struct {
	smb_ucs2_t lower;
	smb_ucs2_t upper;
	unsigned char flags;
} smb_unicode_table_t;

#define TABLE1_BOUNDARY 9450
#define TABLE2_BOUNDARY 64256

static smb_ucs2_t *last_ptr = NULL;

void set_first_token_w(smb_ucs2_t *ptr)
{
	last_ptr = ptr;
}

/****************************************************************************
 Get the next token from a string, return False if none found
 handles double-quotes. 
 Based on a routine by GJC@VILLAGE.COM. 
 Extensively modified by Andrew.Tridgell@anu.edu.au
 bufsize is in bytes.
****************************************************************************/

static smb_ucs2_t sep_list[] = { (smb_ucs2_t)' ', (smb_ucs2_t)'\t',  (smb_ucs2_t)'\n',  (smb_ucs2_t)'\r', 0};
static smb_ucs2_t quotechar = (smb_ucs2_t)'\"';

BOOL next_token_w(smb_ucs2_t **ptr, smb_ucs2_t *buff, smb_ucs2_t *sep, size_t bufsize)
{
	smb_ucs2_t *s;
	BOOL quoted;
	size_t len=1;

	/*
	 * Convert bufsize to smb_ucs2_t units.
	 */

	bufsize /= sizeof(smb_ucs2_t);

	if (!ptr)
		ptr = &last_ptr;
	if (!ptr)
		return(False);

	s = *ptr;

	/*
	 * Default to simple separators.
	 */

	if (!sep)
		sep = sep_list;

	/*
	 * Find the first non sep char.
	 */

	while(*s && strchr_w(sep,*s))
		s++;

	/*
	 * Nothing left ?
	 */

	if (!*s)
		return(False);

	/*
	 * Copy over the token.
	 */

	for (quoted = False; len < bufsize && *s && (quoted || !strchr_w(sep,*s)); s++) {
		if (*s == quotechar) {
			quoted = !quoted;
		} else {
			len++;
			*buff++ = *s;
		}
	}

	*ptr = (*s) ? s+1 : s;  
	*buff = 0;
	last_ptr = *ptr;

	return(True);
}

/****************************************************************************
 Convert list of tokens to array; dependent on above routine.
 Uses last_ptr from above - bit of a hack.
****************************************************************************/

smb_ucs2_t **toktocliplist_w(int *ctok, smb_ucs2_t *sep)
{
	smb_ucs2_t *s=last_ptr;
	int ictok=0;
	smb_ucs2_t **ret, **iret;

	if (!sep)
		sep = sep_list;

	while(*s && strchr_w(sep,*s))
		s++;

	/*
	 * Nothing left ?
	 */

	if (!*s)
		return(NULL);

	do {
		ictok++;
		while(*s && (!strchr_w(sep,*s)))
			s++;
		while(*s && strchr_w(sep,*s))
			*s++=0;
	} while(*s);

	*ctok = ictok;
	s = last_ptr;

	if (!(ret=iret=(smb_ucs2_t **)smb_malloc(ictok*sizeof(smb_ucs2_t *))))
		return NULL;
  
	while(ictok--) {
		*iret++=s;
		while(*s++)
			;
		while(!*s)
			s++;
	}

	return ret;
}

/*******************************************************************
 Compare 2 strings (case sensitive).
********************************************************************/

BOOL strcsequal_w(const smb_ucs2_t *s1,const smb_ucs2_t *s2)
{
	if (s1 == s2)
		return(True);
	if (!s1 || !s2)
		return(False);
  
	return(strcmp_w(s1,s2)==0);
}


/****************************************************************************
 String replace.
****************************************************************************/

void string_replace_w(smb_ucs2_t *s, smb_ucs2_t oldc, smb_ucs2_t newc)
{
	while (*s) {
		if (oldc == *s)
			*s = newc;
		s++;
	}
}

/*******************************************************************
 Skip past some strings in a buffer. n is in bytes.
********************************************************************/

smb_ucs2_t *skip_string_w(smb_ucs2_t *buf,size_t n)
{
	while (n--)
		buf += (strlen_w(buf)*sizeof(smb_ucs2_t)) + 1;
	return(buf);
}

/*******************************************************************
 Count the number of characters in a string. Same as strlen_w in
 smb_ucs2_t string units.
********************************************************************/

size_t str_charnum_w(const smb_ucs2_t *s)
{
	return strlen_w(s);
}

/*******************************************************************
 Trim the specified elements off the front and back of a string.
********************************************************************/

BOOL trim_string_w(smb_ucs2_t *s,const smb_ucs2_t *front,const smb_ucs2_t *back)
{
	BOOL ret = False;
	size_t front_len = (front && *front) ? strlen_w(front) : 0;
	size_t back_len = (back && *back) ? strlen_w(back) : 0;
	size_t s_len;

	while (front_len && strncmp_w(s, front, front_len) == 0) {
		smb_ucs2_t *p = s;
		ret = True;

		while (1) {
			if (!(*p = p[front_len]))
				break;
			p++;
		}
	}

	if(back_len) {
		s_len = strlen_w(s);
		while ((s_len >= back_len) && 
			(strncmp_w(s + s_len - back_len, back, back_len)==0)) {
			ret = True;
			s[s_len - back_len] = 0;
			s_len = strlen_w(s);
		}
	}

	return(ret);
}

/****************************************************************************
 Find the number of 'c' chars in a string.
****************************************************************************/

size_t count_chars_w(const smb_ucs2_t *s,smb_ucs2_t c)
{
	size_t count=0;

	while (*s) {
		if (*s == c)
			count++;
		s++;
	}
	return(count);
}

/*******************************************************************
 Return True if a string consists only of one particular character.
********************************************************************/

BOOL str_is_all_w(const smb_ucs2_t *s,smb_ucs2_t c)
{
	if(s == NULL)
		return False;
	if(!*s)
		return False;

	while (*s) {
		if (*s != c)
			return False;
		s++;
	}
	return True;
}

/****************************************************************************
 Like strncpy but always null terminates. Make sure there is room !
 The variable n should always be one less than the available size and is in
 ucs2 units.
****************************************************************************/

smb_ucs2_t *StrnCpy_w(smb_ucs2_t *dest,const smb_ucs2_t *src,size_t n)
{
	smb_ucs2_t *d = dest;
	if (!dest)
		return(NULL);
	if (!src) {
		*dest = 0;
		return(dest);
	}

	while (n-- && (*d++ = *src++))
		;
	*d = 0;
	return(dest);
}


