#include <pthread.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <smbc_prc_common.h>
#include <smbc_prc_local.h>
#ifdef SMBCD_MFPRINT_USE
#include <ncs_mmesgkey.h>
#include <smbc_prc_mfprintf.h>
#endif

#ifdef FIXEDLINK
extern void _pthread_init __P((void));
#endif


static int               smbc_chld_process_cnt = 0;


int main(int argc, char**argv)
{
///////////////////////////////////////////////////////////
  int                 fromlen = 0;
  int                 newfd = -1;
  int                 fd = -1;
  int                 nfound = 0;
  int                 chilpid = 0 ;
  struct sockaddr_un  s_un ;
  struct stat         sb ;
  fd_set              fds ;
///////////////////////////////////////////////////////////

#ifdef FIXEDLINK
        _pthread_init();
#endif

  signal( SIGPIPE , SIG_IGN);
  signal( SIGCHLD , smbc_prc_sig_chld );

#ifdef SMBCD_MFPRINT_USE
  MPRINTF_CREATE();
  MPRINTF(("[smbc_process]main_srv:: smbc_process  START!\n"));
#endif

  if ( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ) {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif
#ifdef SMBCD_MFPRINT_USE 
    MPRINTF(("[smbc_process]main_srv:: socket ERR\n"));
#endif
    printf("[smbc_process]main_srv:: creating socket failed  error[%d]\n",errorcode);
    goto ERR_RET;
  }

  memset(&s_un, 0, sizeof(s_un));
  memset(&sb, 0, sizeof(stat));


  s_un.sun_family = AF_UNIX;

  if((strlen(SMBC_PRC_SOCK_PATH)+strlen(SMBC_PRC_SOCK_SID_NAME))  > sizeof(s_un.sun_path) )
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]main_srv:: PATH_NAME length ERR\n"));
#endif
      printf("main_srv::path_length error!\n");
      goto ERR_RET;
    }
  if ( stat( SMBC_PRC_SOCK_PATH, &sb ) == -1 ) {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif

#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]main_srv:: stat ERR\n"));
#endif
      printf("[smbc_process]main_srv:: stat error!  errorno[%d]\n",errorcode);
      goto ERR_RET;
  } else {
    strncpy(s_un.sun_path, SMBC_PRC_SOCK_PATH ,strlen(SMBC_PRC_SOCK_PATH));
  }

  strncat(s_un.sun_path, SMBC_PRC_SOCK_SID_NAME , strlen(SMBC_PRC_SOCK_SID_NAME) );


  unlink(s_un.sun_path);
  if ( bind(fd, (struct sockaddr *)&s_un, sizeof(s_un)) < 0 ) {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif
#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]main_srv:: bind ERR\n"));
#endif
    printf("[smbc_process]main_srv:: binding socket failed   errorno[%d]\n",errorcode);
    goto ERR_RET;
  }

  if ( listen(fd, 5) < 0 ) {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif
#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]main_srv:: listen ERR\n"));
#endif
    printf("[smbc_process]main_srv:: listening socket failed   errorno[%d]\n",errorcode);
    goto ERR_RET;
  }

  printf("[smbcd]:: smbcd   PROCESS READY \n");
  while (1)
  {
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    nfound = select(fd+1, &fds, NULL, NULL, NULL);
    if ( nfound <= 0) 
    {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif
      if (errorcode == EINTR) 
      {
         continue;
      }
       else 
      {
#ifdef SMBCD_MFPRINT_USE
         MPRINTF(("[smbc_process]main_srv:: select ERR\n"));
#endif
         printf("[smbc_process]main_srv::select ERR errorno[%d] \n",errorcode);
         break;
      }
    }

accept_retry:
    fromlen = sizeof(s_un);
    newfd = accept(fd, (struct sockaddr *)&s_un, &fromlen);
    if ( newfd < 0 ) {
#ifdef FIXEDLINK
      int errorcode = *__errno();
#else
      int errorcode = errno;
#endif

      if ( errorcode == EINVAL ) 
      {
        goto accept_retry;
      } else if ( errorcode == ECONNABORTED )
      {
	continue;
      }
      continue;
    } else if ( fromlen == 0 ) 
    {
      close(newfd);
      continue;
    }


    smbc_chld_process_cnt++;


    if ( (chilpid = fork()) == 0)
    { 
      close(fd);
      NcsSmbPrcHandler(newfd,smbc_chld_process_cnt);   
      close(newfd);
      exit(0);      
    }
    close(newfd);
  }


ERR_RET:


  if ( fd >= 0 ) 
  {
      close(fd);
  }

  exit(0);

}

void smbc_prc_sig_chld(int signo )
{
////////////////////////////
  pid_t     pid;
  int       stat;
////////////////////////////

   while ( (pid = wait3((int *)&stat, WNOHANG, 0)) > 0) {
     if ( stat == NULL ) {
       //       printf( "[smbc_process] sigchild(pid:[%d], stat:[%d]) ", pid, stat );       
     }
     else {
       printf( "[smbc_process]smbc_prc_sig_chld sigchild(pid:[%d], stat:[%d]) ", pid, stat );
     }
   smbc_chld_process_cnt--;
   }

}

