#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <smbc_prc_common.h>
#include <smbc_prc_local.h>

#ifdef SMBCD_MFPRINT_USE
#include <ncs_mmesgkey.h>
#include <smbc_prc_mfprintf.h>
#endif


wk_smb_msg_t wk_smb_func_cvrt[] = {
  {SMBC_MSG_OPEN           , smbc_param_an_OPEN,         },
  {SMBC_MSG_CLOSE          , smbc_param_an_CLOSE,        },
  {SMBC_MSG_CONNECT        , smbc_param_an_CONNECT,      },
  {SMBC_MSG_DISCONNECT     , smbc_param_an_DISCONNECT,   },
  {SMBC_MSG_OPNFILE        , smbc_param_an_OPNFILE       },
  {SMBC_MSG_CLSFILE        , smbc_param_an_CLSFILE       },
  {SMBC_MSG_READFILE       , smbc_param_an_READFILE      },
  {SMBC_MSG_WRTFILE        , smbc_param_an_WRTFILE       },         
  {SMBC_MSG_DELFILE        , smbc_param_an_DELFILE       },         
  {SMBC_MSG_MKDIR          , smbc_param_an_MKDIR         },       
  {SMBC_MSG_DELDIR         , smbc_param_an_DELDIR        },              
  {SMBC_MSG_CHKDIR         , smbc_param_an_CHKDIR        },             
  {SMBC_MSG_GET_GRPL       , smbc_param_an_GET_GRPL      },               
  {SMBC_MSG_GET_SRVL       , smbc_param_an_GET_SRVL      },            
  {SMBC_MSG_GET_SHRL       , smbc_param_an_GET_SHRL      },             
  {SMBC_MSG_GET_DIRL       , smbc_param_an_GET_DIRL      },          
  {SMBC_MSG_GET_SRVCAP     , smbc_param_an_GET_SRVCAP    },              
#ifdef NCS_SMB_NTAUTH_SUPPORT
  {SMBC_MSG_USR_LIN        , smbc_param_an_USR_LOGIN     },
  {SMBC_MSG_USR_LOUT       , smbc_param_an_USR_LOGOUT    },
  {SMBC_MSG_GET_DCINF      , smbc_param_an_GET_DCInf     },
  {SMBC_MSG_USR_GETGRP     , smbc_param_an_USR_GET_Groups},
#endif /* NCS_SMB_NTAUTH_SUPPORT */
  {SMBC_MSG_SET_DF_CHRC    , smbc_param_an_DF_CHR_CODE   },
  {NULL                    , NULL,                       },
};





void NcsSmbPrcHandler(int smbc_srv_fd ,int smbc_chld_process_cnt )
{

/////////////////////////////////////////////////
smb_ctx        *ctx = 0;
int             ret = 0;
char            msg_type = 0;
int             (*func)() = NULL;
int             i = 0;
int             wk_pos = 0;
u_short         rv_total_len = 0;
u_short         sn_total_len = 0;
int             offset = SMB_MSG_DF_START_POS;
u_char          check_msg_ID = 0;
 u_char          *smbc_rv_buf;
 u_char          *smbc_sn_buf;
/////////////////////////////////////////////////

#ifdef SMBCD_MFPRINT_USE
  MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] START !!  \n",(int)getpid() ));
#endif
    smbc_rv_buf = malloc(SMB_MAX_BUF_SIZE);
    smbc_sn_buf = malloc(SMB_MAX_BUF_SIZE);

    if( (smbc_rv_buf == NULL)||(smbc_sn_buf==NULL))
      {
#ifdef SMBCD_MFPRINT_USE
	MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] BUF_MALLOC ERR!! \n",(int)getpid() ));
#endif
       goto FREE_RET;
       
      }


  while(1)
  {

    memset(  smbc_rv_buf, 0 , SMB_MAX_BUF_SIZE);
    memset(  smbc_sn_buf, 0 , SMB_MAX_BUF_SIZE);

    rv_total_len = smbc_prc_recv_buf( smbc_rv_buf , smbc_srv_fd ); 

    if( rv_total_len == SMBC_MSG_RCV_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] SMBC_MSG_RCV_ERR \n",(int)getpid()));
#endif
       goto FREE_RET;
    }

    memcpy(&check_msg_ID, smbc_rv_buf+SMB_MSG_TOTALLEN_LEN , SMB_MSG_MSG_ID_LEN  );


    if( smbc_chld_process_cnt > SMBC_MAX_CHLD_NUM)
    {

#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] Session Count OVER \n",(int)getpid()));
#endif
        printf("NcsSmbPrcHandler:: Session Count OVER !!!\n");

        memcpy(smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, &check_msg_ID , SMB_MSG_MSG_ID_LEN );

        ret = smbc_prc_cnv_err(SMBERR_IF, SMBERR_TEMPORARY, SMBERR_IFERR); 
        if( ret < 0 )
        {
            goto FREE_RET;
        }

        offset = SMB_MSG_DF_START_POS;
        offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(ret) ,    
                                (char*)&ret ,
                                 offset );
        if(offset == SMBC_PARAM_SET_ERR)
        {
#ifdef SMBCD_MFPRINT_USE
            MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] R_CODE SET_ERR-1  !!  \n",(int)getpid()));
#endif
            goto FREE_RET;
        }

        sn_total_len = (u_short)offset;
        memcpy(smbc_sn_buf, &sn_total_len , SMB_MSG_TOTALLEN_LEN  );

        ret = smbc_prc_send_buf( smbc_sn_buf , sn_total_len , smbc_srv_fd);
        if (ret == SMBC_MSG_SND_ERR  )
        {
#ifdef SMBCD_MFPRINT_USE
           MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] message send ERR-1 !! \n",(int)getpid()));
#endif
           printf("NcsSmbPrcHandler:: send NG !!!\n");
           goto FREE_RET;
        }
        break;
    }

    wk_pos = smbc_prc_get_param(rv_total_len , SMB_PARAM_MSG_TYPE ,smbc_rv_buf ,(char*)&msg_type );
    if( wk_pos == SMBC_PARAM_SERCH_ERR )
    {
#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] msg_type could not found !!  \n",(int)getpid()));
#endif
        goto MSG_ERR;
    }

    for (i=0; wk_smb_func_cvrt[i].type != NULL ; i++)
    {
        if(wk_smb_func_cvrt[i].type == msg_type)
        {
            func = wk_smb_func_cvrt[i].func ;
            break;
	}
    }
    if(func == NULL){
#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] msg_type  ERR!! msg_type->%d  \n",(int)getpid(), msg_type));
#endif
        goto MSG_ERR;
    }
#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] MSG_TYPE->[%d] START \n",(int)getpid(),wk_smb_func_cvrt[i].type));
#endif
    switch(wk_smb_func_cvrt[i].type)
    {
      case SMBC_MSG_OPEN:

           ret = (*func)(smbc_rv_buf , smbc_sn_buf , &ctx);
           break;

      case SMBC_MSG_READFILE:

           ret = (*func)(smbc_rv_buf , smbc_sn_buf , ctx ,smbc_srv_fd , &check_msg_ID );
           if(ret == SMBC_PARAM_AN_SUCCESS)
           {
             continue;
           }
#ifdef SMBCD_MFPRINT_USE
  MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] SMBC_MSG_READFILE ERR !! \n",(int)getpid()));
#endif
           goto FREE_RET;

      case SMBC_MSG_WRTFILE:

           ret = (*func)(smbc_rv_buf , smbc_sn_buf , ctx ,smbc_srv_fd , &check_msg_ID );
           break;

      default:

           ret = (*func)(smbc_rv_buf , smbc_sn_buf , ctx);
           break;

    }

    if( ret != SMBC_PARAM_AN_SUCCESS )
    { 
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] PARAM_AN  ERR!! msg_type->%d  \n",(int)getpid(),wk_smb_func_cvrt[i].type));
#endif
       goto MSG_ERR;
    }

    memcpy( smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, &check_msg_ID , SMB_MSG_MSG_ID_LEN  );
    memcpy( &sn_total_len, smbc_sn_buf , SMB_MSG_TOTALLEN_LEN  );
    ret = smbc_prc_send_buf( smbc_sn_buf , sn_total_len , smbc_srv_fd);
    if( ret == SMBC_MSG_SND_ERR )
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] message send ERR-2 !! \n",(int)getpid() ));
#endif
      goto FREE_RET;
    }

    if( wk_smb_func_cvrt[i].type == SMBC_MSG_CLOSE )
      {
#ifdef SMBCD_MFPRINT_USE
	MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] SMBC_MSG_CLOSE END \n",(int)getpid()));
#endif
        goto FREE_RET;
      }
#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] MSG_END  MSG_TYPE ->[%d]  \n",(int)getpid(),wk_smb_func_cvrt[i].type));
#endif
    continue;
  }


MSG_ERR:


  memset(  smbc_sn_buf, 0 , SMB_MAX_BUF_SIZE);
  memcpy( &check_msg_ID, smbc_rv_buf+SMB_MSG_TOTALLEN_LEN , SMB_MSG_MSG_ID_LEN  );
  memcpy(smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, &check_msg_ID , SMB_MSG_MSG_ID_LEN  );

  ret = smbc_prc_cnv_err(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR); 
  if( ret < 0 )
  {
#ifdef SMBCD_MFPRINT_USE
  MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] FREE_RET !! \n",(int)getpid() ));
#endif
      goto FREE_RET;
  }

  offset = SMB_MSG_DF_START_POS;
  offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(ret) ,    
                            (char*)&ret ,
                             offset );
  if(offset == SMBC_PARAM_SET_ERR)
  {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] R_CODE SET_ERR-2  !!  \n",(int)getpid() ));
#endif
      goto FREE_RET;
  }

  sn_total_len = (u_short)offset;
  memcpy(smbc_sn_buf, &sn_total_len , SMB_MSG_TOTALLEN_LEN );


  ret = smbc_prc_send_buf( smbc_sn_buf , sn_total_len , smbc_srv_fd);
  if (ret == SMBC_MSG_SND_ERR  )
  {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]NcsSmbPrcHandler::pid[%d] message send ERR-3 !! \n",(int)getpid() ));
#endif
     printf("NcsSmbPrcHandler:: send NG !!!\n");
     goto FREE_RET;
  }


FREE_RET:


  if(smbc_sn_buf != NULL)
    {
      free(smbc_sn_buf);
    }
  if(smbc_rv_buf != NULL)
    {
      free(smbc_rv_buf);
    }
  exit(0);

}


int smbc_param_an_OPEN(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf ,smb_ctx** ctx)
{
///////////////////////////////////////////////////////////////////////////////////
  int             smb_ret = 0;
  int             offset = SMB_MSG_DF_START_POS;  
  u_short         total_len = 0;
  smb_ctx*        wk_ctx=0;
///////////////////////////////////////////////////////////////////////////////////


  smb_ret = NcsSmbOpen(&wk_ctx);



  offset = smbc_prc_set_param( smbc_sn_buf ,
                               SMB_PARAM_R_CODE ,
                               sizeof(smb_ret) ,    
                               (char*)&smb_ret ,
                               offset );

  if(offset == SMBC_PARAM_SET_ERR)
  {
#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]smbc_param_an_OPEN:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
    goto ERR_RET;
  }


  total_len = (u_short)offset;
  memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

  *ctx = wk_ctx ;

  return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

  *ctx = NULL;

  return(SMBC_PARAM_AN_ERR);

}



int smbc_param_an_CLOSE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf , smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  int           smb_ret = 0;
  int           offset = SMB_MSG_DF_START_POS;  
  u_short       total_len = 0;
//////////////////////////////////////////////////////////////////////////////////////


  smb_ret = NcsSmbClose(ctx);


  offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,    
                            (char*)&smb_ret ,
                            offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_CLOSE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }
  
   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


  return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  return(SMBC_PARAM_AN_ERR);

}



int smbc_param_an_CONNECT(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  smb_vc              vc;
  int                 offset = SMB_MSG_DF_START_POS;
  int                 smb_ret = 0;
  int                 pos = 0 ;
  u_short             total_len = 0;
//////////////////////////////////////////////////////////////////////////////////////


   bzero(&vc, sizeof(vc));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID ,smbc_rv_buf,(char*)&vc.character );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_GRP_NAME ,smbc_rv_buf ,(char*)vc.group );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SRV_NAME , smbc_rv_buf ,(char*)vc.server );   

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SHRE , smbc_rv_buf ,(char*)vc.share);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf,(char*)vc.username );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD ,smbc_rv_buf ,(char*)vc.password);

   smb_ret = NcsSmbConnect(ctx, &vc, 0);


   offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,    
                            (char*)&smb_ret ,
                            offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_CONNECT:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }
  
   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

   return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

  return(SMBC_PARAM_AN_ERR);
  
}


int smbc_param_an_DISCONNECT(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf , smb_ctx* ctx)
{
/////////////////////////
  int           smb_ret = 0;
  int           offset = SMB_MSG_DF_START_POS;
  u_short       total_len = 0;
/////////////////////////

   smb_ret = NcsSmbDisconnect(ctx , 0 );


  offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,    
                            (char*)&smb_ret ,
                            offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_DISCONNECT:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }


    total_len = (u_short)offset;
    memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


    return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

    return(SMBC_PARAM_AN_ERR);


}


int smbc_param_an_OPNFILE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
///////////////////////////////////////////////////////////////////////////////////////
  smb_file         file;
  smb_finfo        finfo;
  int              offset=SMB_MSG_DF_START_POS;
  int              smb_ret = 0 ;
  u_short          total_len = 0;
  int              pos = 0;
/////////////////////////////////////////////////////////////////////

   bzero(&file,sizeof(smb_file));
   bzero(&finfo,sizeof(smb_finfo));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&file.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_NAME ,smbc_rv_buf ,(char*)&file.name);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_ACC ,smbc_rv_buf ,(char*)&file.access);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_ATTR , smbc_rv_buf ,(char*)&file.attribute);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_DISPOS ,smbc_rv_buf ,(char*)&file.disposition);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_FLAG ,smbc_rv_buf ,(char*)&file.flag);

   smb_ret = NcsSmbOpenFile(ctx,
                        &file,
                        &finfo,
                        0
                        );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,    
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }
  
   if(smb_ret != SMBSUCCESS)
   {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: smb_ret -> [0x%08x]  \n",smb_ret));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.character --> [%d] \n",file.character));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.name --> [%s] \n",file.name));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.access --> [%d] \n",(int)file.access));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.attribute --> [%d] \n",(int)file.attribute));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.disposition --> [%d] \n",(int)file.disposition));
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: file.flag --> [%d] \n",(int)file.flag));
#endif
     printf("NcsSmbOpenFile smb_ret -->0x%08x \n" , smb_ret);
     goto MSG_SND ;
   }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FID ,
                                sizeof(finfo.fid) ,
                                (char*)&finfo.fid ,
                                offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FID SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CRE_TM ,
                                sizeof(finfo.creation_time) ,
                                (char*)&finfo.creation_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_CRE_TM SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ACC_TM ,
                                sizeof(finfo.last_access_time) ,
                                (char*)&finfo.last_access_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_ACC_TM SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_WRT_TM ,
                                sizeof(finfo.last_write_time) ,
                                (char*)&finfo.last_write_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_WRT_TM SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CHG_TM ,
                                sizeof(finfo.change_time) ,
                                (char*)&finfo.change_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_CHG_TM SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ATTR ,
                                sizeof(finfo.attribute) ,
                                (char*)&finfo.attribute,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_ATTR SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_SIZE ,
                                sizeof(finfo.size) ,
                                (char*)&finfo.size ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_OPNFILE:: SMB_PARAM_FI_SIZE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


MSG_SND:


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);




ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}



int smbc_param_an_CLSFILE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////
  int          offset=SMB_MSG_DF_START_POS;
  int          smb_ret = 0 ;
  u_short      total_len = 0;
  int          pos = 0;
  u_short      fid = 0;
//////////////////////////////////////////////////////////////////


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_FID ,smbc_rv_buf ,(char*)&fid);

   smb_ret = NcsSmbCloseFile(
		              ctx,
                              fid,
                                0
                             );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_CLSFILE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

    return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}


int smbc_param_an_READFILE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx ,
                           int smbc_srv_fd ,u_char* check_msg_ID)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       fid = 0;
  u_long        in_offset = 0 ;
  int           size = 0;
  int           write_size = 0;
  u_char*       buf ;
  int           buf_offset = 0;
  int           write_offset = 0;
  int           ret = 0;
//////////////////////////////////////////////////////////////////


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_FID ,smbc_rv_buf ,(char*)&fid);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_OFFSET , smbc_rv_buf ,(char*)&in_offset );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SIZE , smbc_rv_buf ,(char*)&size);

   buf = malloc(size);
   if( buf == NULL )
   {
      goto ERR_RET;
   }
   bzero( buf , size );



   smb_ret = NcsSmbReadFile(
		             ctx,
		             fid,
		             in_offset,
		             buf,
		             &size,
                             0
                            );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_READFILE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_READFILE:: smb_ret -> [0x%08x] !! \n",smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_READFILE:: fid --> [%d] \n",fid));
      MPRINTF(("[smbc_process]smbc_param_an_READFILE:: in_offset --> [%d] \n",(int)in_offset));
      MPRINTF(("[smbc_process]smbc_param_an_READFILE:: size --> [%d] \n",size ));
#endif
      printf("NcsSmbReadFile smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND;
     }



    offset = smbc_prc_set_param( smbc_sn_buf ,
                             SMB_PARAM_SIZE ,
                             sizeof(size),
                             (char*)&size ,
                             offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_READFILE:: SMB_PARAM_SIZE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


    buf_offset = 0;
    write_offset = offset ;
    write_size = 0;
    while( buf_offset < size )
    {
      write_size = MIN(
                       SMB_MAX_BUF_SIZE-write_offset-SMB_PARAM_TAG_LEN-SMB_PARAM_LENGTH_LEN ,
                       size - buf_offset
                       );

      write_offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_BUF ,
                                write_size ,
                                (char*)buf+buf_offset ,
                                write_offset );

      if(write_offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
         MPRINTF(("[smbc_process]smbc_param_an_READFILE:: SMB_PARAM_BUF SET_ERR !! \n"));
#endif
         goto ERR_RET;
      }

      memcpy( smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, check_msg_ID , SMB_MSG_MSG_ID_LEN  );
      total_len = (u_short)write_offset;
      memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );
      ret = smbc_prc_send_buf( smbc_sn_buf , total_len , smbc_srv_fd);

      if (ret == SMBC_MSG_SND_ERR  )
      {
#ifdef SMBCD_MFPRINT_USE
         MPRINTF(("[smbc_process]smbc_param_an_READFILE::message send ERR !! \n"));
#endif
         printf("smbc_param_an_READFILE:: send NG !!!\n");
         goto ERR_RET;
      }
      write_offset = SMB_MSG_DF_START_POS;
      buf_offset += write_size ;
    }


    if(size == 0)
    {
      offset = smbc_prc_set_param( smbc_sn_buf ,
                             SMB_PARAM_BUF ,
                             size,
                             (char*)buf ,
                             offset );

      if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_READFILE:: SMB_PARAM_SIZE SET_ERR !! \n"));
#endif
       goto ERR_RET;
      }

      memcpy( smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, check_msg_ID , SMB_MSG_MSG_ID_LEN  );
      total_len = (u_short)offset;
      memcpy( smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN  );
      ret = smbc_prc_send_buf( smbc_sn_buf , total_len , smbc_srv_fd);
      if( ret == SMBC_MSG_SND_ERR )
      {
#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("[smbc_process]smbc_param_an_READFILE::message send ERR-3 !! \n"));
#endif
        goto ERR_RET;
      }
    }


    if(buf != NULL)
    {
        free(buf);
    }

    return(SMBC_PARAM_AN_SUCCESS);

MSG_SND:

    memcpy( smbc_sn_buf+SMB_MSG_TOTALLEN_LEN, check_msg_ID , SMB_MSG_MSG_ID_LEN  );
    total_len = (u_short)offset;
    memcpy( smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN  );
    ret = smbc_prc_send_buf( smbc_sn_buf , total_len , smbc_srv_fd);
    if( ret == SMBC_MSG_SND_ERR )
      {
#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("[smbc_process]smbc_param_an_READFILE::message send ERR-2 !! \n"));
#endif
        goto ERR_RET;
      }

    if(buf != NULL)
      {
        free(buf);
      }

    return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

    if(buf != NULL)
    {
        free(buf);
    }

  return(SMBC_PARAM_AN_ERR);


}


int smbc_param_an_WRTFILE(u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx ,
                          int smbc_srv_fd , u_char* check_msg_ID)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       fid = 0;
  u_long        in_offset = 0 ;
  u_char*        buf ;
  int           size = 0;
  u_long        mode = 0;
  u_short       length = 0;
  int           buf_offset = 0;
//////////////////////////////////////////////////////////////////


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_FID , smbc_rv_buf ,(char*)&fid);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_OFFSET , smbc_rv_buf ,(char*)&in_offset);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SIZE , smbc_rv_buf ,(char*)&size);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_MODE , smbc_rv_buf ,(char*)&mode);

   buf = malloc( size );
   if( buf == NULL )
   {
      goto ERR_RET;
   }
   bzero( buf , size );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_BUF , smbc_rv_buf ,(char*)buf);
   memcpy( &length , smbc_rv_buf+pos-SMB_PARAM_LENGTH_LEN , SMB_PARAM_LENGTH_LEN );

   if( size > length )
     {
     buf_offset=length;
     while( buf_offset < size )
       {
       memset(  smbc_rv_buf, 0 , SMB_MAX_BUF_SIZE);
       total_len = smbc_prc_buf_ID_PEEK(smbc_rv_buf, check_msg_ID, smbc_srv_fd );
       if( total_len == SMBC_MSG_RCV_ERR )
	 {
            goto ERR_RET;
	 }
       total_len = smbc_prc_recv_buf(smbc_rv_buf, smbc_srv_fd );
       if( total_len == SMBC_MSG_RCV_ERR )
         {
            goto ERR_RET;
         }
       pos = smbc_prc_get_param( total_len , (u_char)SMB_PARAM_BUF , smbc_rv_buf ,(char*)buf+buf_offset);
       length = 0;
       memcpy( &length , smbc_rv_buf+pos-SMB_PARAM_LENGTH_LEN , SMB_PARAM_LENGTH_LEN );
       buf_offset += length ;
       }
     }



   smb_ret = NcsSmbWriteFile(
		             ctx,
		             fid,
		             in_offset,
                             mode,
		             buf,
		             &size,
                             0
                            );



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: smb_ret  -> [0x%08x] !! \n",smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: fid --> [%d] \n",fid));
      MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: in_offset --> [%d] \n",(int)in_offset));
      MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: mode --> [%d] \n",(int)mode));
      MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: size --> [%d] \n",size ));
#endif
      printf("smbc_param_an_WRTFILE smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }


    offset = smbc_prc_set_param( smbc_sn_buf ,
                             SMB_PARAM_SIZE ,
                             sizeof(size),
                             (char*)&size ,
                             offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_WRTFILE:: SMB_PARAM_SIZE SET_ERR !! \n"));
#endif
     goto ERR_RET;
    }

MSG_SND:

    total_len = (u_short)offset;
    memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

    if( buf != NULL)
    {
      free( buf );
    }

    return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  if( buf != NULL)
  {
    free( buf );
  }

  return(SMBC_PARAM_AN_ERR);


}

int smbc_param_an_DELFILE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
/////////////////////////////////////////////////////////////////////
  smb_file      file;
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
/////////////////////////////////////////////////////////////////////

   bzero(&file ,sizeof(smb_file));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&file.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_NAME , smbc_rv_buf ,(char*)&file.name);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_FLAG , smbc_rv_buf ,(char*)&file.flag);


   smb_ret = NcsSmbDeleteFile(ctx,
                        &file,
                        0
                        );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_DELFILE:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


    return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}


int smbc_param_an_MKDIR(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
/////////////////////////////////////////////////////////////////////
  smb_file         file;
  smb_finfo        finfo;
  int              offset=SMB_MSG_DF_START_POS;
  int              smb_ret = 0 ;
  u_short          total_len = 0;
  int              pos = 0;
/////////////////////////////////////////////////////////////////////

   bzero(&file,sizeof(smb_file));
   bzero(&finfo,sizeof(smb_finfo));


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&file.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_NAME , smbc_rv_buf ,(char*)&file.name);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_ACC , smbc_rv_buf ,(char*)&file.access);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_ATTR , smbc_rv_buf ,(char*)&file.attribute);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_DISPOS , smbc_rv_buf ,(char*)&file.disposition);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_FLAG , smbc_rv_buf ,(char*)&file.flag);


   smb_ret = NcsSmbMakeDirectory(ctx,
                        &file,
                        &finfo,
                        0
                        );

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_R_CODE SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: smb_ret --> [0x%08x] !! \n",smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.character --> [%d]\n",file.character));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.name --> [%s] \n",file.name));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.access --> [%d] \n",(int)file.access));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.attribute --> [%d] \n",(int)file.attribute));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.disposition --> [%d] \n",(int)file.disposition));
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: file.flag --> [%d] \n",(int)file.flag));
#endif
      printf("NcsSmbMakeDirectory smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FID ,
                                sizeof(finfo.fid) ,
                                (char*)&finfo.fid ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FID SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CRE_TM ,
                                sizeof(finfo.creation_time) ,
                                (char*)&finfo.creation_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_CRE_TM SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }
   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ACC_TM ,
                                sizeof(finfo.last_access_time) ,
                                (char*)&finfo.last_access_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_ACC_TM  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_WRT_TM ,
                                sizeof(finfo.last_write_time) ,
                                (char*)&finfo.last_write_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_WRT_TM  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CHG_TM ,
                                sizeof(finfo.change_time) ,
                                (char*)&finfo.change_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_CHG_TM  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ATTR ,
                                sizeof(finfo.attribute) ,
                                (char*)&finfo.attribute,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_ATTR  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_SIZE ,
                                sizeof(finfo.size) ,
                                (char*)&finfo.size,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_MKDIR:: SMB_PARAM_FI_SIZE  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }

MSG_SND:


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}



int smbc_param_an_DELDIR(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf , smb_ctx* ctx)
{
/////////////////////////////////////////////////////////////////////
  smb_file        file;
  int             offset=SMB_MSG_DF_START_POS;
  int             smb_ret = 0 ;
  u_short         total_len = 0;
  int             pos = 0;
/////////////////////////////////////////////////////////////////////

   bzero(&file,sizeof(smb_file));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&file.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_NAME , smbc_rv_buf ,(char*)&file.name);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_FLAG , smbc_rv_buf ,(char*)&file.flag);


   smb_ret = NcsSmbDeleteDirectory(ctx,
                        &file,
                        0
                        );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_DELDIR:: SMB_PARAM_R_CODE  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


    return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}



int smbc_param_an_CHKDIR(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
/////////////////////////////////////////////////////////////////////
  smb_file          file;
  int               offset=SMB_MSG_DF_START_POS;
  int               smb_ret = 0 ;
  u_short           total_len = 0;
  int               pos = 0;
/////////////////////////////////////////////////////////////////////

   bzero(&file,sizeof(smb_file));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&file.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_F_NAME , smbc_rv_buf ,(char*)&file.name);


   smb_ret = NcsSmbCheckDirectory(ctx,
                        &file,
                        0
                        );



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_CHKDIR:: SMB_PARAM_R_CODE  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }




   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


    return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:


  return(SMBC_PARAM_AN_ERR);


}



int smbc_param_an_GET_GRPL(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{

//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       max_list_num = 0;
  smb_gp_list    group_list;
  smb_gp_info*   group_info = 0;
  int           i;
//////////////////////////////////////////////////////////////////


   bzero(&group_list , sizeof(smb_gp_list) );

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );

   pos = smbc_prc_get_param( total_len , SMB_PARAM_MAX_LST_N , smbc_rv_buf ,(char*)&max_list_num);

   group_info = malloc(sizeof(smb_gp_info)*max_list_num);
   if( group_info == NULL )
   {
      goto ERR_RET;
   }
   bzero( group_info , sizeof(smb_gp_info)*max_list_num );

   group_list.group_info = (smb_gp_info *)group_info;


   smb_ret = NcsSmbGetGroupList(
		                 ctx,
		                 max_list_num,
		                 &group_list,
		                 0 );

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL:: SMB_PARAM_R_CODE  SET_ERR !! \n"));
#endif
      goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL:: smb_ret --> [0x%08x]  \n",smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL:: max_list_num --> [%d]\n",max_list_num));
#endif
      printf("NcsSmbGetGroupList smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_ENT ,
                                sizeof(group_list.entries) ,
                                (char*)&group_list.entries,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_ENT  SET_ERR \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_TOTAL ,
                                sizeof(group_list.totalnum) ,
                                (char*)&group_list.totalnum,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_TOTAL  SET_ERR \n"));
#endif
     goto ERR_RET;

    }



  for(i = 0; i < group_list.entries; i++) 
  {

     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_CHAR_ID ,
                                  sizeof(group_info[i].character) ,
                                  (char*)&group_info[i].character ,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_CHAR_ID  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }

     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_GL_GP_NAME ,
                                  strlen(group_info[i].gp_name) ,
                                  group_info[i].gp_name,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_GL_GP_NAME  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }

     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_GL_GP_V_MJ ,
                                  sizeof(group_info[i].gp_version_major) ,
                                  (char*)&group_info[i].gp_version_major ,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_GL_GP_V_MJ  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }

     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_GL_GP_V_MN ,
                                  sizeof(group_info[i].gp_version_minor) ,
                                  (char*)&group_info[i].gp_version_minor ,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_GL_GP_V_MN  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }


     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_GL_GP_TYPE ,
                                  sizeof(group_info[i].gp_type) ,
                                  (char*)&group_info[i].gp_type,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_GL_GP_TYPE  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }

     offset = smbc_prc_set_param( smbc_sn_buf ,
                                  SMB_PARAM_GL_GP_MSTBR ,
                                  strlen(group_info[i].gp_master_browser) ,
                                  group_info[i].gp_master_browser ,
                                  offset );
    if(offset == SMBC_PARAM_SET_ERR)
      {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_GRPL::SMB_PARAM_GL_GP_MSTBR  SET_ERR  i=%d \n",i));
#endif
       goto ERR_RET;
      }
  }

MSG_SND:

   if( group_info != NULL )
   {
     free(group_info);
     group_info = 0;
   }

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

   if( group_info != NULL )
   {
     free(group_info);
     group_info = 0;
   }

   return(SMBC_PARAM_AN_ERR);

}

int smbc_param_an_GET_SRVL(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       max_list_num = 0;
  smb_vc        vc;
  smb_sv_list   server_list ;
  smb_sv_info*   server_info;
  int           i=0;
//////////////////////////////////////////////////////////////////

   bzero(&vc, sizeof(vc));
   bzero(&server_list , sizeof(server_list) );



   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_MAX_LST_N , smbc_rv_buf ,(char*)&max_list_num);

   server_info = malloc(sizeof(smb_sv_info)*max_list_num);
   if( server_info == NULL )
   {
      goto ERR_RET;
   }
   bzero( server_info , sizeof(smb_sv_info)*max_list_num );

   server_list.server_info = (smb_sv_info *)server_info;


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_GRP_NAME , smbc_rv_buf ,(char*)vc.group);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf ,(char*)vc.username);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD , smbc_rv_buf ,(char*)vc.password);


      smb_ret = NcsSmbGetServerList(
                                 ctx,
                                 &vc,
                                 max_list_num,
                                 &server_list,
                                 0 );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_R_CODE  SET_ERR \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL:: smb_ret --> [0x%08x]  \n",smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL:: max_list_num  --> [%d] \n", max_list_num ));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL:: vc.group  --> [%s] \n", vc.group ));
#endif
      printf("NcsSmbGetServerList smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_ENT ,
                                sizeof(server_list.entries) ,
                                (char*)&server_list.entries,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_ENT  SET_ERR \n"));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_TOTAL ,
                                sizeof(server_list.totalnum) ,
                                (char*)&server_list.totalnum,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_TOTAL  SET_ERR \n"));
#endif
     goto ERR_RET;
    }

for(i = 0; i < server_list.entries; i++) {


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CHAR_ID ,
                                sizeof(server_info[i].character) ,
                                (char*)&server_info[i].character ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_CHAR_ID SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SVI_SV_NAME ,
                                strlen(server_info[i].sv_name) ,
                                server_info[i].sv_name,
                                offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_SVI_SV_NAME SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SVI_SV_V_MJ ,
                                sizeof(server_info[i].sv_version_major) ,
                                (char*)&server_info[i].sv_version_major ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_SVI_SV_V_MJ  SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SVI_SV_V_MN ,
                                sizeof(server_info[i].sv_version_minor) ,
                                (char*)&server_info[i].sv_version_minor ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_SVI_SV_V_MN  SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SVI_SV_TYPE ,
                                sizeof(server_info[i].sv_type) ,
                                (char*)&server_info[i].sv_type,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_SVI_SV_TYPE  SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SVI_SV_COMM ,
                                strlen(server_info[i].sv_comment) ,
                                server_info[i].sv_comment ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVL::SMB_PARAM_SVI_SV_COMM  SET_ERR  i=%d \n",i));
#endif
     goto ERR_RET;
    }

}

MSG_SND:


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

  if( server_info != NULL )
  {
    free(server_info);
    server_info = 0;
  }

  return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  if( server_info != NULL )
  {
    free(server_info);
    server_info = 0;
  }

  return(SMBC_PARAM_AN_ERR);

}



int smbc_param_an_GET_SHRL(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       max_list_num = 0;
  smb_vc        vc;
  smb_sh_list   share_list ;
  smb_sh_info*   share_info = 0 ;
  int           i = 0;
//////////////////////////////////////////////////////////////////

   bzero(&vc, sizeof(vc));
   bzero(&share_list , sizeof(smb_sh_list) );


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_MAX_LST_N , smbc_rv_buf ,(char*)&max_list_num);

   share_info = malloc(sizeof(smb_sh_info)*max_list_num);
   if( share_info == NULL )
     {
      goto ERR_RET;
     }
   bzero( share_info , sizeof(smb_sh_info)*max_list_num );
   share_list.share_info = (smb_sh_info *)share_info;


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SRV_NAME , smbc_rv_buf ,(char*)vc.server);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf ,(char*)vc.username);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD , smbc_rv_buf ,(char*)vc.password);



   smb_ret = NcsSmbGetShareList(
                                 ctx,
                                 &vc,
                                 max_list_num,
                                 &share_list,
                                 0 );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_R_CODE  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL:: smb_ret --> [0x%08x]  \n", smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL:: max_list_num  --> [%d] \n", max_list_num ));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL:: vc.character  --> [%d] \n", vc.character ));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL:: vc.server  --> [%s] \n", vc.server ));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL:: vc.username  --> [%s] \n", vc.username ));
#endif
      printf("NcsSmbGetShareList smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_GRP_NAME ,
                                strlen(vc.group) ,
                                vc.group ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_GRP_NAME  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_ENT ,
                                sizeof(share_list.entries) ,
                                (char*)&share_list.entries,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_ENT  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }
   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_TOTAL ,
                                sizeof(share_list.totalnum) ,
                                (char*)&share_list.totalnum,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_TOTAL  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


  for(i = 0; i < share_list.entries; i++) 
  {


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CHAR_ID ,
                                sizeof(share_info[i].character) ,
                                (char*)&share_info[i].character ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_CHAR_ID  SET_ERR  i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SHI_SH_NAME ,
                                strlen(share_info[i].sh_name) ,
                                share_info[i].sh_name,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_SHI_SH_NAME  SET_ERR  i=%d  \n",i));
#endif
     goto ERR_RET;
    }
   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_SHI_SH_COMM ,
                                strlen(share_info[i].sh_comment) ,
                                share_info[i].sh_comment ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SHRL::SMB_PARAM_SHI_SH_COMM  SET_ERR  i=%d  \n",i));
#endif
     goto ERR_RET;
    }

}
MSG_SND:


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

  if( share_info != NULL )
  {
    free(share_info);  
    share_info = 0;
  }

  return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:
 
  if( share_info != NULL )
  {
    free(share_info);  
    share_info = 0;
  }

  return(SMBC_PARAM_AN_ERR);

}


int smbc_param_an_GET_DIRL(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       max_list_num = 0;
  smb_vc        vc;
  smb_dr_list   dr_list ;
  smb_dr_info*   directory_info = 0;
  u_char        directory[SMB_MAXFNAMELEN*2+1]  ;
  int           i = 0;
//////////////////////////////////////////////////////////////////


   bzero(&vc, sizeof(vc));
   bzero(&dr_list , sizeof(smb_dr_list) );
   bzero(directory , SMB_MAXFNAMELEN*2+1 );

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_MAX_LST_N , smbc_rv_buf ,(char*)&max_list_num);

   directory_info = malloc(sizeof(smb_dr_info)*max_list_num);
   if( directory_info == NULL )
   {
      goto ERR_RET;
   }
   bzero( directory_info , sizeof(smb_dr_info)*max_list_num );

   dr_list.directory_info = (smb_dr_info *)directory_info;



   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SRV_NAME , smbc_rv_buf ,(char*)vc.server);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SHRE , smbc_rv_buf ,(char*)vc.share);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf ,(char*)vc.username);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD , smbc_rv_buf ,(char*)vc.password);


   pos = smbc_prc_get_param( total_len , SMB_PARAM_DIR , smbc_rv_buf ,directory);


   smb_ret = NcsSmbDirectoryList(
                                 ctx,
                                 &vc,
                                 directory,
                                 max_list_num,
                                 &dr_list,
                                 0 );




   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_R_CODE  SET_ERR   \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::smb_ret --> [0x%08x]  \n",smb_ret));
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::directory  --> [%s] \n", directory ));
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::max_list_num  --> [%d] \n", max_list_num ));
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::vc.character  --> [%d] \n", vc.character ));
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::vc.server  --> [%s] \n", vc.server ));
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::vc.username  --> [%s] \n", vc.username ));
       /*
	 MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::vc.password  --> [%s] \n", vc.password ));
       */
       MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::vc.share  --> [%s] \n", vc.share ));
#endif
      printf("NcsSmbDirectoryList smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }




   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_GRP_NAME ,
                                strlen(vc.group) ,
                                vc.group ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_GRP_NAME  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_ENT ,
                                sizeof(dr_list.entries) ,
                                (char*)&dr_list.entries,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_ENT  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_TOTAL ,
                                sizeof(dr_list.totalnum) ,
                                (char*)&dr_list.totalnum,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_TOTAL  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


 for(i = 0; i < dr_list.entries; i++) 
 {

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CHAR_ID ,
                                sizeof(directory_info[i].character) ,
                                (char*)&directory_info[i].character ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_CHAR_ID  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_DLI_DR_NAME ,
                                strlen(directory_info[i].dr_name) ,
                                directory_info[i].dr_name ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_DLI_DR_NAME  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FID ,
                                sizeof(directory_info[i].finfo.fid) ,
                                (char*)&directory_info[i].finfo.fid ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FID  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CRE_TM ,
                                sizeof(directory_info[i].finfo.creation_time) ,
                                (char*)&directory_info[i].finfo.creation_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_CRE_TM  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ACC_TM ,
                                sizeof(directory_info[i].finfo.last_access_time) ,
                                (char*)&directory_info[i].finfo.last_access_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_ACC_TM  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_WRT_TM ,
                                sizeof(directory_info[i].finfo.last_write_time) ,
                                (char*)&directory_info[i].finfo.last_write_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_WRT_TM  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_CHG_TM ,
                                sizeof(directory_info[i].finfo.change_time) ,
                                (char*)&directory_info[i].finfo.change_time,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_CHG_TM  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_ATTR ,
                                sizeof(directory_info[i].finfo.attribute) ,
                                (char*)&directory_info[i].finfo.attribute,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_ATTR  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_FI_SIZE ,
                                sizeof(directory_info[i].finfo.size) ,
                                (char*)&directory_info[i].finfo.size,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DIRL::SMB_PARAM_FI_SIZE  SET_ERR i=%d  \n",i));
#endif
     goto ERR_RET;
    }


 }

MSG_SND:

   if( directory_info != NULL )
     {
       free(directory_info);
       directory_info = 0;
     }

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);



ERR_RET:

   if( directory_info != NULL )
     {
       free( directory_info );
       directory_info = 0;
     }

   return(SMBC_PARAM_AN_ERR);


}

int smbc_param_an_GET_SRVCAP(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  smb_vc           vc;
  int              offset = SMB_MSG_DF_START_POS;
  int              smb_ret = 0;
  int              pos = 0 ;
  u_short          total_len = 0;
  smb_server_cap   capability ;
//////////////////////////////////////////////////////////////////////////////////////

   bzero(&vc, sizeof(vc));
   bzero(&capability , sizeof(smb_server_cap) );

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_SRV_NAME , smbc_rv_buf ,(char*)vc.server);



   smb_ret = NcsSmbGetServerCapability(
            			        ctx,
			                &vc,
			                &capability,
			                0
                                        );


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP::SMB_PARAM_R_CODE  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP:: smb_ret --> [0x%08x] \n", smb_ret));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP:: vc.character  --> [%d] \n", vc.character ));
      MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP:: vc.server  --> [%s] \n", vc.server ));
#endif
      printf("NcsSmbGetServerCapability smb_ret -->0x%08x \n" , smb_ret);
      goto MSG_SND ;
     }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_GRP_NAME ,
                                strlen(vc.group) ,
                                vc.group ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP::SMB_PARAM_GRP_NAME  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CP_FLAG ,
                                sizeof(capability.flag) ,
                                (char*)&capability.flag,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP::SMB_PARAM_CP_FLAG  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }



   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CP_SEC_MD ,
                                sizeof(capability.sec_mode) ,
                                (char*)&capability.sec_mode,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_SRVCAP::SMB_PARAM_CP_SEC_MD  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


MSG_SND:

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);



ERR_RET:

   return(SMBC_PARAM_AN_ERR);


}


#ifdef NCS_SMB_NTAUTH_SUPPORT

int smbc_param_an_USR_LOGIN(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  smb_vc           vc;
  int              offset = SMB_MSG_DF_START_POS;
  int              smb_ret = 0;
  int              pos = 0 ;
  u_short          total_len = 0;
//////////////////////////////////////////////////////////////////////////////////////


   bzero(&vc, sizeof(vc));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_GRP_NAME , smbc_rv_buf ,(char*)vc.group);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf ,(char*)vc.username);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD , smbc_rv_buf ,(char*)vc.password);


   smb_ret = NcsSmbUserLogin(
			   ctx,
			   &vc,
			   0
                             );



   offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,
                            (char*)&smb_ret ,
                            offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_LOGIN::SMB_PARAM_R_CODE  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );



   return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

  return(SMBC_PARAM_AN_ERR);

}


int smbc_param_an_USR_LOGOUT(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  int         offset = SMB_MSG_DF_START_POS;
  int         smb_ret = 0;
  u_short     total_len = 0;
//////////////////////////////////////////////////////////////////////////////////////


  smb_ret = NcsSmbUserLogout( ctx , 0 );

  offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,
                            (char*)&smb_ret ,
                            offset );

  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_USR_LOGOUT::SMB_PARAM_R_CODE  SET_ERR  \n"));
#endif
      goto ERR_RET;
    }


    total_len = (u_short)offset;
    memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


    return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

    return(SMBC_PARAM_AN_ERR);


}


int smbc_param_an_GET_DCInf(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////////////////////////
  smb_vc             vc;
  int                offset = SMB_MSG_DF_START_POS;
  int                smb_ret = 0;
  int                pos = 0 ;
  u_short            total_len = 0;
  smb_dc_info        dcinfo;
//////////////////////////////////////////////////////////////////////////////////////

   bzero(&vc, sizeof(vc));
   bzero( &dcinfo , sizeof(smb_dc_info));

   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_GRP_NAME , smbc_rv_buf ,(char*)vc.group);


   smb_ret = NcsSmbGetDCInfo(ctx, &vc, &dcinfo, 0);


  offset = smbc_prc_set_param( smbc_sn_buf ,
                            SMB_PARAM_R_CODE ,
                            sizeof(smb_ret) ,
                            (char*)&smb_ret ,
                            offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DCInf::SMB_PARAM_R_CODE  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }

   if(smb_ret != SMBSUCCESS)
   {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DCInf:: smb_ret --> [0x%08x] \n",smb_ret));
     MPRINTF(("[smbc_process]smbc_param_an_GET_DCInf:: vc.character  --> [%d] \n", vc.character ));
     MPRINTF(("[smbc_process]smbc_param_an_GET_DCInf:: vc.group  --> [%s] \n", vc.group ));
#endif
     printf("NcsSmbGetDCInfo smb_ret -->0x%08x \n" , smb_ret);
     goto MSG_SND ;
   }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_DC_ADDR ,
                                sizeof(dcinfo.address) ,
                                (char*)&dcinfo.address,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_GET_DCInf::SMB_PARAM_DC_ADDR  SET_ERR  \n"));
#endif
     goto ERR_RET;
    }


MSG_SND:

   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);



ERR_RET:

  return(SMBC_PARAM_AN_ERR);


}
int smbc_param_an_USR_GET_Groups(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
/////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_short       max_list_num = 0;
  smb_vc        vc;
  smb_ugp_list  ugp_list;
  smb_ugp_info* ugp_info = 0 ;
  int           i=0;

//////////////////////////////////////////////////////////////////


   bzero(&vc, sizeof(vc));
   bzero(&ugp_list , sizeof(ugp_list) );


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );


   pos = smbc_prc_get_param( total_len , SMB_PARAM_MAX_LST_N , smbc_rv_buf ,(char*)&max_list_num);

   ugp_info = malloc(sizeof(smb_ugp_info)*max_list_num);
   if( ugp_info == NULL )
   {
#ifdef SMBCD_MFPRINT_USE
      MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::malloc ERR \n"));
#endif
      goto ERR_RET;
   }
   bzero( ugp_info , sizeof(smb_ugp_info)*max_list_num );

   ugp_list.ugp_info = (smb_ugp_info *)ugp_info ;

   pos = smbc_prc_get_param( total_len , SMB_PARAM_CHAR_ID , smbc_rv_buf ,(char*)&vc.character);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_GRP_NAME , smbc_rv_buf ,(char*)vc.group);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_USER_NAME , smbc_rv_buf ,(char*)vc.username);

   pos = smbc_prc_get_param( total_len , SMB_PARAM_PASS_WORD , smbc_rv_buf ,(char*)vc.password);


   smb_ret = NcsSmbUserGetGroups(
                                 ctx,
                                 &vc,
                                 max_list_num,
                                 &ugp_list,
                                 0 );




   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::SMB_PARAM_R_CODE  SET_ERR \n"));
#endif
     goto ERR_RET;
    }


   if(smb_ret != SMBSUCCESS)
     {
#ifdef SMBCD_MFPRINT_USE
       MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups:: smb_ret --> [0x%08x] \n",smb_ret));
       MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups:: max_list_num -> [%d]\n",max_list_num));
       MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups:: vc.character -> [%d]\n",vc.character));
       MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups:: vc.group -> [%s]\n",vc.group));
       MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups:: vc.username -> [%s]\n",vc.username));       
#endif
       printf("smbc_param_an_USR_GET_Groups:: smb_ret -->0x%08x \n" , smb_ret);
       goto MSG_SND ;
     }


   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_ENT ,
                                sizeof(ugp_list.entries) ,
                                (char*)&ugp_list.entries,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::SMB_PARAM_ENT  SET_ERR \n"));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_TOTAL ,
                                sizeof(ugp_list.totalnum) ,
                                (char*)&ugp_list.totalnum,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::SMB_PARAM_TOTAL  SET_ERR \n"));
#endif
     goto ERR_RET;
    }

 for(i = 0; i < ugp_list.entries; i++) 
 {

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_CHAR_ID ,
                                sizeof(ugp_info[i].character) ,
                                (char*)&ugp_info[i].character ,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::SMB_PARAM_CHAR_ID  SET_ERR i=%d \n",i));
#endif
     goto ERR_RET;
    }

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_UGI_UGP_NAME ,
                                strlen(ugp_info[i].ugp_name) ,
                                ugp_info[i].ugp_name,
                                offset );
  if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_USR_GET_Groups::SMB_PARAM_UGI_UGP_NAME  SET_ERR i=%d \n",i));
#endif
     goto ERR_RET;
    }
 }

MSG_SND:

  if( ugp_info != NULL )
    {
       free(ugp_info);
       ugp_info = 0;
    }


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );


   return(SMBC_PARAM_AN_SUCCESS);

ERR_RET:

  if( ugp_info != NULL )
    {
      free(ugp_info);
      ugp_info = 0;
    }

  return(SMBC_PARAM_AN_ERR);

}
#endif /* NCS_SMB_NTAUTH_SUPPORT */

int smbc_param_an_DF_CHR_CODE(const u_char* smbc_rv_buf ,u_char* smbc_sn_buf, smb_ctx* ctx)
{
//////////////////////////////////////////////////////////////////
  int           offset=SMB_MSG_DF_START_POS;
  int           smb_ret = 0 ;
  u_short       total_len = 0;
  int           pos = 0;
  u_char        char_code = 0;
//////////////////////////////////////////////////////////////////


   memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );



   pos = smbc_prc_get_param( total_len , SMB_PARAM_DEF_CHR_CODE , smbc_rv_buf ,(char*)&char_code);


   NcsSmbSetDefaultCharCode(
                            char_code
                            );


   smb_ret = SMBSUCCESS;

   offset = smbc_prc_set_param( smbc_sn_buf ,
                                SMB_PARAM_R_CODE ,
                                sizeof(smb_ret) ,
                                (char*)&smb_ret ,
                                offset );
   if(offset == SMBC_PARAM_SET_ERR)
    {
#ifdef SMBCD_MFPRINT_USE
     MPRINTF(("[smbc_process]smbc_param_an_DF_CHAR_CODE::SMB_PARAM_R_CODE  SET_ERR \n"));
#endif
     goto ERR_RET;
    }


   total_len = (u_short)offset;
   memcpy(smbc_sn_buf, &total_len , SMB_MSG_TOTALLEN_LEN );

   return(SMBC_PARAM_AN_SUCCESS);


ERR_RET:

   return(SMBC_PARAM_AN_ERR);

}



int smbc_prc_get_param(int max_srch_cnt ,
                                u_char srch_param , 
                          const u_char *smbc_rv_buf ,
                                char* set_param )
{
////////////////////////////////////////////////////////////////////////////////////
int           position = SMBC_PARAM_SERCH_ERR;
int           i=0;
u_short       length = 0;
int           ret = 0;
////////////////////////////////////////////////////////////////////////////////////

    for (i=SMB_MSG_DF_START_POS; i <= max_srch_cnt ; i++ )
      {
        ret = memcmp(&srch_param, smbc_rv_buf+i, SMB_PARAM_TAG_LEN );
        if( ret == 0 )
	  {
              position = i + SMB_PARAM_LENGTH_LEN +SMB_PARAM_TAG_LEN ;  
              memcpy( &length , smbc_rv_buf+position-SMB_PARAM_LENGTH_LEN , SMB_PARAM_LENGTH_LEN );
              memcpy( set_param , smbc_rv_buf+position , length );
              break;
	  }
         memcpy(&length , smbc_rv_buf+i+SMB_PARAM_TAG_LEN , SMB_PARAM_LENGTH_LEN );
         i += length + SMB_PARAM_LENGTH_LEN ;
      }

    if( (position < SMB_MIN_BUF_SIZE) || (position > SMB_MAX_BUF_SIZE))
      {
         position = SMBC_PARAM_SERCH_ERR;
      }

    return (position); 
}


int smbc_prc_set_param( u_char      *smbc_sn_buf ,
                   u_char      param_tag ,
                   u_short     length ,
                   u_char      *param_value ,
                   int         in_offset
                  )
{
//////////////////////////////////////////////////////////////////
//                                                              //
//                                                              //
//                                                              //
//                                                              //
//                                                              //
//////////////////////////////////////////////////////////////////
/////////////////
int offset = 0;
/////////////////


    offset = in_offset ;

    if(offset+length+SMB_PARAM_TAG_LEN+SMB_PARAM_LENGTH_LEN > SMB_MAX_BUF_SIZE )
      {
        printf("smbc_prc_set_param::length OVER!!! %d\n",offset+length+SMB_PARAM_TAG_LEN+SMB_PARAM_LENGTH_LEN);
#ifdef SMBCD_MFPRINT_USE
        MPRINTF(("smbc_prc_set_param:: MAX_LENGTH  OVER !! %d \n"));
#endif
        return(SMBC_PARAM_SET_ERR);
      }


    memcpy(smbc_sn_buf+offset, &param_tag , SMB_PARAM_TAG_LEN );
    offset += SMB_PARAM_TAG_LEN;

    memcpy(smbc_sn_buf+offset, &length , SMB_PARAM_LENGTH_LEN);
    offset += SMB_PARAM_LENGTH_LEN;

    if( length == 0)
    {
      return(offset);
    }

    memcpy(smbc_sn_buf+offset, param_value , length);
    offset += length;

    return(offset);

}


int smbc_prc_recv_buf( u_char* smbc_rv_buf ,int smbc_srv_fd )
{
/////////////////////////////////////////////////////////////
int             ret = -1;
int             rest_size = 0 ;
int             wk_total = 0;
u_short         total_len = 0;
fd_set          readfds;
struct timeval  timeout;
int             selret = -1;
/////////////////////////////////////////////////////////////

    ret = smbc_prc_recv_min(smbc_srv_fd , smbc_rv_buf ,0 );
    if( ret == SMBC_MSG_RCV_ERR)
    {
         goto RECV_ERR;
    }

    memcpy(&total_len , smbc_rv_buf , SMB_MSG_TOTALLEN_LEN );

    if( (total_len < SMB_MIN_BUF_SIZE) || (total_len > SMB_MAX_BUF_SIZE ))
      {
         printf("smbc_prc_recv_buf::total_len ERR %d \n" ,total_len);
         goto RECV_ERR;
      }

    rest_size = total_len - SMB_MIN_BUF_SIZE;
    wk_total = SMB_MIN_BUF_SIZE;
    while( 1 ) 
    {
        FD_ZERO(&readfds);
        FD_SET(smbc_srv_fd, &readfds);
        timeout.tv_sec = 600;
        timeout.tv_usec = 0;
        selret = select(smbc_srv_fd+1, &readfds, NULL, NULL, &timeout);
        if ( selret == 0 )
        {
           // TIME OUT
           goto RECV_ERR;
        }
        if(FD_ISSET (smbc_srv_fd, &readfds))
        {
          ret = recv( smbc_srv_fd, smbc_rv_buf + wk_total , rest_size , 0 );
          if( ret > 0 ) 
          {
            wk_total += ret;
            rest_size -= ret;
           } else {
            printf("smbc_prc_recv_buf:: recv NG!!!!  smbc_srv_fd -->  %d \n", smbc_srv_fd);
            total_len = 0;
            break;
           }
	}
        if ( rest_size <= 0 )
        {
            break;
	}
    }


    if( rest_size != 0 )
    {
      goto RECV_ERR;
    }

   return(total_len);

RECV_ERR:

#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("[smbc_process]smbc_prc_recv_buf:: RECV_ERR !! \n"));
#endif
    printf("smbc_prc_recv_buf::pid[%d] recv NG!!!!  \n", (int)getpid() );
    return(SMBC_MSG_RCV_ERR);

}


int smbc_prc_recv_min( int smbc_srv_fd , u_char* smbc_rv_buf , int recv_option)
{
////////////////////////////////////////////////
fd_set          readfds;
int             selret = -1;
int ret = -1;
int             rest_size = 0 ;
int             wk_total = 0;
////////////////////////////////////////////////

    rest_size = SMB_MIN_BUF_SIZE;

    while(1)
    {
      FD_ZERO(&readfds);
      FD_SET(smbc_srv_fd, &readfds);
      selret = select(smbc_srv_fd+1, &readfds, NULL, NULL, NULL);
      if ( selret == 0 )
	{
         goto RECV_ERR;
	}
      if(FD_ISSET (smbc_srv_fd, &readfds))
	{
         ret = recv( smbc_srv_fd , smbc_rv_buf ,SMB_MIN_BUF_SIZE , recv_option );
         if( ret > 0 )
	   {
            wk_total += ret;
            rest_size -= ret;
	   } else {
            goto RECV_ERR;
	   }
         if ( rest_size <= 0 )
	   {
            break;
	   }
	}
    }

    if( rest_size != 0 )
    {
      goto RECV_ERR;
    }

    return(SMB_MIN_BUF_SIZE);

RECV_ERR:

   printf("smbc_prc_recv_min::pid[%d] recv_min NG!!!!  \n", (int)getpid() );
   return(SMBC_MSG_RCV_ERR);

}




int smbc_prc_send_buf( const u_char* smbc_sn_buf , u_short total_len ,int smbc_srv_fd)
{
/////////////////////////////////////////////////
int             ret = -1;
int             rest_size = 0 ;
int             wk_total = 0;
fd_set          writefds;
struct timeval timeout;
int             selret = -1;
/////////////////////////////////////////////////


    rest_size = total_len ;


    while( 1 )
      {

        FD_ZERO(&writefds);
        FD_SET(smbc_srv_fd, &writefds);
        timeout.tv_sec = 600;
        timeout.tv_usec = 0;
        select(smbc_srv_fd+1, NULL, &writefds, NULL, &timeout);
        if ( selret == 0 )
        {
          goto ERR_RET;
        }

        if(FD_ISSET (smbc_srv_fd, &writefds))
        {
          ret = send( smbc_srv_fd, smbc_sn_buf + wk_total , rest_size , 0 );
          if( ret > 0 )
            {
              wk_total += ret;
              rest_size -= ret;
            } else {
              printf("smbc_lib_send_buf:: read NG!!!!  smbc_srv_fd -->  %d \n", smbc_srv_fd );
              goto ERR_RET;
            }
          if ( rest_size <= 0 )
            {
              break;
            }
        }
    }

    if( rest_size != 0)
    {
      goto ERR_RET;
    }

    return(ret);


ERR_RET:

#ifdef SMBCD_MFPRINT_USE
    MPRINTF(("smbc_prc_send_buf::pid[%d] SEND_ERR !! \n",(int)getpid() ));
#endif
    return(SMBC_MSG_SND_ERR);


}



int smbc_prc_cnv_err(int class, int condition, int code)
{
//////////////////////////////////////////////////////////////////
//                                                              //
//                                                              //
//                                                              //
//                                                              //
//                                                              //
//////////////////////////////////////////////////////////////////
/////////////////////////////
        int     ret = 0;
/////////////////////////////

        ret |= (condition & 0xff) << 24;
        ret |= (class & 0xff) << 16;
        ret |= (code & 0xffff);

        return ret;

}



 
u_short  smbc_prc_buf_ID_PEEK(char* smbc_rv_buf,u_char *check_msg_ID, int smbc_srv_fd )
{
////////////////////////////////////////////////
fd_set          readfds;
struct timeval  timeout;
int             selret = -1;
int ret = -1;
int             rest_size = 0 ;
int             wk_total = 0;
static pthread_mutex_t recvDataMutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t recvDataCond = PTHREAD_COND_INITIALIZER;
static int recvDataProcessed  = 0;
u_char   message_ID = 0;
////////////////////////////////////////////////


    pthread_mutex_lock(&recvDataMutex);

    rest_size = SMB_MIN_BUF_SIZE;

    while(1)
    {
      FD_ZERO(&readfds);
      FD_SET(smbc_srv_fd, &readfds);
      timeout.tv_sec = 600;
      timeout.tv_usec = 0;
      selret = select(smbc_srv_fd+1, &readfds, NULL, NULL, &timeout);
      if ( selret == 0 )
	{
         // TIME OUT
         goto RECV_ERR;
	}
      if(FD_ISSET (smbc_srv_fd, &readfds))
	{
         ret = recv( smbc_srv_fd , smbc_rv_buf ,SMB_MIN_BUF_SIZE , MSG_PEEK );
         if( ret > 0 )
	 {
            wk_total += ret;
            rest_size -= ret;
	 } 
         else
         {
            goto RECV_ERR;
	 }
         if ( rest_size == 0 )
         {
            memcpy(&message_ID , smbc_rv_buf+SMB_MSG_TOTALLEN_LEN , SMB_MSG_MSG_ID_LEN);
            if(*check_msg_ID == message_ID)
            {
              break;
            }
            while(!recvDataProcessed)
            {
              pthread_cond_wait(&recvDataCond, &recvDataMutex);
            }
            recvDataProcessed = 0;
	 }
         else
         {
            goto RECV_ERR;
	 }
	}
    }// while end

    recvDataProcessed = 1;
    pthread_cond_signal(&recvDataCond);
    pthread_mutex_unlock(&recvDataMutex);

   return(SMB_MIN_BUF_SIZE);

RECV_ERR:


    recvDataProcessed = 1;
    pthread_cond_signal(&recvDataCond);
    pthread_mutex_unlock(&recvDataMutex);

    printf("smbc_lib_recv_min:: recv_min NG!!!!  smbc_srv_fd -->  %d \n", smbc_srv_fd);
    return(SMBC_MSG_RCV_ERR);

}

