#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include <netinet/in.h>
#include "includes.h"
#include <smbc_lib.h>
#include <syslog.h>
#include <net/if_types.h>
#include <net/if_dl.h>
#ifndef SMBCD_USE
#include <mfprintf.h>
#include <ncs_mmesgkey.h>
#endif
#define FTAPI_ABORT_EXTERN	/* Oe[uQ */

#define	MAILSLOT_BROWSE	"\\MAILSLOT\\BROWSE"
#define	SMBSERVER	"*SMBSERVER"

static void *smbc_mf = NULL;

// -------------------------------------------------------------
// internal library
// -------------------------------------------------------------

static void smb_cleanup( smb_ctx* ctx );
#ifdef notdef
static void server_fn(const char *name, uint32 m, const char *comment, void *state);
#endif
int smb_get_myname(char *name);
int smb_get_mygroup(char *group);
int smb_get_mpx(int *mpx);
int smb_get_myip(u_long *ip, int intf);
int smb_get_wins_addr(u_long *primary, u_long *secondary, char *scopeid, int intf);
#if 0
static int cli_GetBackupList(smb_ctx* ctx, const char *myname, const char *group_name, char *server);
static int cli_SendBackupListReq(int nbt_fd, char *mailslot,
                      char *buf, int len,
                      const char *srcname, int src_type,
                      const char *dstname, int dest_type,
                      u_long dest_ip, u_long src_ip,
                      int dest_port, int src_port);
static int cli_GetBackupListResp(int nbt_fd, char *mailslot, char *buf, int bufsiz);
#endif
int smbc_errno(struct cli_state *cli);
static int smbc_dos_errno(uint8 eclass, uint32 ecode);
int smb_getnodestatus(const NBTNamePack* rmt, NBTNodeName* nn, NBTNodeStatistics* ns, u_short suffix);
static int  NcsGetSmbConf( char* tag, char* menu, char* str, int str_max );
static int  NcsGetNbtdConf( char* tag, char* menu, char* str, int str_max );
extern void set_smb_default_char_code(u_char char_code);
extern u_char get_smb_default_char_code(void);
extern void set_smb_char_code(u_char char_code);
extern u_char get_smb_char_code(void);

typedef struct stIFConf_tag
{
	u_char	iftype;
	char	ifname[IFNAMSIZ];
#define IFCONF_MAC_MAX	20
	u_char	mac[IFCONF_MAC_MAX];
	int		mac_len;
	
} stIFConf;

static int smb_GetIFConf(stIFConf* dst, stIFConf* src);
static int smb_GetIFBCAddr(u_char iftype, u_long* bcnaddr, const char* ifname);

#define NcsGetXXXConf   smbcd_NcsGetXXXConf
int     NcsGetXXXConf( char* file, char* tag, char* menu, char* str, int str_max );     // get_conf.c

#define	smbc_openlog		ncs_smbc_openlog
long	smbc_openlog = 0;		// openlogςۂ
int nbt_connectTimeout(int *s, NBTNamePack* rmt, const NBTNamePack* local, u_long *ip, int num);

#ifdef SMB_DEBUG_LOGPRINT
#define SMBC_LOG(body)	{ printf body; printf("\n");}
#else
#define SMBC_LOG(body)	
#endif

/**********************************************************
 * nbt_connectTimeout
 * pthreadłnbt_connect^CAEgȂ̂
 * ^CAEg[`쐬
 *
 *********************************************************/
static void connectAlarm(int signo)
{
	DPRINTF(("### connect TIMEOUT ###\n"));
	return;	// interupts connect()
}
#define	CONNECT_TIMEO	10		// 10 sec

int nbt_connectTimeout(int *s, NBTNamePack* rmt, const NBTNamePack* local, u_long *ip, int num)
{
	int ret = 0;
	sig_t	oldAlrmp = NULL;
	int i = 0;
	int	abort = FTAPI_ABORT_REQOFF;	/* ABORT */
	int	conn_errno = 0;

	for(i = 0; i < num; i++) {
		*s = nbt_open(NBT_STYPE_SESSION, htonl(INADDR_ANY), htons(NBT_PORT_ANY));
		if(*s < 0){
			printf("nbt_connectTimeout: failed nbt_open\n");
#ifndef SMBCD_USE
			smbc_mfprintf("nbt_connectTimeout: failed nbt_open\n");
#endif
			nbt_close(*s);
			*s = -1;
			ret = -1;
			break;
		}
		rmt->naddr = ip[i];

#ifndef SMBCD_USE
		/* fv`FbN */
		abort = FtapiAbortReqCheck();
		if(abort == FTAPI_ABORT_REQON) {
			printf("nbt_connectTimeout: abort\n");
			smbc_mfprintf("nbt_connectTimeout: abort\n");
			nbt_close(*s);
			*s = -1;
			ret = -1;
			break;
		}
#endif

		oldAlrmp = signal(SIGALRM, connectAlarm);
#ifdef SMBC_USE_PTL
		pthread_alarm_np(FTAPI_ABORT_INTERVAL);
#else
		alarm(FTAPI_ABORT_INTERVAL);
#endif
		ret = nbt_connect(*s, rmt, local, -1);
#ifdef SMBC_USE_PTL
		pthread_alarm_np(0);
#else
		alarm(0);
#endif
		signal(SIGALRM, oldAlrmp);

#ifndef SMBCD_USE
		/* fv`FbN */
		abort = FtapiAbortReqCheck();
		if(abort == FTAPI_ABORT_REQON) {
			printf("nbt_connectTimeout: abort\n");
			smbc_mfprintf("nbt_connectTimeout: abort\n");
			nbt_close(*s);
			*s = -1;
			ret = -1;
			break;
		}
#endif
		if(ret == 0) {
			break;
		}

		conn_errno = ERRNO;
		if( (conn_errno == ETIMEDOUT) || (conn_errno == EINTR) )
		{
			fd_set	rfds;	/* Read			*/
			fd_set	wfds;	/* Write		*/
			struct timeval	interval;	/* C^[o		*/
			int	retrycnt = 0;		/* ĎgCJE^	*/
			int	retrymax = 0;		/* ĎgC	*/
			int	retryendinterval = 0;	/* ĎŏIC^[o	*/
			int	option = 0;		/* \PbgIvVerrno	*/
			int	optionlen = sizeof(option);	/* \PbgIvVerrno	*/

			/* connectĎ */
			/* gC񐔐ݒ	*/
			retrymax = ((CONNECT_TIMEO - FTAPI_ABORT_INTERVAL)/FTAPI_ABORT_INTERVAL) + 1;	
			retryendinterval = (CONNECT_TIMEO - FTAPI_ABORT_INTERVAL)%FTAPI_ABORT_INTERVAL;		/* ŏIC^[oݒ	*/
			if( retryendinterval == 0 )
			{				/* ŏIC^[o0̏ꍇČvZ	*/
				retrymax -= 1;
				retryendinterval = FTAPI_ABORT_INTERVAL;
			}
			interval.tv_sec = FTAPI_ABORT_INTERVAL;		/* ftHgC^[oݒ	*/
			interval.tv_usec = 0;

			for( retrycnt = 0; retrycnt < retrymax; retrycnt ++ )
			{
				FD_ZERO( &rfds );
				FD_ZERO( &wfds );
				FD_SET( *s, &rfds );
				FD_SET( *s, &wfds );	/* Read&Writeݒ	*/

#ifndef SMBCD_USE
				/* fv`FbN(select<connect>O) */
				abort = FtapiAbortReqCheck();
				if( abort == FTAPI_ABORT_REQON )
				{
					printf("nbt_connectTimeout: abort\n");
					smbc_mfprintf("nbt_connectTimeout: abort\n");
					nbt_close(*s);
					*s = -1;
					ret = -1;
					break;
				}
#endif
				if( retrycnt == (retrymax - 1) )
				{
					interval.tv_sec = retryendinterval;	/* ŏIC^[o	*/
				}
				ret = select( (*s + 1), &rfds, &wfds, NULL, &interval );
				if( ret > 0 )
				{
					if( (FD_ISSET( *s, &rfds ) != 0) ||
					    (FD_ISSET( *s, &wfds ) != 0)    )
					{
						if( getsockopt( *s, SOL_SOCKET, SO_ERROR,
										&option, &optionlen ) < 0 )
						{
							nbt_close(*s);
							*s = -1;
							ret = -1;
							break;
						}
						if( option == ETIMEDOUT ){	// Operation timed out
							nbt_close(*s);
							*s = -1;
							ret = -1;
							break;
						}
						if( option == ECONNREFUSED ){	// Connection refused
							nbt_close(*s);
							*s = -1;
							ret = -1;
							break;
						}
						if( option != 0 ){
							nbt_close(*s);
							*s = -1;
							ret = -1;
							break;
						}
#ifndef SMBCD_USE
						/* fv`FbN(select<connect>) */
						abort = FtapiAbortReqCheck();
						if( abort == FTAPI_ABORT_REQON )
						{
							nbt_close(*s);
							*s = -1;
							ret = -1;
							break;
						}
#endif
						break;
					}
				}
				else if ( ret < 0 )
				{
					nbt_close(*s);
					*s = -1;
					ret = -1;
					break;
				}
				/* else if ( ret == 0 ) { continue; } */
			}
			if( retrycnt >= retrymax )
			{
				/* ^CAEg	*/
				nbt_close(*s);
				*s = -1;
				ret = -1;
				break;
			}

			if( *s == -1 ){
				break;				// break loop
			}

		}
		else
		{
			if( conn_errno == ECONNREFUSED ){	// Connection refused
#ifndef SMBCD_USE
				smbc_mfprintf("nbt_connectTimeout: Connection refused(name=%s, naddr=0x%x, num=%d)\n", rmt->name, rmt->naddr, num);
#endif
			}
			else {
#ifndef SMBCD_USE
				smbc_mfprintf("nbt_connectTimeout: Connection failed(name=%s, naddr=0x%x, num=%d)\n", rmt->name, rmt->naddr, num);
#endif
			}
			nbt_close(*s);
			*s = -1;
			ret = -1;
			break;
		}

#ifndef SMBCD_USE
		smbc_mfprintf("nbt_connectTimeout: timeout(name=%s, naddr=0x%x, num=%d)\n", rmt->name, rmt->naddr, num);
#endif
		nbt_close(*s);
		*s = -1;
		ret = -1;
	}

	return ret;
}

/**********************************************************
 * NcsSmbSetDefaultCharCode
 *
 *
 *********************************************************/
void NcsSmbSetDefaultCharCode( u_char char_code )
{
	SMBC_LOG(("NcsSmbSetDefaultCharCode(0x%x) enter", char_code));
	set_smb_default_char_code(char_code);
	SMBC_LOG(("NcsSmbSetDefaultCharCode exit"));
}

static char * smb_strrchr(char *s, int c);
/**********************************************************
 * NcsSmbOpen
 *
 *
 *********************************************************/
int NcsSmbOpen( smb_ctx** ctx )
{
	smb_ctx	*context = 0;
	int	res = 0;

	SMBC_LOG(("NcsSmbOpen(0x%x) enter", ctx));
	if(!ctx) {
		SMBC_LOG(("NcsSmbOpen exit"));
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if( smbc_openlog == 0 ){
		openlog( "smb client", (LOG_PID | LOG_NDELAY), LOG_USER );
		smbc_openlog = 1;
	}

	context = (smb_ctx *)smb_malloc(sizeof(smb_ctx));
	if(!context) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpen: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	bzero(context, sizeof(smb_ctx));
	context->cli = NULL;
	context->ss_client = -1;

	res = nbt_setnodetype(NBT_ONT_H);
	if(res < 0) {
		printf("NcsSmbOpen:failed nbt_setnodetype (%d)\n", res);
		smb_cleanup(context);
		SMBC_LOG(("NcsSmbOpen exit"));
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	smb_get_myname(context->local.name);
	if(context->local.name[0] == NULL) {
		strcpy(context->local.name, "*");
	}
#ifdef SMBC_USE_PTL
	/*
	 *mutex init
	 */
	res = pthread_mutex_init(&context->mutex, NULL);
	if(res != 0){
		smb_cleanup(context);
		printf("smbc_lib: pthread_mutex_init() error[%d].",res);
		SMBC_LOG(("NcsSmbOpen exit"));
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
#endif
	smb_srandom((u_long)context);
	*ctx = context;

	SMBC_LOG(("NcsSmbOpen exit"));
	return SMBSUCCESS;
}

int NcsSmbClose( smb_ctx* ctx )
{
	int	smberr = SMBSUCCESS;
#ifdef SMBC_USE_PTL
	int	status = 0;
#endif

	SMBC_LOG(("NcsSmbClose(0x%x) enter", ctx));
	if( !ctx ) {
		SMBC_LOG(("NcsSmbClose exit"));
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if(ctx->cli) {
		if(ctx->cli->cnum != 0xffff) {
			if(!cli_tdis(ctx->cli)) {
				smberr = smbc_errno(ctx->cli); 
			}
		}
		if(ctx->cli->vuid != UID_FIELD_INVALID) {
			if(!cli_ulogoff(ctx->cli)) {
				smberr = smbc_errno(ctx->cli); 
			}
		}
	}
	smb_cleanup(ctx);
#ifdef SMBC_USE_PTL
	/*
	 * mutex destruct
	 */
	status = pthread_mutex_destroy(&ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_destroy() error[%d].",status);
		smberr = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
#endif
	SAFE_FREE(ctx);
	SMBC_LOG(("NcsSmbClose exit"));
	return smberr;
}

/**********************************************************
 * NcsSmbConnect
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbConnect(smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback)
{
	int				ret = 0;
	int				status = 0;
	pthread_t 		thread_id;
	smbc_connect_t  *th_arg = NULL;

	SMBC_LOG(("NcsSmbConnect(0x%x, 0x%x, 0x%x) enter", ctx, vcp, callback));
	if(!ctx || !vcp) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnect invalid arg. \n");
#endif
		SMBC_LOG(("NcsSmbConnect exit(0x%x)", ret));
		return ret;
	}
	if(ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnect invalid cmdseq. \n");
#endif
		SMBC_LOG(("NcsSmbConnect exit(0x%x)", ret));
		return ret;
	}

	if(callback){
		th_arg = (smbc_connect_t *)smb_malloc(sizeof(smbc_connect_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnect: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbConnectThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnect failed: pthread_create. \n");
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnect failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbConnect exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbConnectSub(ctx, vcp);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnect failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbConnect exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbConnect exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbConnectThread(void *arg)
{
	int		ret = 0;
	int		status = 0;
	smbc_connect_t  *th_arg = (smbc_connect_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbConnectSub(th_arg->ctx, th_arg->vcp);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_CONNECT, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);

#ifdef DEBUG
	printf("[NcsSmbConnectThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbConnectSub( smb_ctx* ctx, smb_vc* vcp)
#else
int NcsSmbConnect( smb_ctx* ctx, smb_vc* vcp, SMBC_CALLBACK* callback )
#endif
{
	int	res = 0;

	struct in_addr dest_ip;
	struct nmb_name called;
	NBTAddrPack	rmt;
	char		group[SMB_MAXSRVNAMELEN+1];
	char		server[SMB_MAXHOSTNAMELEN+1];
	char		share[SMB_MAXSHARENAMELEN+1];
	char		username[SMB_MAXUSERNAMELEN+1];
	char		password[SMB_MAXPASSWORDLEN+1];
	u_short		char_code = vcp->character;
	int		name_type = 0;
	int		NcsSmbConnect_negprot_fail = 0;
	u_long		target_ip[SMB_MAX_IP];
	int		num = SMB_MAX_IP;
	int		param = TRUE;

	if(!ctx || !vcp) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub invalid arg. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}

	if(ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub invalid command sequence. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	ctx->cli = cli_initialise(NULL);
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub failed: cli_initialise. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	SMBC_LOG(("input server name [%s]", vcp->server));
	SMBC_LOG(("input share name [%s]", vcp->share));
	SMBC_LOG(("input user name [%s]", vcp->username));

	bzero(&dest_ip, sizeof(dest_ip));
	bzero(&called, sizeof(called));
	bzero(&rmt, sizeof(rmt));
	bzero(group, sizeof(group));
	bzero(server, sizeof(server));
	bzero(share, sizeof(share));
	bzero(username, sizeof(username));
	bzero(password, sizeof(password));
	set_smb_char_code(char_code);

	if(vcp->group[0] != 0) {
		SMBC_LOG(("input group name [%s]", vcp->group));
		strncpy(group, vcp->group, SMB_MAXSRVNAMELEN);
	} else {
		smb_get_mygroup(group);
	}

	strncpy(server, vcp->server, SMB_MAXHOSTNAMELEN);
	strncpy(share, vcp->share, SMB_MAXSHARENAMELEN);
	strncpy(username, vcp->username, SMB_MAXUSERNAMELEN);
	strncpy(password, vcp->password, SMB_MAXPASSWORDLEN);

	/*
	 * IPC$ɂ͐ڑłȂ悤ɂ
	 */
	if (strcasecmp(share,"IPC$")==0) {
		syslog( LOG_ERR, "SMBC lib. cannot connect to IPC$." );
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub cannot connect to IPC$. \n");
#endif
		return smbc_errcode(SMBERR_DOS, SMBERR_PERMANENT, SMBERR_NOACCESS);
	}
	bzero(&ctx->rmt, sizeof(ctx->rmt));
	ctx->rmt.suffix = SUFFIX_SERV;
	name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);
	if(!name_type) {
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. cannot resolve server name(server=%s).",  server);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub failed: smb_gethostbyname. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	} else if(name_type != 1) {
		int res = 0;
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;
		int ii = 0;

		/*
		 * ftHgł"*SMBSERVER"w肷B
		 * ^[QbgPCWindowsNT4.0 SP6aȍ~OSł
		 * ̓ꖼŐڑ\
		 */
		strncpy(server, SMBSERVER, SMB_MAXSRVNAMELEN);
		for(ii = 0; ii < num; ii++) {
			ctx->rmt.naddr = target_ip[ii];
			memset(&nn, 0, sizeof(nn));
			memset(&ns, 0, sizeof(ns));
			res = smb_getnodestatus(&ctx->rmt, &nn, &ns, SUFFIX_SERV);
			if(res == LIBNBT_ERR_NOERR) {
				SMBC_LOG(("smb_getnodestatus: name = %s", nn.name)); 
				strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
				if(ii > 0) {
					/* ڑAhX̑1ɂĂ */
					u_long tmp_ip;
					tmp_ip = target_ip[0];
					target_ip[0] = target_ip[ii];
					target_ip[ii] = tmp_ip;
				}
				break;
			}
		}
	}
	strncpy(ctx->rmt.name, server, SMB_MAXSRVNAMELEN);

NcsSmbConnect_retry:
	res = nbt_connectTimeout(&ctx->ss_client, &ctx->rmt, &ctx->local, target_ip, num);
	if(res < 0) {
		printf("NcsSmbConnect: failed nbt_connectTimeout (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub failed: nbt_connectTimeout. \n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. failed to connect server(server=%s).", ctx->rmt.name );
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}

	res = nbt_getfd(ctx->ss_client);
	if(res < 0) {
		printf("NcsSmbConnect: failed nbt_getfd (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub failed: nbt_getfd. \n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	ctx->cli->fd = res;
	if(setsockopt(ctx->cli->fd, IPPROTO_TCP, TCP_NODELAY, (char *)&param, sizeof(param)) < 0) {
		syslog( LOG_WARNING, "SMBC lib. setsockopt (TCP_NODELAY)");
	}
	ctx->cli->port = 139;
	fstrcpy(ctx->cli->desthost, server);
	fstrcpy(ctx->cli->user_name, username);
	fstrcpy(ctx->cli->domain, group);
	fstrcpy(ctx->cli->share, share);
	fstrcpy(ctx->cli->called.name, server);
	ctx->cli->called.name_type = SUFFIX_SERV;
	ctx->cli->dest_ip.s_addr = htonl(ctx->rmt.naddr);

	if(!cli_negprot(ctx->cli, &char_code)) {
		if(ctx->cli->fd < 0) {
			if(NcsSmbConnect_negprot_fail++ < 5) {
				nbt_close(ctx->ss_client);
				usleep(smb_random()%1000+100);
				goto NcsSmbConnect_retry;
			}
			printf("cli_negprot fail %d\n", ERRNO);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnectSub failed: cli_negprot. \n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}

		printf("retry NULL domain\n");
#ifndef SMBCD_USE
		smbc_mfprintf("retry NULL domain\n");
#endif
		ctx->cli->domain[0] = NULL;
		if(!cli_negprot(ctx->cli, &char_code)) {
			int smberr;
			if(ctx->cli->fd < 0) {
				if(NcsSmbConnect_negprot_fail++ < 5) {
					nbt_close(ctx->ss_client);
					usleep(smb_random()%1000+100);
					goto NcsSmbConnect_retry;
				}
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}
			printf("NcsSmbConnect: failed negprot\n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnectSub failed: cli_negprot 2. \n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to negotiate(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		}
	}
#ifdef DEBUG
	SMBC_LOG(("server_type [%s]", ctx->cli->server_type));
	SMBC_LOG(("server_os [%s]", ctx->cli->server_os));
	SMBC_LOG(("server_domain [%s]", ctx->cli->server_domain));
	SMBC_LOG(("capabilities [0x%x]", ctx->cli->capabilities));
	SMBC_LOG(("protocol [%d]", ctx->cli->protocol));
#endif	// DEBUG

	/*
	 * 2002/11/21 terao add
	 */
	if(ctx->cli->server_domain) {
		strncpy(vcp->group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
		strncpy(group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
	}

	if (!cli_session_setup(ctx->cli, username,
		password, strlen(password), 
		password, strlen(password), 
		"", &char_code)) {

		/*
		 * if a password was not supplied
		 * then try again with a null username
		 */
		 if (!password[0] || username[0] ||
				!cli_session_setup(ctx->cli, "", "", 0, 
				"", 0, group, &char_code)) {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}

			printf("NcsSmbConnect: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnectSub failed: cli_session_setup (try again with a null username). \n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 } else {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}

			printf("NcsSmbConnect: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbConnectSub failed: cli_session_setup. \n");
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 }
	}

	if (!cli_send_tconX(ctx->cli, share, "?????",
					password, strlen(password)+1, &char_code)) {
		int smberr;
		if(ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else {
			smberr = smbc_errno(ctx->cli);
		}

		printf("NcsSmbConnect: cli_send_tconX failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbConnectSub failed: cli_send_tconX. \n");
#endif
		syslog( LOG_ERR, "SMBC lib. failed to tree connect(server=%s, share=%s).", ctx->rmt.name, share);
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smberr;
	}

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbDisconnect
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbDisconnect( smb_ctx *ctx, SMBC_CALLBACK* callback )
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_disconnect_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbDisconnect(0x%x, 0x%x) enter", ctx, callback));
	if(!ctx) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnect failed: invalid arg. \n");
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ); 
		SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnect failed: cli is null. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnect failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_disconnect_t *)smb_malloc(sizeof(smbc_disconnect_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDisconnect: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbDisconnectThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDisconnect: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDisconnect failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbDisconnectSub(ctx);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDisconnect failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbDisconnect exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbDisconnectThread(void* arg)
{
	int			ret = 0;
	int			status = 0;
	smbc_disconnect_t	*th_arg = (smbc_disconnect_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DISCONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DISCONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbDisconnectSub(th_arg->ctx);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DISCONNECT, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_DISCONNECT, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbDisconnectThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbDisconnectSub( smb_ctx* ctx )
#else
int NcsSmbDisconnect( smb_ctx* ctx, SMBC_CALLBACK* callback )
#endif
{
	int	smberr = SMBSUCCESS;

	if(!ctx) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectSub failed: ctx is null. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectSub failed: cli is null. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnectSub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(!cli_tdis(ctx->cli)) {
		printf("NcsSmbDisconnect: cli_tdis failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnect: cli_tdis failed: \n");
#endif
		syslog( LOG_ERR, "SMBC lib. failed to tree disconnect(server=%s).", ctx->rmt.name);
		if(ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else {
			smberr = smbc_errno(ctx->cli); 
		}
	} else if(!cli_ulogoff(ctx->cli)) {
		printf("NcsSmbDisconnect: cli_ulogoff failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDisconnect: cli_ulogoff failed: \n");
#endif
		syslog( LOG_ERR, "SMBC lib. failed to user logoff(server=%s).", ctx->rmt.name);
		if(ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else {
			smberr = smbc_errno(ctx->cli); 
		}
	}
	cli_shutdown(ctx->cli);
	ctx->cli = NULL;

	if(ctx->ss_client >= 0) {
	        nbt_close(ctx->ss_client);
		ctx->ss_client = -1;
	}

	return( smberr );
}

/**********************************************************
 * NcsSmbOpenFile
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbOpenFile( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback )
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_openfile_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbOpenFile(0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, file, finfo, callback));
	if(!ctx || !file || !finfo) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFile failed: ctx=%d, file=%d, finfo=%d \n", ctx, file, finfo);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFile failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFile failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_openfile_t *)smb_malloc(sizeof(smbc_openfile_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbOpenFile: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->file = file;
		th_arg->finfo = finfo;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbOpenFileThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbOpenFile: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbOpenFile failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbOpenFileSub(ctx, file, finfo);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbOpenFile failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbOpenFile exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbOpenFileThread(void* arg)
{
	int		ret = 0;
	int		status = 0;
	smbc_openfile_t *th_arg = (smbc_openfile_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_OPENFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_OPENFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbOpenFileSub(th_arg->ctx, th_arg->file, th_arg->finfo);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_OPENFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_OPENFILE, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbOpenFileThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbOpenFileSub( smb_ctx* ctx, smb_file* file, smb_finfo* finfo )
#else
int NcsSmbOpenFile( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback )
#endif
{
	const char	*fname = file->name;
	int		flags = 0;
	int		share_mode = 0;
	int		fid = 0;
	int		file_create = 0;
	u_short		char_code = file->character;
	int		ret = SMBSUCCESS;

	if(!ctx || !file || !finfo) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileSub failed: ctx=%d, file=%d, finfo=%d \n", ctx, file, finfo);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileSub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFileSub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input file name [%s]", file->name));
	SMBC_LOG(("input file flag [0x%x]", file->flag));
	SMBC_LOG(("input file disposition [0x%x]", file->disposition));
	SMBC_LOG(("input file attribute [0x%x]", file->attribute));
	set_smb_char_code(char_code);

	if(file->disposition == SMB_CREATE_NEW 
			|| file->disposition == SMB_CREATE_ALWAYS) {

		file_create = 1;
	}

	if(file_create && (file->flag & SMB_CREATE_DIR) 
				&& (strchr(fname, '\\') != NULL)) { 
		char	dir[SMB_MAXFNAMELEN+1];
		char	*dirp = 0;

		bzero(dir, sizeof(dir));
		strncpy(dir, fname, SMB_MAXFNAMELEN);
		dirp = dir;
		if(dir[0] == '\\') {
			dirp ++;	/* skip 1st */
		}
		while(dirp) {
			dirp = strchr(dirp, '\\');
			if(!dirp) {
				break;
			}
			*dirp = 0;
			if(!cli_chkpath(ctx->cli, dir, &char_code)) {
				if(!cli_mkdir(ctx->cli, dir, &char_code)) {
					printf("NcsSmbOpenFile: cli_mkdir(%s) failed: \n", dir);
#ifndef SMBCD_USE
					smbc_mfprintf("NcsSmbOpenFileSub failed: cli_mkdir. \n");
#endif
					return smbc_errno(ctx->cli);
				}
			}
			*dirp++ = '\\';
		}
	}

	if(file->attribute & SMB_ATTR_READONLY) {
		flags |= O_RDONLY;
	}
	if(file->attribute & SMB_ATTR_NORMAL) {
		flags |= O_RDWR;
	}
	if(file->disposition == SMB_CREATE_NEW) {
		flags |= (O_CREAT | O_EXCL);
	}
	if(file->disposition == SMB_CREATE_ALWAYS) {
		flags |= O_CREAT;
	}
	if(file->disposition == SMB_OPEN_ALWAYS) {
		flags |= O_CREAT;
	}
	if(file->disposition == SMB_OPEN_EXISTING) {
		flags &= ~O_CREAT;
	}
	if(!file_create && file->flag & SMB_OPEN_READONLY) {
		flags &= ~O_RDWR;
		flags |= O_RDONLY;
	}
	share_mode = DENY_NONE;

	fid = cli_open(ctx->cli, fname, flags, share_mode, &char_code);
	if(fid < 0) {
		printf("NcsSmbOpenFile: cli_open: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbOpenFile: cli_open: \n");
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
		syslog( LOG_ERR, "SMBC lib. cannot open file(server=%s file=%s).", ctx->rmt.name, fname);
	} else {
		finfo->fid = (u_short)fid;
	}

	return( ret );
}

/**********************************************************
 * NcsSmbCloseFile
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbCloseFile( smb_ctx* ctx, u_short fid, SMBC_CALLBACK* callback )
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_closefile_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbCloseFile(0x%x, 0x%x, 0x%x) enter", ctx, fid, callback));
	if(!ctx) {
		printf("NcsSmbCloseFile failed: ctx is NULL. \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFile failed: ctx is NULL. \n");
#endif
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
		return ret;
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFile failed: cli is NULL. \n");
#endif
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFile failed: cnum is 0xffff. \n");
#endif
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
		return ret;
	}

	if(callback){
		th_arg = (smbc_closefile_t *)smb_malloc(sizeof(smbc_closefile_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCloseFile: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->fid = fid;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbCloseFileThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCloseFile: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCloseFile failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbCloseFileSub(ctx, fid);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCloseFile failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbCloseFile exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbCloseFileThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_closefile_t	*th_arg = (smbc_closefile_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CLOSEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CLOSEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbCloseFileSub(th_arg->ctx, th_arg->fid);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CLOSEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_CLOSEFILE, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbCloseFileThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbCloseFileSub( smb_ctx* ctx, u_short fid )
#else
int NcsSmbCloseFile( smb_ctx* ctx, u_short fid, SMBC_CALLBACK* callback )
#endif
{
	int ret = SMBSUCCESS;
	if(!ctx) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileSub failed: ctx is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileSub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFileSub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input file id [%d]", (int)fid));

	if(!cli_close(ctx->cli, fid)) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCloseFile: cli_close failed: \n");
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
		syslog( LOG_ERR, "SMBC lib. cannot close file.(server=%s)", ctx->rmt.name);
	}
	return( ret );
}

/**********************************************************
 * NcsSmbReadFile
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbReadFile( smb_ctx* ctx, u_short fid, u_long offset, u_char* buf, int* size, SMBC_CALLBACK* callback )
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_readfile_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbReadFile(0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, fid, offset, buf, size, callback));
	if(!ctx || !buf || !size) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFile failed: ctx=%d, buf=%d, size=%d. \n", ctx, buf, size);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFile failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFile failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_readfile_t *)smb_malloc(sizeof(smbc_readfile_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbReadFile: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->fid = fid;
		th_arg->offset = offset;
		th_arg->buf = buf;
		th_arg->size = size;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbReadFileThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbReadFile: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbReadFile failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbReadFileSub(ctx, fid, offset, buf, size);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbReadFile failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbReadFile exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbReadFileThread(void *arg)
{
	int		ret = 0;
	int		status = 0;
	smbc_readfile_t *th_arg = (smbc_readfile_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_READFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_READFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbReadFileSub(th_arg->ctx, th_arg->fid, th_arg->offset, th_arg->buf, th_arg->size);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_READFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_READFILE, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbReadFileThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbReadFileSub( smb_ctx* ctx, u_short fid, u_long offset, u_char* buf, int* size )
#else
int NcsSmbReadFile( smb_ctx* ctx, u_short fid, u_long offset, u_char* buf, int* size, SMBC_CALLBACK* callback )
#endif
{
	int	ret_size = 0;
	int	ret = SMBSUCCESS;

	if(!ctx || !buf || !size) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileSub failed: ctx=%d, buf=%d, size=%d. \n", ctx, buf, size);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileSub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFileSub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input fid (%d)", (int)fid));

	ret_size = cli_read(ctx->cli, fid, buf, offset, *size);
	if(ret_size  <= 0) {
		printf("NcsSmbReadFile: cli_read failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbReadFile: cli_read failed: \n");
#endif
		syslog( LOG_ERR, "SMBC lib. cannot read file.(server=%s)", ctx->rmt.name);
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
	} else {
		*size = ret_size;
	}
	return( ret );
}

/**********************************************************
 * NcsSmbWriteFile
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbWriteFile( smb_ctx* ctx, u_short fid, u_long offset, u_long mode, u_char* buf, int* size, SMBC_CALLBACK* callback )
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_writefile_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbWriteFile(0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, fid, offset, mode, buf, size, callback));
	if(!ctx || !buf || !size) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFile failed: ctx=%d, buf=%d, size=%d. \n", ctx, buf, size);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFile cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFile cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_writefile_t *)smb_malloc(sizeof(smbc_writefile_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbWriteFile: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->fid = fid;
		th_arg->offset = offset;
		th_arg->mode = mode;
		th_arg->buf = buf;
		th_arg->size = size;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbWriteFileThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbWriteFile: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbWriteFile failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbWriteFileSub(ctx, fid, offset, mode, buf, size);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbWriteFile failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbWriteFile exit(0x%x)", ret));
	return(ret);
}
void *NcsSmbWriteFileThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_writefile_t	*th_arg = (smbc_writefile_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_WRITEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_WRITEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbWriteFileSub(th_arg->ctx, th_arg->fid, th_arg->offset, th_arg->mode, th_arg->buf, th_arg->size);


	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_WRITEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_WRITEFILE, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);

#ifdef DEBUG
	printf("[NcsSmbWriteFileThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbWriteFileSub( smb_ctx* ctx, u_short fid, u_long offset, u_long mode, u_char* buf, int* size )
#else
int NcsSmbWriteFile( smb_ctx* ctx, u_short fid, u_long offset, u_long mode, u_char* buf, int* size, SMBC_CALLBACK* callback )
#endif
{
	uint16	write_mode = 0;
	int	ret_size = 0;
	int	ret = SMBSUCCESS;
	int	mpx = 0;

	if(!ctx || !buf || !size) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileSub failed: ctx=%d, buf=%d, size=%d. \n", ctx, buf, size);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileThread failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFileThread failed: cnum is 0xffff. \n");
#endif
        	return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input fid (%d)", (int)fid));
	SMBC_LOG(("input data offset (0x%x)", offset));
	SMBC_LOG(("input data mode (0x%x)", mode));
	SMBC_LOG(("input data size (%d)", *size));

	smb_get_mpx(&mpx);

	ret_size = cli_write(ctx->cli, fid, write_mode, buf, offset, *size, mpx);
	if(ret_size  <= 0) {
		printf("NcsSmbWriteFile: cli_write: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbWriteFile: cli_write: \n");
#endif
		syslog( LOG_ERR, "SMBC lib. cannot write file.(server=%s)", ctx->rmt.name);
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
	} else {
		*size = ret_size;
	}
	return( ret );
}

/**********************************************************
 * NcsSmbDeleteFile
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbDeleteFile( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback)
{
	pthread_t		thread_id;
	int		ret = 0;
	int		status = 0;
	smbc_deletefile_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbDeleteFile enter(0x%x, 0x%x, 0x%x)", ctx, file, callback));
	if(!ctx || !file) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFile failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFile failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFile failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_deletefile_t *)smb_malloc(sizeof(smbc_deletefile_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteFile: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->file = file;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbDeleteFileThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteFile: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteFile failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbDeleteFileSub(ctx, file);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteFile failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbDeleteFile exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbDeleteFileThread( void* arg )
{
	int		ret = 0;
	int			status = 0;
	smbc_deletefile_t		*th_arg = (smbc_deletefile_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELETEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELETEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbDeleteFileSub(th_arg->ctx, th_arg->file);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELETEFILE, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_DELETEFILE, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbDeleteFileThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbDeleteFileSub( smb_ctx* ctx, smb_file* file )
#else
int NcsSmbDeleteFile( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback )
#endif
{
	u_short	char_code = file->character;
	int	ret = SMBSUCCESS;

	if(!ctx || !file) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileSub failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileSub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFileSub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input file name (%s)", file->name));
	set_smb_char_code(char_code);

	if(!cli_unlink(ctx->cli, file->name, &char_code)) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteFile: cli_unlink(%s) failed: \n", file->name);
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
		syslog( LOG_ERR, "SMBC lib. cannot delete file(server=%s, file=%s).", ctx->rmt.name, file->name);
	}
	return( ret );
}

/**********************************************************
 * NcsSmbMakeDirectory
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbMakeDirectory( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback)
{
	pthread_t	thread_id;
	int		ret = 0;
	int		status = 0;
	smbc_makedir_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbMakeDirectory(0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, file, finfo, callback));
	if(!ctx || !file || !finfo) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectory failed: ctx=%d, file=%d, finfo=%d. \n", ctx, file, finfo);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectory failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectory failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_makedir_t *)smb_malloc(sizeof(smbc_makedir_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectory: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->file = file;
		th_arg->finfo = finfo;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbMakeDirectoryThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectory: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectory failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbMakeDirectorySub(ctx, file, finfo);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectory failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbMakeDirectory exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbMakeDirectoryThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_makedir_t		*th_arg = (smbc_makedir_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectoryThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_MAKEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectoryThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_MAKEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbMakeDirectorySub(th_arg->ctx, th_arg->file, th_arg->finfo);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectoryThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_MAKEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_MAKEDIR, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbMakeDirectoryThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbMakeDirectorySub( smb_ctx* ctx, smb_file* file, smb_finfo* finfo )
#else
int NcsSmbMakeDirectory( smb_ctx* ctx, smb_file* file, smb_finfo* finfo, SMBC_CALLBACK* callback )
#endif
{
	const char	*fname = file->name;
	u_short		char_code  = file->character;
	int		ret = SMBSUCCESS;

	if(!ctx || !file || !finfo) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: ctx=%d, file=%d, finfo=%d. \n", ctx, file, finfo);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: cli is NULL. \n");
#endif
        	return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input file name (%s)", file->name));
	set_smb_char_code(char_code);

	if(file->flag & SMB_CREATE_SUBDIR) {
		char	dir[SMB_MAXFNAMELEN+1];
		char	*dirp = 0;

		bzero(dir, sizeof(dir));
		strncpy(dir, fname, SMB_MAXFNAMELEN);
		dirp = dir;
		if(dir[0] == '\\') {
			dirp ++;	/* skip 1st */
		}
		while(dirp) {
			dirp = strchr(dirp, '\\');
			if(!dirp) {
				break;
			}
			*dirp = 0;
			if(!cli_chkpath(ctx->cli, dir, &char_code)) {
				if(!cli_mkdir(ctx->cli, dir, &char_code)) {
					printf("NcsSmbOpenFile: cli_mkdir(%s) failed: \n", dir);
#ifndef SMBCD_USE
					smbc_mfprintf("NcsSmbMakeDirectorySub failed: cnum is 0xffff. \n");
#endif
					syslog( LOG_ERR, "SMBC lib. invalid derectory(server=%s, dir=%s).", ctx->rmt.name, dir);
					return smbc_errno(ctx->cli);
				}
			}
			*dirp++ = '\\';
		}
	}

	if(!cli_mkdir(ctx->cli, file->name, &char_code)) {
		printf("NcsSmbMakeDirectory: cli_mkdir(%s) failed: \n", file->name);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub: cli_mkdir(%s) failed: \n", file->name);
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
		syslog( LOG_ERR, "SMBC lib. cannot make derectory(server=%s, dir=%s).", ctx->rmt.name, file->name);
	}

	return( ret );
}

/**********************************************************
 * NcsSmbDeleteDirectory
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbDeleteDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback)
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_deledir_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbDeleteDirectory(0x%x, 0x%x, 0x%x) enter", ctx, file, callback));
	if(!ctx || !file) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbMakeDirectorySub failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_deledir_t *)smb_malloc(sizeof(smbc_deledir_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteDirectory: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->file = file;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbDeleteDirectoryThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDeleteDirectory: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectorySub failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbDeleteDirectorySub(ctx, file);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbMakeDirectorySub failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbDeleteDirectory exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbDeleteDirectoryThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_deledir_t		*th_arg = (smbc_deledir_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectoryThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectoryThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbDeleteDirectorySub(th_arg->ctx, th_arg->file);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectoryThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_DELEDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_DELEDIR, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbDeleteDirectoryThread] return\n");
#endif
	return(NULL);
}
#endif

#ifdef SMBC_USE_PTL
int NcsSmbDeleteDirectorySub( smb_ctx* ctx, smb_file* file )
#else
int NcsSmbDeleteDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback )
#endif
{
	u_short	char_code = file->character;
	int	ret = SMBSUCCESS;

	if(!ctx || !file) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	SMBC_LOG(("input file name (%s)", file->name));
	set_smb_char_code(char_code);

	if(!cli_chkpath(ctx->cli, file->name, &char_code)) {
		printf("NcsSmbDeleteDirectory: cli_chkpath(%s) failed\n", file->name);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub: cli_chkpath(%s) failed\n", file->name);
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			int class, condition, code;
			ret = smbc_errno(ctx->cli);

			class = (ret & 0xff000000) >> 24;
			condition = (ret & 0xff0000) >> 16;
			code = (ret & 0xffff);
			if(class == SMBERR_DOS && code == ERRbadfile) {
				ret = smbc_errcode(class, condition, SMBERR_BADPATH);
			}
		}
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_chkpath. \n");
#endif
		return ret;
	}

	if(file->flag & SMB_REMOVE_RECURSION) {
		char	path[SMB_MAXFNAMELEN+3];
		char	dir[SMB_MAXFNAMELEN+1];
		int	res = 0;
		int	listnum = 1;
		int	totalnum = 0;
		smb_dr_info	subdir[listnum];
		uint16 attribute = (aDIR | aSYSTEM | aHIDDEN);
		int	all_clear = 0;

		bzero(dir, sizeof(dir));
		bzero(path, sizeof(path));
		strncpy(dir, file->name, SMB_MAXFNAMELEN);

		while(!all_clear) {
			while(listnum > 0) { 
				strncpy(path, dir, SMB_MAXFNAMELEN);
				strncat(path, "\\*", 2);
				printf("delete %s\n", path);
				res = cli_unlink(ctx->cli, path, &char_code);
				if(!res) {
					int smberr = smbc_errno(ctx->cli);
					if((smberr & 0xffff) != ERRbadfile) {
						printf("NcsSmbDeleteDirectory: cli_unlink(%s) failed 0x%x:\n", path, smberr);
#ifndef SMBCD_USE
						smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_unlink. \n");
#endif
						syslog( LOG_ERR, "SMBC lib. cannot dellete derectory or file(server=%s, %s).", ctx->rmt.name, path);
						return smberr;
					}
				}

				listnum = 1;
				bzero(subdir, sizeof(subdir));
				res = cli_list(ctx->cli, path, attribute, &totalnum, &listnum, &subdir, NULL, &char_code);
				if(res < 0) {
					printf("NcsSmbDeleteDirectory: cli_list(%s) failed:\n", path);
#ifndef SMBCD_USE
					smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_list. \n");
#endif
					syslog( LOG_ERR, "SMBC lib. cannot get derectory list(server=%s, path=%s).", ctx->rmt.name, path);
					return smbc_errno(ctx->cli);
				}
				if(subdir[0].dr_name[0] != 0) {
					strncat(dir, "\\", 1);
					strncat(dir, subdir[0].dr_name, strlen(subdir[0].dr_name));
				}
			}
			printf("current dir = %s\n", dir);
			printf("file->name = %s\n", file->name);

			while(strcmp(dir, file->name) != 0) {
				char *p = 0;
	
				printf("deldir %s\n", dir);
				if(!cli_rmdir(ctx->cli, dir, &char_code)) {
					int smberr = smbc_errno(ctx->cli);
					if((smberr & 0xffff) != ERRdirnotempty) {
						printf("NcsSmbDeleteDirectory: cli_rmdir(%s) failed 0x%x:\n", dir, smberr);
#ifndef SMBCD_USE
						smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_rmdir. \n");
#endif
						syslog( LOG_ERR, "SMBC lib. cannot dellete derectory(server=%s, dir=%s).", ctx->rmt.name, dir);
						return smberr;
					}
				}
	
				p = smb_strrchr(dir, '\\');
				*p = 0;
			}
			strncpy(path, file->name, SMB_MAXFNAMELEN);
			strncat(path, "\\*", 2);
			listnum = 1;
			res = cli_list(ctx->cli, path, attribute, &totalnum, &listnum, &subdir, NULL, &char_code);
			if(res < 0) {
				printf("NcsSmbDeleteDirectory: cli_list(%s) failed:\n", path);
#ifndef SMBCD_USE
				smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_list. \n");
#endif
				syslog( LOG_ERR, "SMBC lib. cannot get derectory list(server=%s, path=%s).", ctx->rmt.name, path);
				return smbc_errno(ctx->cli);
			}
			if(listnum == 0) {
				all_clear = 1;
			}
		}
	}

	if(!cli_rmdir(ctx->cli, file->name, &char_code)) {
		printf("NcsSmbDeleteDirectory: cli_rmdir(%s) failed: \n", file->name);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDeleteDirectorySub failed: cli_rmdir. \n");
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			ret = smbc_errno(ctx->cli);
		}
		syslog( LOG_ERR, "SMBC lib. cannot dellete derectory(server=%s, dir=%s).", ctx->rmt.name, file->name);
	}

	return( ret );
}

/**********************************************************
 * NcsSmbCheckDirectory
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbCheckDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback)
{
	pthread_t	thread_id;
	int		ret = 0;
	int		status = 0;
	smbc_chkdir_t	*th_arg = NULL;

	SMBC_LOG(("NcsSmbCheckDirectory(0x%x, 0x%x, 0x%x) enter", ctx, file, callback));
	if(!ctx || !file) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectory failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return ret;
	}
	if(!ctx->cli) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectory failed: cli is NULL. \n");
#endif
		return ret;
	}
	if(ctx->cli->cnum == 0xffff) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
		SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectory failed: cnum is 0xffff. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_chkdir_t *)smb_malloc(sizeof(smbc_chkdir_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCheckDirectory: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->file = file;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbCheckDirectoryThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCheckDirectory: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbCheckDirectory failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbCheckDirectorySub(ctx, file);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
			printf("NcsSmbCheckDirectory: pthread_mutex_unlock() error[%d].",status);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbCheckDirectory exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbCheckDirectoryThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_chkdir_t		*th_arg = (smbc_chkdir_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectoryThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CHKDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectoryThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CHKDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbCheckDirectorySub(th_arg->ctx, th_arg->file);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectoryThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_CHKDIR, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_CHKDIR, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbCheckDirectoryThread] return\n");
#endif
	return(NULL);
}
#endif /* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbCheckDirectorySub( smb_ctx* ctx, smb_file* file )
#else
int NcsSmbCheckDirectory( smb_ctx* ctx, smb_file* file, SMBC_CALLBACK* callback )
#endif
{
	u_short	char_code = file->character;
	int	ret = SMBSUCCESS;

	if(!ctx || !file) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectorySub failed: ctx=%d, file=%d. \n", ctx, file);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(!ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectorySub failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	if(ctx->cli->cnum == 0xffff) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectorySub failed: cnum is 0xffff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_CMDSEQ);
	}
	set_smb_char_code(char_code);

	if(!cli_chkpath(ctx->cli, file->name, &char_code)) {
		printf("NcsSmbCheckDirectory: cli_chkpath(%s) failed: \n", file->name);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbCheckDirectorySub failed: cli_chkpath. \n");
#endif
		if(ctx->cli->fd == -1) {
			ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			int class, condition, code;
			ret = smbc_errno(ctx->cli);

			class = (ret & 0xff000000) >> 24;
			condition = (ret & 0xff0000) >> 16;
			code = (ret & 0xffff);
			if(class == SMBERR_DOS && code == ERRbadfile) {
				ret = smbc_errcode(class, condition, SMBERR_BADPATH);
			}
		}
	}
	return( ret );
}

/**********************************************************
 * NcsSmbGetGroupList
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbGetGroupList( smb_ctx* ctx, u_short max_list_num, smb_gp_list* group_list, SMBC_CALLBACK* callback )
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_getgrplist_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbGetGroupList(0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, max_list_num, group_list, callback));
	if(!ctx || !group_list) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetGroupList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupList failed: ctx=%d, group_list=%d. \n", ctx, group_list);
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_getgrplist_t *)smb_malloc(sizeof(smbc_getgrplist_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetGroupList: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->max_list_num = max_list_num;
		th_arg->group_list = group_list;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbGetGroupListThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetGroupList: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetGroupList failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetGroupList exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbGetGroupListSub(ctx, max_list_num, group_list);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetGroupList failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetGroupList exit(0x%x)", ret));
			return(ret);
		}
	}

	SMBC_LOG(("NcsSmbGetGroupList exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbGetGroupListThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_getgrplist_t		*th_arg = (smbc_getgrplist_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETGRPLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETGRPLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbGetGroupListSub(th_arg->ctx, th_arg->max_list_num, th_arg->group_list);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETGRPLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_GETGRPLST, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbGetGroupListThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbGetGroupListSub( smb_ctx* ctx, u_short max_list_num, smb_gp_list* group_list )
#else
int NcsSmbGetGroupList( smb_ctx* ctx, u_short max_list_num, smb_gp_list* group_list, SMBC_CALLBACK* callback )
#endif
{
	smb_vc	vc;
	int	res = 0;
	smb_gp_info	*group_info = 0;
	int	listnum = max_list_num;
	int	totalnum = 0;
	smb_ctx	*browse_ctx = 0;
	int	smberr = SMBSUCCESS;
	u_short	char_code = get_smb_default_char_code();
	int	i;

	if(!ctx || !group_list) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub failed: ctx=%d, group_list=%d. \n", ctx, group_list);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	group_list->entries = 0;

	group_info = (smb_gp_info *)smb_malloc(sizeof(smb_gp_info)*max_list_num);
	if(!group_info) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	bzero(group_info, sizeof(smb_gp_info)*max_list_num);
	bzero(&vc, sizeof(vc));

	res = NcsSmbOpen(&browse_ctx);
	if(res != SMBSUCCESS) {
		printf("get group list failed:\n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub failed: NcsSmbOpen. \n");
#endif
		SAFE_FREE(group_info);
		return res;
	}

	smb_get_mygroup(vc.group);
	res = cli_SmbOpenPipe(browse_ctx, &vc, NULL);
	if(res != SMBSUCCESS) {
		printf("get group list failed:\n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub failed: cli_SmbOpenPipe. \n");
#endif
		smberr = res;
		SAFE_FREE(group_info);
		return res;
	}

	if(cli_NetServerEnum(browse_ctx->cli, (char *)&vc.group, 
			SV_TYPE_DOMAIN_ENUM, &totalnum, &listnum, group_info, NULL, &char_code) < 0) {
		printf("get group list failed:\n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub failed: cli_NetServerEnum. \n");
#endif
		if(browse_ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			smberr = smbc_errno(browse_ctx->cli);
		}
		NcsSmbClose(browse_ctx);
		SAFE_FREE(group_info);
		syslog( LOG_ERR, "SMBC lib. cannot get group list.");
		return smberr;
	}

	res = cli_SmbClosePipe(browse_ctx, NULL);
	if(res != SMBSUCCESS) {
		printf("get group list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetGroupListSub failed: cli_SmbClosePipe. \n");
#endif
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(group_info);
		return smberr;
	}
	group_list->entries = listnum;
	group_list->totalnum = totalnum;
	for(i = 0; i < listnum; i++) {
		group_list->group_info[i] = group_info[i];
	}
	NcsSmbClose(browse_ctx);
	SAFE_FREE(group_info);

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbGetServerList
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbGetServerList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sv_list* server_list, SMBC_CALLBACK* callback )
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_getsrvlist_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbGetServerList(0x%x, 0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, vcp, max_list_num, server_list, callback));
	if(!ctx || !vcp || !server_list) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetServerList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerList failed: ctx=%d, vcp=%d, server_list=%d. \n", ctx, vcp, server_list);
#endif
		return ret;
	}
	if(vcp->group[0] == NULL) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetServerList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerList failed: group is NULL. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_getsrvlist_t *)smb_malloc(sizeof(smbc_getsrvlist_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerList: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->max_list_num = max_list_num;
		th_arg->server_list = server_list;
		th_arg->callback = callback;
		status = pthread_create(&thread_id, NULL, NcsSmbGetServerListThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerList: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerList failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetServerList exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbGetServerListSub(ctx, vcp, max_list_num, server_list);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerList failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetServerList exit(0x%x)", ret));
			return(ret);
		}
	}

	SMBC_LOG(("NcsSmbGetServerList exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbGetServerListThread( void* arg )
{
	int			ret = 0;
	int			status = 0;
	smbc_getsrvlist_t		*th_arg = (smbc_getsrvlist_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbGetServerListSub(th_arg->ctx, th_arg->vcp, th_arg->max_list_num, th_arg->server_list );

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	th_arg->callback->func(SMBC_GETSRVLST, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbGetServerListThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbGetServerListSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sv_list* server_list )
#else
int NcsSmbGetServerList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sv_list* server_list, SMBC_CALLBACK* callback )
#endif
{
	int	res = 0;
	smb_sv_info	*server_info = 0;
	int	listnum = max_list_num;
	int	totalnum = 0;
	smb_ctx	*browse_ctx = 0;
	int	smberr = SMBSUCCESS;
	u_short	char_code = vcp->character;
	char	group[SMB_MAXSRVNAMELEN+1];
	int	i;

	if(!ctx || !vcp || !server_list) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: ctx=%d, vcp=%d, server_list=%d. \n", ctx, vcp, server_list);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(vcp->group[0] == NULL) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: group is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	SMBC_LOG(("input group name [%s]", vcp->group));
	set_smb_char_code(char_code);
	server_list->entries = 0;

	server_info = (smb_sv_info *)smb_malloc(sizeof(smb_sv_info)*max_list_num);
	if(!server_info) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	bzero(server_info, sizeof(smb_sv_info)*max_list_num);
	res = NcsSmbOpen(&browse_ctx);
	if(res != SMBSUCCESS) {
		SAFE_FREE(server_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: NcsSmbOpen. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	bzero(group, sizeof(group));
	if(vcp->group[0] != 0) {
		strncpy(group, vcp->group, SMB_MAXSRVNAMELEN);
	} else {
		smb_get_mygroup(group);
	}
	res = cli_SmbOpenPipe(browse_ctx, vcp, NULL);
	if(res != SMBSUCCESS) {
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(server_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: cli_SmbOpenPipe. \n");
#endif
		return res;
	}
	
	if(cli_NetServerEnum(browse_ctx->cli, (char *)group, 
			SV_TYPE_ALL, &totalnum, &listnum, server_info, NULL, &char_code) < 0 ) {
		printf("get server list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: cli_NetServerEnum. \n");
#endif
		if(browse_ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			smberr = smbc_errno(browse_ctx->cli);
		}
		NcsSmbClose(browse_ctx);
		SAFE_FREE(server_info);
		syslog( LOG_ERR, "SMBC lib. cannot get server list(%s).", group);
		return smberr;
	}

	res = cli_SmbClosePipe(browse_ctx, NULL);
	if(res != SMBSUCCESS) {
		printf("get server list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerListSub failed: cli_SmbClosePipe. \n");
#endif
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(server_info);
		return smberr;
	}
	server_list->entries = listnum;
	server_list->totalnum = totalnum;
	for(i = 0; i < listnum; i++) {
		server_list->server_info[i] = server_info[i];
	}
	NcsSmbClose(browse_ctx);
	SAFE_FREE(server_info);

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbGetShareList
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbGetShareList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sh_list* share_list, SMBC_CALLBACK* callback )
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_getshrlist_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbGetShareList(0x%x, 0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, vcp, max_list_num, share_list, callback));
	if(!ctx || !vcp || !share_list) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetShareList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareList failed: ctx=%d, vcp=%d, share_list=%d. \n", ctx,vcp,share_list);
#endif
		return ret;
	}
	if(vcp->server[0] == NULL) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetShareList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareList failed: server is NULL. \n");
#endif
		return ret;
	}

	if(callback){
		th_arg = (smbc_getshrlist_t *)smb_malloc(sizeof(smbc_getshrlist_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetShareList: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->max_list_num = max_list_num;
		th_arg->share_list = share_list;
		th_arg->callback = callback;

		status = pthread_create(&thread_id, NULL, NcsSmbGetShareListThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetShareList: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetShareList failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetShareList exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbGetShareListSub(ctx, vcp, max_list_num, share_list);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetShareList failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetShareList exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbGetShareList exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbGetShareListThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_getshrlist_t	*th_arg = (smbc_getshrlist_t *)arg;

	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSHRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSHRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	ret = NcsSmbGetShareListSub(th_arg->ctx, th_arg->vcp, th_arg->max_list_num, th_arg->share_list);

	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSHRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListThread failed: pthread_mutex_unlock. \n");
#endif
		return(NULL);
	}

	th_arg->callback->func(SMBC_GETSHRLST, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbGetShareListThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbGetShareListSub( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sh_list* share_list )
#else
int NcsSmbGetShareList( smb_ctx* ctx, smb_vc* vcp, u_short max_list_num, smb_sh_list* share_list, SMBC_CALLBACK* callback)
#endif
{
	int	res = 0;
	smb_sh_info	*share_info = 0;
	int	listnum = max_list_num;
	int	totalnum = 0;
	smb_ctx	*browse_ctx = 0;
	int	smberr = SMBSUCCESS;
	u_short	char_code = vcp->character;
	int	i;

	if(!ctx || !vcp || !share_list) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: ctx=%d, vcp=%d, share_list=%d. \n", ctx,vcp,share_list);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(vcp->server[0] == NULL) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: server is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	share_list->entries = 0;
	SMBC_LOG(("input group name [%s]", vcp->group));
	SMBC_LOG(("input server name [%s]", vcp->server));
	set_smb_char_code(char_code);

	share_info = (smb_sh_info *)smb_malloc(sizeof(smb_sh_info)*max_list_num);
	if(!share_info) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	bzero(share_info, sizeof(smb_sh_info)*max_list_num);
	res = NcsSmbOpen(&browse_ctx);
	if(res != SMBSUCCESS) {
		SAFE_FREE(share_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: NcsSmbOpen. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	res = cli_SmbOpenPipe(browse_ctx, vcp, NULL);
	if(res != SMBSUCCESS) {
		printf("get share list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: cli_SmbOpenPipe. \n");
#endif
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(share_info);
		return smberr;
	}

	if(cli_RNetShareEnum(browse_ctx->cli, &totalnum, &listnum, share_info, NULL, &char_code) < 0) {
		printf("get share list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: cli_RNetShareEnum. \n");
#endif
		if(browse_ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			smberr = smbc_errno(browse_ctx->cli);
		}
		NcsSmbClose(browse_ctx);
		syslog( LOG_ERR, "SMBC lib. cannot get share list(%s).", vcp->server);
		SAFE_FREE(share_info);
		return smberr;
	}

	res = cli_SmbClosePipe(browse_ctx, NULL);
	if(res != SMBSUCCESS) {
		printf("get share list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetShareListSub failed: cli_SmbClosePipe. \n");
#endif
		smberr = res;
		NcsSmbClose(browse_ctx);
		SAFE_FREE(share_info);
		return smberr;
	}
	share_list->entries = listnum;
	share_list->totalnum = totalnum;
	for(i = 0; i < listnum; i++) { 
		share_list->share_info[i] = share_info[i];
	}
	NcsSmbClose(browse_ctx);
	SAFE_FREE(share_info);

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbDirectoryList
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbDirectoryList( smb_ctx* ctx, smb_vc* vcp, const char* directory, u_short max_list_num, smb_dr_list* dr_list, SMBC_CALLBACK* callback)
{
	pthread_t		thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_getdirlist_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbGetShareList(0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, vcp, directory, max_list_num, dr_list, callback));
	if(!ctx || !vcp || !directory || !dr_list) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbDirectoryList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryList failed: ctx=%d, vcp=%d, directory=%d, dr_list=%d. \n", ctx,vcp,directory,dr_list);
#endif
		return ret;
	}
	if(vcp->server[0] == NULL) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbDirectoryList exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryList failed: server is NULL. \n");
#endif
		return ret;
	}

	if(callback){
		/*
		 * set arg for pthread_create
		 */
		th_arg = (smbc_getdirlist_t *)smb_malloc(sizeof(smbc_getdirlist_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDirectoryList: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->directory = directory;
		th_arg->max_list_num = max_list_num;
		th_arg->dr_list= dr_list;
		th_arg->callback = callback;

		/*
		 * pthread_create
		 */
		status = pthread_create(&thread_id, NULL, NcsSmbDirectoryListThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDirectoryList: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDirectoryList failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDirectoryList exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbDirectoryListSub(ctx, vcp, directory, max_list_num, dr_list);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbDirectoryList failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbDirectoryList exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbDirectoryList exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbDirectoryListThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_getdirlist_t		*th_arg = (smbc_getdirlist_t *)arg;

	/*
	 * pthread_detach
	 */
	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETDIRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * mutex lock
	 */
	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETDIRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec Sub
	 */
	ret = NcsSmbDirectoryListSub(th_arg->ctx, th_arg->vcp, th_arg->directory, th_arg->max_list_num, th_arg->dr_list);

	/*
	 * mutex unlock
	 */
	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETDIRLST, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec callback
	 */
	th_arg->callback->func(SMBC_GETDIRLST, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbDirectoryListThread] return\n");
#endif
	return(NULL);
}
#endif	/* SMBC_USE_PTL */

#ifdef SMBC_USE_PTL
int NcsSmbDirectoryListSub( smb_ctx* ctx, smb_vc* vcp, const char* directory, u_short max_list_num, smb_dr_list* dr_list )
#else
int NcsSmbDirectoryList( smb_ctx* ctx, smb_vc* vcp, const char* directory, u_short max_list_num, smb_dr_list* dr_list, SMBC_CALLBACK* callback )
#endif
{
	int	res = 0;
	smb_dr_info	*directory_info = 0;
	int	listnum = max_list_num;
	int	totalnum = 0;
	smb_ctx	*browse_ctx = 0;
	uint16 attribute = (aDIR | aSYSTEM | aHIDDEN);
	u_short	char_code = vcp->character;
	int	i;
	BOOL	is_multi_last_char = FALSE;
	char	path[SMB_MAXFNAMELEN+3];
	char	*p = 0;
	int	end_path = 0;
	int	len = 0;
	int	smberr = SMBSUCCESS;
	
	if(!ctx || !vcp || !directory || !dr_list) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: ctx=%d, vcp=%d, directory=%d, dr_list=%d. \n", ctx,vcp,directory,dr_list);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	if(vcp->server[0] == NULL) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: server is NULL. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	dr_list->entries = 0;
	SMBC_LOG(("input server name [%s]", vcp->server));
	SMBC_LOG(("input share name [%s]", vcp->share));
	SMBC_LOG(("input directory name [%s]", directory));
	set_smb_char_code(char_code);

	directory_info = (smb_dr_info *)smb_malloc(sizeof(smb_dr_info)*max_list_num);
	if(!directory_info) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub: smb_malloc() error.\n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	bzero(directory_info, sizeof(smb_dr_info)*max_list_num);
	res = NcsSmbOpen(&browse_ctx);
	if(res != SMBSUCCESS) {
		SAFE_FREE(directory_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: NcsSmbOpen. \n");
#endif
		return res;
	}

	res = NcsSmbConnect(browse_ctx, vcp, NULL);
	if(res != SMBSUCCESS) {
		syslog( LOG_ERR, "SMBC lib. cannot get directory list(server=%s, share=%s).", vcp->server, vcp->share);
		NcsSmbClose(browse_ctx);
		SAFE_FREE(directory_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: NcsSmbConnect. \n");
#endif
		return res;
	}

	bzero(path, sizeof(path));
	strncpy(path, directory, SMB_MAXFNAMELEN);
	/*
	 * fBNgXg擾Ƃ́A̍Ō"\*"
	 * T[oɃNGXgB
	 * <information>
	 * "\\\\"̓T[oɂ"\"ƓƂ݂Ȃ邪A
	 * "****""*"Ƃ͋ʂ悤łB
	 */
	end_path = strlen(path);
	p = path;
	len = 0;
	is_multi_last_char = FALSE;
	while(len < end_path) {
		int tmplen = 0;

		/* oCg擪oCg */
#ifdef SMBCD_TXTENC_USE
		tmplen = txtencGetSizeOfMbChar(char_code, *p);
#else
		tmplen = 1;
#endif
		if(tmplen == 0) {
			/* ̐擪oCgƂĕsȒl̏ꍇioOj */
			/* 1oCgƂĈ */
			tmplen = 1;
			is_multi_last_char = FALSE;
		}
		else if(tmplen == 1) {	/* 1oCg */
			is_multi_last_char = FALSE;
		} else {		/* oCg */
			is_multi_last_char = TRUE;
		}
		p += tmplen;
		len += tmplen;
	}
	if(!is_multi_last_char) {	/* Ō̕oCgł͂Ȃ */
		if( end_path >= 2 && 
		    (path[end_path-2] == '\\' && path[end_path-1] == '*')) {
			/* Ō̕"\*"̂Ƃ͉Ȃ */
		} else if(end_path >= 1 && path[end_path-1] == '\\') {
			/* Ō̕'\'ŏIƂ'*'t */
			path[end_path] = '*';
		} else {
			/* L̂łȂƂ"\*"t */
			path[end_path] = '\\';
			path[end_path+1] = '*';
		}
	} else {
		/* Ō̕oCgŏIĂꍇA"\*"t */
		path[end_path] = '\\';
		path[end_path+1] = '*';
	}

	res = cli_list(browse_ctx->cli, path, attribute, &totalnum, &listnum, directory_info, NULL, &char_code);
	if(res < 0) {
		if(browse_ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else { 
			smberr = smbc_errno(browse_ctx->cli);
		}
		if(smberr == ((SMBERR_TEMPORARY << 24) | (ERRDOS << 16) | SMBERR_BADFILE)) {
			/*
			 * t@C܂̓fBNgȂƂSMBERR_BADFILEƂȂ
			 * fBXÑ[gLAɃt@CȂƂ
			 * ̃G[ԂĂB
			 * jtH[}bg̃tbs[fBXNiA:)LꍇȂ
			 * ̏ꍇɂ̓pX`FbNApX݂ꍇɂ
			 * t@CGgOƂĐI
			 * 
			 * t@Cȏ゠ꍇA邢͋L|Cg[gȊO
			 * i"."".."݂̂ŁjG[ɂ͂ȂȂ
			 */
			if(*directory == NULL || strcmp(directory, "\\") == 0) {
				if(cli_chkpath(browse_ctx->cli, directory, &char_code)) {
					dr_list->entries = 0;	/*  */
					NcsSmbClose(browse_ctx);
					SAFE_FREE(directory_info);
#ifndef SMBCD_USE
					smbc_mfprintf("NcsSmbDirectoryListSub failed: cli_chkpath. \n");
#endif
					return SMBSUCCESS;
				}
			}
		}
		printf("get directory list failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: cli_list. \n");
#endif
		syslog( LOG_ERR, "SMBC lib. cannot get directory list(server=%s, share=%s).", vcp->server, vcp->share);
		NcsSmbClose(browse_ctx);
		SAFE_FREE(directory_info);
		return smberr;
	}

	dr_list->entries = listnum;
	dr_list->totalnum = totalnum;
	for(i = 0; i < listnum; i++) {
		dr_list->directory_info[i] = directory_info[i];
	}
	res = NcsSmbDisconnect(browse_ctx, NULL);
	if(res != SMBSUCCESS) {
		smberr = smbc_errno(browse_ctx->cli); 
		NcsSmbClose(browse_ctx);
		SAFE_FREE(directory_info);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbDirectoryListSub failed: NcsSmbDisconnect. \n");
#endif
		return smberr;
	}
	NcsSmbClose(browse_ctx);
	SAFE_FREE(directory_info);

	return( SMBSUCCESS );
}

/**********************************************************
 * NcsSmbGetServerCapability
 *
 *
 *********************************************************/
#ifdef SMBC_USE_PTL
int NcsSmbGetServerCapability( smb_ctx* ctx, smb_vc* vcp, smb_server_cap* capability, SMBC_CALLBACK* callback)
{
	pthread_t	thread_id;
	int			ret = 0;
	int			status = 0;
	smbc_getsrvcap_t		*th_arg = NULL;

	SMBC_LOG(("NcsSmbGetServerCapability(0x%x, 0x%x, 0x%x, 0x%x) enter", ctx, vcp, capability, callback));
	if(!ctx || !vcp || !capability) {
		ret = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
		SMBC_LOG(("NcsSmbGetServerCapability exit(0x%x)", ret));
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapability failed: ctx=%d, vcp=%d, capability=%d. \n", ctx,vcp,capability);
#endif
		return ret;
	}

	if(callback){
		/*
		 * set arg for pthread_create
		 */
		th_arg = (smbc_getsrvcap_t *)smb_malloc(sizeof(smbc_getsrvcap_t));
		if(!th_arg) {
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapability: smb_malloc() error.\n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_IFERR);
		}
		th_arg->ctx = ctx;
		th_arg->vcp = vcp;
		th_arg->capability = capability;
		th_arg->callback = callback;
		/*
		 * pthread_create
		 */
		status = pthread_create(&thread_id, NULL, NcsSmbGetServerCapabilityThread, th_arg);
		if(status != 0){
			printf("smbc_lib: pthread_create() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapability: pthread_create() error[%d].",status);
#endif
			SAFE_FREE(th_arg);
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}else{
			ret = SMBSUCCESS;
		}
	}else{
		status = pthread_mutex_trylock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapability failed: pthread_mutex_trylock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetServerCapability exit(0x%x)", ret));
			return(ret);
		}
		
		ret = NcsSmbGetServerCapabilitySub(ctx, vcp, capability);
		
		status = pthread_mutex_unlock(&ctx->mutex);
		if(status != 0){
			printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapability failed: pthread_mutex_unlock. \n");
#endif
			ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
			SMBC_LOG(("NcsSmbGetServerCapability exit(0x%x)", ret));
			return(ret);
		}
	}
	SMBC_LOG(("NcsSmbGetServerCapability exit(0x%x)", ret));
	return(ret);
}

void *NcsSmbGetServerCapabilityThread( void* arg )
{
	int		ret = 0;
	int		status = 0;
	smbc_getsrvcap_t		*th_arg = (smbc_getsrvcap_t *)arg;

	/*
	 * pthread detach
	 */
	status = pthread_detach(pthread_self());
	if(status != 0){
		printf("smbc_lib: pthread_detach() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: pthread_detach. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVCAP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * mutex lock
	 */
	status = pthread_mutex_trylock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_trylock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: pthread_mutex_trylock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVCAP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec Sub
	 */
	ret = NcsSmbGetServerCapabilitySub(th_arg->ctx, th_arg->vcp, th_arg->capability);

	/*
	 * mutex unlock
	 */
	status = pthread_mutex_unlock(&th_arg->ctx->mutex);
	if(status != 0){
		printf("smbc_lib: pthread_mutex_unlock() error[%d].",status);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: pthread_mutex_unlock. \n");
#endif
		ret = smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		th_arg->callback->func(SMBC_GETSRVCAP, ret, th_arg->callback->param);
		SAFE_FREE(th_arg);
		return(NULL);
	}

	/*
	 * exec callback func
	 */
	th_arg->callback->func(SMBC_GETSRVCAP, ret, th_arg->callback->param);
	SAFE_FREE(th_arg);
#ifdef DEBUG
	printf("[NcsSmbgetServerCapabilityThread] return\n");
#endif
	return(NULL);
}
#endif

#ifdef SMBC_USE_PTL
int NcsSmbGetServerCapabilitySub( smb_ctx* ctx, smb_vc* vcp, smb_server_cap* capability )
#else
int NcsSmbGetServerCapability( smb_ctx* ctx, smb_vc* vcp, smb_server_cap* capability, SMBC_CALLBACK* callback )
#endif
{
	int	res;

	char		server[SMB_MAXHOSTNAMELEN+1];
	smb_ctx		*local_ctx = 0;
	int	smberr = SMBSUCCESS;
	int	name_type = 0;
	u_short	char_code = vcp->character;
	int	NcsSmbGetServerCapability_negprot_fail = 0;
	u_long		target_ip[SMB_MAX_IP];
	int		num = SMB_MAX_IP;

	if(!ctx || !vcp || !capability) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilitySub failed: ctx=%d, vcp=%d, capability=%d. \n", ctx,vcp,capability);
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_PARAM);
	}
	SMBC_LOG(("input server name [%s]", vcp->server));

	res = NcsSmbOpen(&local_ctx);
	if(res != SMBSUCCESS) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: NcsSmbOpen. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}

	local_ctx->cli = cli_initialise(NULL);
	if(!local_ctx->cli) {
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: cli is NULL. \n");
#endif
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	set_smb_char_code(char_code);

	strncpy(server, vcp->server, SMB_MAXHOSTNAMELEN);

	bzero(&local_ctx->rmt, sizeof(local_ctx->rmt));
	local_ctx->rmt.suffix = SUFFIX_SERV;
	name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);

	if(!name_type) {
		cli_shutdown(local_ctx->cli);
		local_ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. cannot resolve server name(server=%s).",  server);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: name_type is 0. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	} else if(name_type != 1) {
		int res = 0;
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;

		int ii = 0;

		/*
		 * ftHgł"*SMBSERVER"w肷B
		 * ^[QbgPCWindowsNT4.0 SP6aȍ~OSł
		 * ̓ꖼŐڑ\
		 */
		strncpy(server, SMBSERVER, SMB_MAXSRVNAMELEN);
		for(ii = 0; ii < num; ii++) {
			local_ctx->rmt.naddr = target_ip[ii];
			memset(&nn, 0, sizeof(nn));
			memset(&ns, 0, sizeof(ns));
			res = smb_getnodestatus(&local_ctx->rmt, &nn, &ns, SUFFIX_SERV);
			if(res == LIBNBT_ERR_NOERR) {
				SMBC_LOG(("smb_getnodestatus: name = %s", nn.name)); 
				strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
				if(ii > 0) {
					/* ڑAhX̑1ɂĂ */
					u_long tmp_ip;
					tmp_ip = target_ip[0];
					target_ip[0] = target_ip[ii];
					target_ip[ii] = tmp_ip;
				}
				break;
			}
		}
	}
	strncpy(local_ctx->rmt.name, server, SMB_MAXSRVNAMELEN);

NcsSmbGetServerCapability_retry:
	res = nbt_connectTimeout(&local_ctx->ss_client, &local_ctx->rmt, &local_ctx->local, target_ip, num);
	if(res < 0) {
		printf("failed nbt_connectTimeout (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: nbt_connectTimeout. \n");
#endif
		NcsSmbClose(local_ctx);
		syslog( LOG_ERR, "SMBC lib. failed to connect server(server=%s).", local_ctx->rmt.name );
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}

	res = nbt_getfd(local_ctx->ss_client);
	if(res < 0) {
		printf("failed nbt_getfd (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: nbt_getfd. \n");
#endif
		NcsSmbClose(local_ctx);
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	local_ctx->cli->fd = res;
	local_ctx->cli->port = 139;
	fstrcpy(local_ctx->cli->desthost, server);
	fstrcpy(local_ctx->cli->called.name, server);
	local_ctx->cli->called.name_type = SUFFIX_SERV;
	local_ctx->cli->dest_ip.s_addr = htonl(local_ctx->rmt.naddr);

	if(!cli_negprot(local_ctx->cli, &char_code)) {
		if(local_ctx->cli->fd < 0) {
			if(NcsSmbGetServerCapability_negprot_fail++ < 5) {
				nbt_close(local_ctx->ss_client);
				usleep(smb_random()%1000+100);
				goto NcsSmbGetServerCapability_retry;
			}
			printf("cli_negprot fail %d\n", ERRNO);
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: cli_negprot. \n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}
		printf("retry NULL domain\n");
		local_ctx->cli->domain[0] = NULL;
		if(!cli_negprot(local_ctx->cli, &char_code)) {
			if(local_ctx->cli->fd < 0) {
				if(NcsSmbGetServerCapability_negprot_fail++ < 5) {
					nbt_close(local_ctx->ss_client);
					usleep(smb_random()%1000+100);
					goto NcsSmbGetServerCapability_retry;
				}
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else { 
				smberr = smbc_errno(local_ctx->cli);
			}
			NcsSmbClose(local_ctx);
			syslog( LOG_ERR, "SMBC lib. failed to negotiate.");
#ifndef SMBCD_USE
			smbc_mfprintf("NcsSmbGetServerCapabilityThread failed: cli_negprot(retry). \n");
#endif
			return smberr;
		}
	}
	capability->flag = local_ctx->cli->capabilities; 
	capability->sec_mode = local_ctx->cli->sec_mode; 
	/*
	 * 2002/11/21 terao add
	 */
	if(local_ctx->cli->server_domain) {
		strncpy(vcp->group, local_ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
	}

	NcsSmbClose(local_ctx);
	return( SMBSUCCESS );
}


// -------------------------------------------------------------
// internal library
// -------------------------------------------------------------

int cli_SmbOpenPipe( smb_ctx* ctx, smb_vc* vcp, SMBC_PROC func)
{
	int	res = 0;

	char		myname[SMB_MAXSRVNAMELEN+1];
        char            group[MAX(SMB_MAXSRVNAMELEN,SMB_MAXHOSTNAMELEN)+1];
	char		server[SMB_MAXSRVNAMELEN+1];
	char		username[SMB_MAXUSERNAMELEN+1];
	char		password[SMB_MAXPASSWORDLEN+1];
	u_short		char_code = vcp->character;
	int		name_type = 0;
	int		cli_SmbOpenPipe_negprot_fail = 0;
	u_long		target_ip[SMB_MAX_IP];
	int		num = SMB_MAX_IP;

	if(!ctx->cli) {
		ctx->cli = cli_initialise(NULL);
		if(!ctx->cli) {
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: cli_initialise. \n");
#endif
			return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
		}
	}
	set_smb_char_code(char_code);

	bzero(myname, sizeof(myname));
	bzero(group, sizeof(group));
	bzero(server, sizeof(server));
	bzero(username, sizeof(username));
	bzero(password, sizeof(password));

	if(vcp->group[0] != 0) {
		strncpy(group, vcp->group, sizeof(group)-1);
	} else {
		smb_get_mygroup(group);
	}
	if(vcp->server[0] != 0) {
		strncpy(server, (char *)vcp->server, SMB_MAXSRVNAMELEN);
		ctx->rmt.suffix = SUFFIX_SERV;
		name_type = smb_gethostbyname(server, SUFFIX_SERV, target_ip, &num);
		if(!name_type) {
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			syslog( LOG_ERR, "SMBC lib. cannot resolve server name(%s).",  server);
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: smb_gethostbyname(%s)[0x20]. \n", server);
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else if(name_type != 1) {
			int res = 0;
			NBTNodeName nn;
			NBTNodeStatistics ns;
			int findnum = 0;
			int ii = 0;

			/*
			 * ftHgł"*SMBSERVER"w肷B
			 * ^[QbgPCWindowsNT4.0 SP6aȍ~OSł
			 * ̓ꖼŐڑ\
			 */
			strncpy(server, SMBSERVER, SMB_MAXSRVNAMELEN);
			for(ii = 0; ii < num; ii++) {
				ctx->rmt.naddr = target_ip[ii];
				memset(&nn, 0, sizeof(nn));
				memset(&ns, 0, sizeof(ns));
				res = smb_getnodestatus(&ctx->rmt, &nn, &ns, SUFFIX_SERV);
				if(res >= 0) {
					if(ii > 0) {
						/* ڑAhX̑1ɂĂ */
						u_long tmp_ip;
						tmp_ip = target_ip[0];
						target_ip[0] = target_ip[ii];
						target_ip[ii] = tmp_ip;
					}

					SMBC_LOG(("smb_getnodestatus: name = %s", nn.name));
					strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
					bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
					strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
					break;
				}
			}
		}
		strncpy(ctx->rmt.name, server, SMB_MAXSRVNAMELEN);
	} else {
		NBTNodeName nn;
		NBTNodeStatistics ns;
		int findnum = 0;
		int ii = 0;

		/* get master browser */ 
		name_type = smb_gethostbyname(group, SUFFIX_LMB, target_ip, &num);
		if(!name_type) {
			name_type = smb_gethostbyname(group, SUFFIX_DMB, target_ip, &num);
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: smb_gethostbyname(%s)[0x1d]. \n", group);
#endif
			if(!name_type) {
#ifndef SMBCD_USE
				smbc_mfprintf("cli_SmbOpenPipe failed: smb_gethostbyname(%s)[0x1b]. \n", group);
#endif
				return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			}
		}
		/*
		 * ftHgł"*SMBSERVER"w肷B
		 * ^[QbgPCWindowsNT4.0 SP6aȍ~OSł
		 * ̓ꖼŐڑ\
		 */
		strncpy(server, SMBSERVER, SMB_MAXSRVNAMELEN);
		for(ii = 0; ii < num; ii++) {
			ctx->rmt.naddr = target_ip[ii];
			memset(&nn, 0, sizeof(nn));
			memset(&ns, 0, sizeof(ns));
			res = smb_getnodestatus(&ctx->rmt, &nn, &ns, SUFFIX_SERV);
			if(res == LIBNBT_ERR_NOERR) {
				SMBC_LOG(("smb_getnodestatus: name = %s", nn.name)); 
				strncpy(server, nn.name, SMB_MAXSRVNAMELEN);
				bzero((char *)vcp->server, SMB_MAXSRVNAMELEN);
				strncpy((char *)vcp->server, server, SMB_MAXSRVNAMELEN);
				if(ii > 0) {
					/* ڑAhX̑1ɂĂ */
					u_long tmp_ip;
					tmp_ip = target_ip[0];
					target_ip[0] = target_ip[ii];
					target_ip[ii] = tmp_ip;
				}
				break;
			}
		}
	}
	if(vcp->username[0] != 0) {
		strncpy(username, vcp->username, SMB_MAXUSERNAMELEN); 
		SMBC_LOG(("input user name [%s]", username));
	}
	if(vcp->password[0] != 0) {
		strncpy(password, vcp->password, SMB_MAXPASSWORDLEN); 
		SMBC_LOG(("input password ****"));
	}

cli_SmbOpenPipe_retry:
	strncpy(ctx->rmt.name, server, SMB_MAXSRVNAMELEN);
	ctx->rmt.suffix = SUFFIX_SERV;
	res = nbt_connectTimeout(&ctx->ss_client, &ctx->rmt, &ctx->local, target_ip, num);
	if(res < 0) {
		printf("failed nbt_connectTimeout (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("cli_SmbOpenPipe failed: nbt_connectTimeout. \n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		syslog( LOG_ERR, "SMBC lib. failed to connect server(%s).", ctx->rmt.name );
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}

	res = nbt_getfd(ctx->ss_client);
	if(res < 0) {
		printf("failed nbt_getfd (%d)\n", res);
#ifndef SMBCD_USE
		smbc_mfprintf("cli_SmbOpenPipe failed: nbt_getfd. \n");
#endif
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	ctx->cli->fd = res;
	ctx->cli->port = 139;
	fstrcpy(ctx->cli->domain, group);
	fstrcpy(ctx->cli->share, "IPC$");
#ifdef USE_BROADCAST
	fstrcpy(ctx->cli->desthost, server);
	fstrcpy(ctx->cli->called.name, server);
#else
	fstrcpy(ctx->cli->desthost, group);
	fstrcpy(ctx->cli->called.name, group);
#endif
	ctx->cli->called.name_type = SUFFIX_SERV;
	ctx->cli->dest_ip.s_addr = htonl(ctx->rmt.naddr);

	if(!cli_negprot(ctx->cli, &char_code)) {
		if(ctx->cli->fd < 0) {
			if(cli_SmbOpenPipe_negprot_fail++ < 5) {
				nbt_close(ctx->ss_client);
				usleep(smb_random()%1000+100);
				goto cli_SmbOpenPipe_retry;
			}
			printf("cli_negprot fail %d\n", ERRNO);
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: cli_negprot. \n");
#endif
			return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		}
		printf("retry NULL domain\n");
		ctx->cli->domain[0] = NULL;
		if(!cli_negprot(ctx->cli, &char_code)) {
			int smberr;
			if(ctx->cli->fd < 0) {
				if(cli_SmbOpenPipe_negprot_fail++ < 5) {
					nbt_close(ctx->ss_client);
					usleep(smb_random()%1000+100);
					goto cli_SmbOpenPipe_retry;
				}
				printf("failed negprot\n");
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			syslog( LOG_ERR, "SMBC lib. failed to negotiate.(server=%s)", ctx->rmt.name);
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: cli_negprot(retry). \n");
#endif
			return smberr;
		}
	}
	SMBC_LOG(("server_type [%s]", ctx->cli->server_type));
	SMBC_LOG(("server_os [%s]", ctx->cli->server_os));
	SMBC_LOG(("server_domain [%s]", ctx->cli->server_domain));
	SMBC_LOG(("capabilities [0x%x]", ctx->cli->capabilities));
	SMBC_LOG(("protocol [%d]", ctx->cli->protocol));

	/*
	 * 2002/11/21 terao add
	 */
	if(ctx->cli->server_domain) {
		strncpy(vcp->group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
		strncpy(group, ctx->cli->server_domain, SMB_MAXSRVNAMELEN);
	}

	if (!cli_session_setup(ctx->cli, username,
		password, strlen(password), 
		password, strlen(password), 
		"", &char_code)) {

		/*
		 * if a password was not supplied
		 * then try again with a null username
		 */
		 if (!password[0] || username[0] ||
				!cli_session_setup(ctx->cli, "", "", 0, 
				"", 0, group, &char_code)) {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}
			printf("cli_SmbOpenPipe: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: cli_session_setup(user=%s). \n", username);
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 } else {
			int smberr;
			if(ctx->cli->fd == -1) {
				smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
			} else {
				smberr = smbc_errno(ctx->cli);
			}

			printf("cli_SmbOpenPipe: cli_session_setup failed: \n");
#ifndef SMBCD_USE
			smbc_mfprintf("cli_SmbOpenPipe failed: cli_session_setup(user=%s). \n", username);
#endif
			syslog( LOG_ERR, "SMBC lib. failed to user logon(server=%s).", ctx->rmt.name);
			cli_shutdown(ctx->cli);
			ctx->cli = NULL;
			return smberr;
		 }
	}

        if (!cli_send_tconX(ctx->cli, "IPC$", "IPC", password, strlen(password), &char_code)) {
		int smberr;
		if(ctx->cli->fd == -1) {
			smberr = smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
		} else {
			smberr = smbc_errno(ctx->cli);
		}
		printf("cli_SmbOpenPipe: cli_send_tconX failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("cli_SmbOpenPipe failed: cli_send_tconX(IPC$). \n");
#endif
		syslog( LOG_ERR, "SMBC lib. failed to tree connect(server=%s, share=IPC$).", ctx->rmt.name);
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
		return smberr;
	}

	return( SMBSUCCESS );
}

int cli_SmbClosePipe( smb_ctx* ctx, SMBC_PROC func)
{
	if(!cli_tdis(ctx->cli)) {
		printf("tree disconnect failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("cli_SmbClosePipe failed: cli_tdis. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}
	if(!cli_ulogoff(ctx->cli)) {
		printf("user logof failed: \n");
#ifndef SMBCD_USE
		smbc_mfprintf("cli_SmbClosePipe failed: cli_ulogoff. \n");
#endif
		return smbc_errcode(SMBERR_IF, SMBERR_PERMANENT, SMBERR_REMOTE);
	}
	cli_shutdown(ctx->cli);
	ctx->cli = NULL;

	if(ctx->ss_client >= 0) {
	        nbt_close(ctx->ss_client);
		ctx->ss_client = -1;
	}
	return( SMBSUCCESS );
}

static void smb_cleanup( smb_ctx* ctx )
{
	if(ctx->cli) {
		cli_shutdown(ctx->cli);
		ctx->cli = NULL;
	}
	if(ctx->ss_client >= 0) {
	        nbt_close(ctx->ss_client);
		ctx->ss_client = -1;
	}
}


#ifdef notdef
/****************************************************************************
list a server name
****************************************************************************/
static void server_fn(const char *name, uint32 m, 
                      const char *comment, void *state)
{
 	pstring uname, ucomment;
	
	dos_to_unix(uname, strncpy_fill(uname, name, 16));
	dos_to_unix(ucomment, comment);
	printf("\t%s     %s\n", uname, ucomment);
}
#endif

int smb_get_myname(char *name)
{
	char	msn_computer_name[SMB_MAXSRVNAMELEN+1];
	int	ret = 0;

	memset(msn_computer_name, 0, sizeof(msn_computer_name));

	ret = NcsGetSmbConf( "smb", "msn_computer_name", msn_computer_name, sizeof(msn_computer_name) );
	if(ret) {
		printf("smb_get_myname: msn_computer_name is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_myname: msn_computer_name is not found\n");
#endif
		return 1;
	}
	strncpy(name, msn_computer_name, SMB_MAXSRVNAMELEN+1);
	return 0;
}

int smb_get_mygroup(char *group)
{
	char	msn_workgroup_name[SMB_MAXSRVNAMELEN+1];
	int	ret = 0;

	memset(msn_workgroup_name, 0, sizeof(msn_workgroup_name));
	ret = NcsGetSmbConf( "smb", "msn_workgroup_name", msn_workgroup_name, sizeof(msn_workgroup_name) );
	if(ret) {
		printf("smb_get_mygroup: msn_workgroup_name is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_mygroup: msn_workgroup_name is not found\n");
#endif
		return 1;
	}
	strncpy(group, msn_workgroup_name, SMB_MAXSRVNAMELEN+1);
	SMBC_LOG(("smb_get_mygroup: msn_workgroup_name = %s", msn_workgroup_name));
	return 0;
}

int smb_get_mpx(int *mpx)
{
	char	buf[16];
	int	ret = 0;

	memset(buf, 0, sizeof(buf));
	ret = NcsGetSmbConf( "smb", "msn_mpx", buf, sizeof(buf) );
	if(ret) {
		printf("smb_get_mpx: msn_mpx is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_mpx: msn_mpx is not found\n");
#endif
		*mpx = 0;
		return 1;
	}
	*mpx = atol(buf);
	return 0;
}

int smb_get_myip(u_long *ip, int intf)
{
	char	buf[16];
	int	ret = 0;
	char	interface[8];

	snprintf(interface, sizeof(interface), "ip%d", intf);
	memset(buf, 0, sizeof(buf));
	ret = NcsGetSmbConf( "ipif", interface, buf, sizeof(buf) );
	if(ret) {
		printf("smb_get_myip: %s is not found\n", interface);
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_myip: %s is not found\n", interface);
#endif
		return 1;
	}
	*ip = atol(buf);
	return 0;
}

int smb_get_wins_addr(u_long *primary, u_long *secondary, char *scopeid, int intf)
{
	char	buff1[32];
	char	wins_available_s[16], ip_s[16];
	char	primary_wins_addr_s[16], secondary_wins_addr_s[16];
	int	ret = 0;
	u_long	wins_available = 0;
	u_long	ip = 0;
	u_long	primary_wins_addr = 0;
	u_long	secondary_wins_addr = 0;
	int	len = 0;

	memset(buff1, 0, sizeof(buff1));
	memset(wins_available_s, 0, sizeof(wins_available_s));
	memset(primary_wins_addr_s, 0, sizeof(primary_wins_addr_s));

	snprintf( buff1, sizeof(buff1), "ip%d", intf);
	ret = NcsGetSmbConf( "ipif", buff1, ip_s, sizeof(ip_s) );
	if(ret) {
		printf("smb_get_wins_addr: interface is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_wins_addr: interface is not found\n");
#endif
		return 1;
	}
	ip = atol(ip_s);
	if(!ip) {
		*primary = 0;
		*secondary = 0;
		return 1;
	}

	memset(wins_available_s, 0, sizeof(wins_available_s));
	snprintf( buff1, sizeof(buff1), "wins_available%d", intf);
	ret = NcsGetSmbConf( "ipif", buff1, wins_available_s, sizeof(wins_available_s) );
	if(ret) {
		printf("smb_get_wins_addr: wins_available is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_wins_addr: wins_available is not found\n");
#endif
		return 1;
	}
	wins_available = atol(wins_available_s);
	if(!wins_available) {
		*primary = 0;
		*secondary = 0;
		return 1;
	}
	memset(primary_wins_addr_s, 0, sizeof(primary_wins_addr_s));
	snprintf( buff1, sizeof(buff1), "primary_wins_addr%d", intf );
	ret = NcsGetSmbConf( "ipif", buff1, primary_wins_addr_s, sizeof(primary_wins_addr_s) );
	if(ret) {
		printf("smb_get_wins_addr: primary_wins_addr is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_wins_addr: primary_wins_addr is not found\n");
#endif
	} else {
		*primary = atol(primary_wins_addr_s);
	}

	memset(secondary_wins_addr_s, 0, sizeof(secondary_wins_addr_s));
	snprintf( buff1, sizeof(buff1), "secondary_wins_addr%d", intf );
	ret = NcsGetSmbConf( "ipif", buff1, secondary_wins_addr_s, sizeof(secondary_wins_addr_s) );
	if(ret) {
		printf("smb_get_wins_addr: secondary_wins_addr is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_wins_addr: secondary_wins_addr is not found\n");
#endif
	} else {
		*secondary = atol(secondary_wins_addr_s);
	}

	snprintf( buff1, sizeof(buff1), "scope_id%d", intf );
	ret = NcsGetSmbConf( "ipif", buff1, scopeid, NBT_LENGTH_SCOPEID );
	if(ret) {
		scopeid[0] = 0;
		printf("smb_get_wins_addr: scopeid is not found\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_get_wins_addr: scopeid is not found\n");
#endif
	}

	return 0;
}

#if 0
static int cli_GetBackupList(smb_ctx* ctx, const char *myname, const char *group_name, char *server)
{
  char outbuf[15];
  char *p = 0;
  u_long group_ip = 0, my_ip = 0;
  struct sockaddr_in sock_out;
  NBTNamePack	np;
  NBTAddrPack	ap;
  int res = 0;
  int nbt_fd = -1;
  char backup_list_buf[1024];
  int name_type = 0;
  u_long	target_ip[SMB_MAX_IP];
  int		num = SMB_MAX_IP;

  bzero(&np, sizeof(np));
  bzero(&ap, sizeof(ap));

  name_type = smb_gethostbyname(group_name, SUFFIX_LMB, target_ip, &num);
  if(!name_type) {
    name_type = smb_gethostbyname(group_name, SUFFIX_DMB, target_ip, &num);
    if(!name_type) {
      return False;
    }
  }

  nbt_fd = nbt_open(NBT_STYPE_DATAGRAM, htonl(INADDR_ANY), htons(138));
  if(nbt_fd < 0) {
    printf("nbt_open failed (%d)\n", nbt_fd);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupList: nbt_open failed (%d)\n", nbt_fd);
#endif
    return False;
  }

  DEBUG_PRINT(5, ("Socket bound to IP:%s, port: %d\n", inet_ntoa(sock_out.sin_addr), ntohs(sock_out.sin_port)));

  /* Now, build the request */

  memset(backup_list_buf, '\0', sizeof(backup_list_buf));
  memset(outbuf, '\0', sizeof(outbuf));

  p = outbuf;

  SCVAL(p, 0, ANN_GetBackupListReq);
  p++;

  SCVAL(p, 0, 4); /* Count pointer ... */
  p++;

  SIVAL(p, 0, 1); /* The sender's token ... */
  p += 4;

  res = cli_SendBackupListReq(nbt_fd, MAILSLOT_BROWSE, outbuf, 
		    PTR_DIFF(p, outbuf), myname, 0, group_name, 
		    SUFFIX_LMB, group_ip, my_ip, 138, sock_out.sin_port);
  if(res < 0) {
    printf("cli_SendBackupListReq failed (%d)\n", nbt_fd);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_SendBackupListReq failed (%d)\n", nbt_fd);
#endif
    nbt_close(nbt_fd);
    return False;
  }

  /* We should check the error and return if we got one */

  /* Now, get the response ... */
  res = cli_GetBackupListResp(nbt_fd, MAILSLOT_BROWSE, backup_list_buf, sizeof(backup_list_buf));
  if(res < 0) {
    printf("cli_SendBackupListReq failed (%d)\n", nbt_fd);
    nbt_close(nbt_fd);
    return False;
  }

  /* Should check the response here ... FIXME */

  nbt_close(nbt_fd);

  bzero(server, SMB_MAXSRVNAMELEN);
  strncpy(server, backup_list_buf, SMB_MAXSRVNAMELEN);

  return True;
}

/*
 * cli_send_mailslot, send a mailslot for client code ...
 */

static int cli_SendBackupListReq(int nbt_fd, char *mailslot, 
		      char *buf, int len,
		      const char *srcname, int src_type, 
		      const char *dstname, int dest_type,
		      u_long dest_ip, u_long src_ip,
		      int dest_port, int src_port)
{
  char *p2 = 0;
  char data[MAX_DGRAM_SIZE];
  int datasize = 0;
  int res = 0;
  NBTNamePack srcnp, dstnp;

  bzero(data, sizeof(data));
  bzero(&srcnp, sizeof(srcnp));
  bzero(&dstnp, sizeof(dstnp));

  /* Setup the smb part. */
  smb_set_message(data,17,17 + len,False);

  SCVAL(data,smb_com,SMBtrans);
  SSVAL(data,smb_vwv1,len);
  SIVAL(data,smb_vwv6,1000);	/* timeout @terao add */
  SSVAL(data,smb_vwv11,len);
  SSVAL(data,smb_vwv12,70 + strlen(mailslot));
  SSVAL(data,smb_vwv13,3);
  SSVAL(data,smb_vwv14,1);
  SSVAL(data,smb_vwv15,1);
  SSVAL(data,smb_vwv16,2);
  p2 = smb_buf(data);
  pstrcpy(p2,mailslot);
  p2 = skip_string(p2,1);

  memcpy(p2,buf,len);
  p2 += len;

  datasize = PTR_DIFF(p2,data+4);

  bzero(&srcnp, sizeof(srcnp));
  bzero(&dstnp, sizeof(dstnp));
  strncpy(srcnp.name, srcname, 15); 
  srcnp.suffix = 0x00;
  srcnp.naddr = src_ip; 
  srcnp.group = NBT_G_GROUP;
  strncpy(dstnp.name, dstname, 15); 
#ifdef notdef
  dstnp.suffix = SUFFIX_LMB;
  dstnp.naddr = ((src_ip.s_addr & 0x00ffffff) | 0xff000000);
  dstnp.group = NBT_G_GROUP;
  res = nbt_senddatagram(nbt_fd, data+4, datasize, &srcnp, &dstnp);
#else
  dstnp.suffix = SUFFIX_DMB;
  dstnp.naddr = dest_ip;
  dstnp.group = NBT_G_GROUP;
  res = nbt_senddatagram(nbt_fd, data+4, datasize, &srcnp, &dstnp);
#endif

  return res;

}

/*
 * cli_get_response: Get a response ...
 */
static int cli_GetBackupListResp(int nbt_fd, char *mailslot, char *buf, int bufsiz)
{
  char	data[MAX_DGRAM_SIZE];
  NBTNamePack	srcnp, dstnp;
  int	retval = 0;
  int	ecode = 0;
  int	rsize = 0;
  char	*p = 0;
  u_char	cmd = 0;
  u_char	wcnt = 0;
  u_short	bcnt = 0;
  u_char	opcode = 0;
  u_short	srvcnt = 0;

  bzero(data, sizeof(data));
  bzero(&srcnp, sizeof(srcnp));
  bzero(&dstnp, sizeof(dstnp));
  rsize = nbt_recvdatagram(nbt_fd, data, MAX_DGRAM_SIZE, &srcnp, &dstnp, &retval, &ecode);

  if(rsize < 0) { 
    printf("nbt_recvdatagram: failed retval = %d, ecode = %d\n", retval, ecode);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupListResp: nbt_recvdatagram retval = %d, ecode = %d\n", retval, ecode);
#endif
    return -1;
  }

  if((u_char)*data == 0xff && strncmp(data+1, "SMB", 3)) {
    printf("nbt_recvdatagram: invalid header \n");
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupListResp: nbt_recvdatagram: invalid header \n");
#endif
    return -1;
  }

  p = data;
  p -= 4;
  cmd = CVAL(p,smb_com); 
  if(cmd != SMBtrans) {
    printf("nbt_recvdatagram: invalid command %d\n", cmd);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupListResp: nbt_recvdatagram: invalid command %d\n", cmd);
#endif
    return -1;
  }
  p += smb_wct;		/* skip smb header */
  wcnt = CVAL(p, 0);
  p++;
  printf("wcnt =%d(0x%x)\n", wcnt, wcnt); 
  p += wcnt * 2;	/* skip word parameter */

  bcnt = SVAL(p, 0);
  p += 2;
  if(strncmp(p, mailslot, strlen(mailslot))) {
    printf("nbt_recvdatagram: invalid byte data %s\n", p);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupListResp: nbt_recvdatagram: invalid byte data %s\n", p);
#endif
    return -1;
  }
  p += strlen(mailslot) + 1;
  opcode = CVAL(p, 0);
  if(opcode != ANN_GetBackupListResp) {
    printf("nbt_recvdatagram: invalid byte data %s\n", p);
#ifndef SMBCD_USE
    smbc_mfprintf("cli_GetBackupListResp: nbt_recvdatagram: invalid byte data %s\n", p);
#endif
    return -1;
  }
  srvcnt = CVAL(p, 1);	/* ??? */

  p += 6;

  strncpy(buf, p, bufsiz);
  printf("buf = %s\n", buf); 
  return 0;
}
#endif

int smbc_errno(struct cli_state *cli)
{
	uint8	eclass = 0;
	uint32	ecode = 0;

	cli_dos_error(cli, &eclass, &ecode);
	if(ecode) {
		smbc_errprint(cli);
	}
	return smbc_dos_errno(eclass, ecode);
}

void smbc_errprint(struct cli_state *cli)
{
	if(!cli) {
		return;
	}
#ifndef SMBCD_USE
	smbc_mfprintf("[smbc error status]\n");
	smbc_mfprintf("\tprotocol=%d\n", cli->protocol);
	smbc_mfprintf("\tsec_mode=%d\n", cli->sec_mode);
	smbc_mfprintf("\tdesthost=%s\n", cli->desthost);
	smbc_mfprintf("\tuser_name=%s\n", cli->user_name);
	smbc_mfprintf("\tdomain=%s\n", cli->domain);
	smbc_mfprintf("\tserver_type=%s\n", cli->server_type);
	smbc_mfprintf("\tserver_os=%s\n", cli->server_os);
	smbc_mfprintf("\tserver_domain=%s\n", cli->server_domain);
	smbc_mfprintf("\tcalled.name=%s, called.scope=%s, called.name_type=%d\n", cli->called.name, cli->called.scope, cli->called.name_type);
	smbc_mfprintf("\tcalling.name=%s, calling.scope=%s, calling.name_type=%d\n", cli->calling.name, cli->calling.scope, cli->calling.name_type);
	smbc_mfprintf("\tdest_ip=0x%x\n", cli->dest_ip.s_addr);
	smbc_mfprintf("\tNT error=0x%x\n", cli->nt_error);
#endif
	return;
}


static int smbc_dos_errno(uint8 eclass, uint32 ecode)
{
	uint8	econd = 0;

	if(eclass == ERRDOS) {
		switch(ecode) {
			case ERRbadfile:
			case ERRbadpath:
			case ERRnofids:
			case ERRnoaccess:
			case ERRbadaccess:
			case ERRnomem:
			case ERRnofiles:
			case ERRlock:
			case ERRfilexists:
				econd = SMBERR_TEMPORARY;
				break;
			default:
				econd = SMBERR_PERMANENT;
				break;
		}
	} else if(eclass == ERRSRV) {
		switch(ecode) {
			case ERRbadpw:
			case ERRaccess:
			case ERRnoresource:
				econd = SMBERR_TEMPORARY;
				break;
			default:
				econd = SMBERR_PERMANENT;
				break;
		}
	} else if(eclass == ERRHRD) {
		switch(ecode) {
			case ERRnotready:
			case ERRseek:
				econd = SMBERR_TEMPORARY;
				break;
			default:
				econd = SMBERR_PERMANENT;
				break;
		}
	}

	return smbc_errcode(eclass, econd, ecode);
}


int smbc_errcode(int class, int condition, int code)
{
	int	ret = 0;

	ret |= (condition & 0xff) << 24;
	ret |= (class & 0xff) << 16;
	ret |= (code & 0xffff);

	return ret;
}


int smb_getnodestatus(const NBTNamePack* rmt, NBTNodeName* nn, NBTNodeStatistics* ns, u_short suffix)
{
	int		ns_client = 0;
	int		res = 0;
	int		nn_num = SMB_MAX_NODE;
	int		findnum = 0;
	NBTNodeName	*tmp_nn = 0;
	int		i = 0;
	BOOL		found = FALSE;

	tmp_nn = (NBTNodeName *)malloc(sizeof(NBTNodeName)*nn_num);
	if(!tmp_nn) {
		return -1;
	}
	ns_client = nbt_open(NBT_STYPE_NAME, htonl(INADDR_ANY), htons(NBT_PORT_ANY));
	if(ns_client < 0) {
		free(tmp_nn);
		return -1;
	}
	res = nbt_getnodestatus(ns_client, rmt, tmp_nn, nn_num, &findnum, ns, NULL);
	if(res == LIBNBT_ERR_NOERR) {
		for(i = 0; i < MIN(findnum, nn_num); i++) {
			if(tmp_nn[i].suffix == suffix) {
				/* ŏɈvsuffix̂̂gp */
				*nn = tmp_nn[i];
				found = TRUE;
				break;
			}
		}
		if(!found) {	/* Ȃ */
			res = -1;	/* G[ɂĂ */
			if(findnum > nn_num) {
				NBTNodeName *tmp_nn2 = 0;
				/*
				 * NetBIOSẅɓ肫ȂƂ
				 * ̈g債čēxO擾
				 */
				nn_num = findnum;
				tmp_nn2 = (NBTNodeName *)realloc(tmp_nn, sizeof(NBTNodeName)*nn_num);
				if(!tmp_nn2) {
					nbt_close(ns_client);
					free(tmp_nn);
					return -1;
				}
				tmp_nn = tmp_nn2;
				res = nbt_getnodestatus(ns_client, rmt, tmp_nn, findnum, &findnum, ns, NULL);
				if(res == LIBNBT_ERR_NOERR) {
					for(i = 0; i < MIN(findnum, nn_num); i++) {
						if(tmp_nn[i].suffix == suffix) {
							/* ŏɈvsuffix̂̂gp */
							*nn = tmp_nn[i];
							found = TRUE;
						}
					}
				}
				if(!found) {
					res = -1;
				}
			}
		}
	}
	nbt_close(ns_client);
	free(tmp_nn);
	return res;
}

/*
 * w肳ꂽĐIPAhX擾
 *  NetBIOSiWINS/NBTu[hLXgj-> zXgiDNSj ƂȂ
 * ߂l@AhX̃^Cv@@0:s, 1: NB, 2:IPAhX, 3:DNS
 */
int smb_gethostbyname(const char *target_name, u_short suffix, u_long *target_ip, int *num)
{
	NBTNamePack	np;
	NBTAddrPack	ap[*num];	/* Ăяonum1ȏłȂ΂ȂȂ */
	int		findnum = 0;
	u_long	ip_addr[*num];	/* Ăяonum1ȏłȂ΂ȂȂ */
	int		name_type = 0;	/* 0:s, 1: NB, 2:IPAhX, 3:DNS */
	u_long	primary_winsaddr = 0, secondary_winsaddr = 0;
	char	scopeid[NBT_LENGTH_SCOPEID];
	u_long	ip[3];
	u_char	iftye[3] = {IFT_ETHER, IFT_IEEE80211, IFT_IP1394};
	int		ns_client = -1;
	int		i = 0, j = 0;

	if(!target_ip || !num) {
		return 0;
	}
	if(*num <= 0) {
		return 0;
	}

#ifdef notdef	/* @terao20031128 */
	if ((ip_addr[0] = inet_addr(target_name)) != (u_long)(-1)) {
			target_ip[0] = ip_addr[0];
			*num = 1;
			name_type = 2;
			return name_type;
	}
#else
	if ((ip_addr[0] = smb_inet_addr(target_name)) != (u_long)(-1)) {
			target_ip[0] = ip_addr[0];
			*num = 1;
			name_type = 2;
			return name_type;
	}
#endif

	bzero(&np, sizeof(np));
	bzero(ap, sizeof(ap));
	bzero(scopeid, sizeof(scopeid));
	bzero(ip_addr, sizeof(ip_addr));
	/*
	 * O̒15ȉ̂ƂNetBIOS
	 */
	if(strlen(target_name) < NBT_LENGTH_NETBIOSNAME) {
		int res = 0;

		ns_client = nbt_open(NBT_STYPE_NAME, htonl(INADDR_ANY), htons(NBT_PORT_ANY));
		strncpy(np.name, target_name, NBT_LENGTH_NETBIOSNAME);
		np.suffix = suffix;
		/*
		 * WINST[oAhX͊eC^[tFCX
		 * iEthernet//IEEE1394jƂɂ̂
		 * ꂼꌟ
		 */
		for(i = 0; i < 3; i++) {
			u_long	bcnaddr;

			/*
			 * I/F݂邩`FbN
			 */
			smb_get_myip(&ip[i], i+1);
			if(ip[i] == 0) {
				continue;
			}
			findnum = 0;
			bzero(ip_addr, sizeof(ip_addr));
			bzero(ap, sizeof(ap));
			bzero(scopeid, sizeof(scopeid));
			primary_winsaddr = secondary_winsaddr = 0;
			/*
			 * I/FɃu[hLXgAhXݒ肷
			 */
			if(smb_GetIFBCAddr(iftye[i], &bcnaddr, NULL) != 0) {
				bcnaddr = 0xffffffff;
			}
			nbt_setbcaddr(ns_client, bcnaddr);
			smb_get_wins_addr(&primary_winsaddr, &secondary_winsaddr, scopeid, i+1);
			if(scopeid[0] != NULL) {
				printf("set scopeid %s\n", scopeid);
#ifndef SMBCD_USE
				smbc_mfprintf("\tscopeid=%s, scopeid\n");
#endif
				nbt_setscopeid(scopeid);
			}
			if(primary_winsaddr) {
				nbt_setnodetype(NBT_ONT_P);
				nbt_setwinsaddr(htonl(primary_winsaddr));
				res = nbt_findname(ns_client, &np, ap, *num, &findnum, NULL);
				if (res == 0) {
					if(findnum > *num) {
						findnum = *num;
					}
					for(j = 0; j < findnum; j++) {
						ip_addr[j] = ap[j].naddr; 
					}
					name_type = 1;
					break;
				}
#ifndef SMBCD_USE
				smbc_mfprintf("\tprimary_winsaddr=0x%x, primary_winsaddr\n");
#endif
			}
			if(secondary_winsaddr) {
				nbt_setnodetype(NBT_ONT_P);
				nbt_setwinsaddr(htonl(secondary_winsaddr));
				res = nbt_findname(ns_client, &np, ap, *num, &findnum, NULL);
				if (res == 0) {
					if(findnum > *num) {
						findnum = *num;
					}
					for(j = 0; j < findnum; j++) {
						ip_addr[j] = ap[j].naddr; 
					}
					name_type = 1;
					break;
				}
#ifndef SMBCD_USE
				smbc_mfprintf("\tsecondary_winsaddr=0x%x, secondary_winsaddr\n");
#endif
			}
			nbt_setnodetype(NBT_ONT_B);
			res = nbt_findname(ns_client, &np, ap, *num, &findnum, NULL);
			if (res == 0) {
				if(findnum > *num) {
					findnum = *num;
				}
				for(j = 0; j < findnum; j++) {
					ip_addr[j] = ap[j].naddr; 
				}
				name_type = 1;
				break;
			}
			printf("could not resolve name: %s<%x> err=%d\n", target_name, np.suffix, res);
#ifndef SMBCD_USE
			smbc_mfprintf("smb_gethostbyname: could not resolve name: %s<%x> err=%d\n", target_name, np.suffix, res);
#endif
			/*
			 * NBTOłȂƂ
			 * ZbgXR[vIDNA
			 */
			if(scopeid[0] != NULL) {
				bzero(scopeid, sizeof(scopeid));
				nbt_setscopeid(scopeid);
			}
		}
		nbt_close(ns_client);
	}

	if(!name_type) {
		struct hostent *host = 0;
		host = gethostbyname(target_name);
		if(host == 0) {
			printf("smb_gethostbyname: failed gethostbyname %s\n", target_name);
#ifndef SMBCD_USE
			smbc_mfprintf("smb_gethostbyname: failed gethostbyname %s\n", target_name);
#endif
		} else if (host == (struct  hostent *)-1) {
			printf("smb_gethostbyname: dns name not found\n");
#ifndef SMBCD_USE
			smbc_mfprintf("smb_gethostbyname: dns name not found\n");
#endif
		} else if (host->h_length != 4) {
			printf("smb_gethostbyname: target not found\n");
#ifndef SMBCD_USE
			smbc_mfprintf("smb_gethostbyname: target not found\n");
#endif
		} else {
			findnum = 1;
			memcpy(&ip_addr[0], host->h_addr_list[0], host->h_length);
			name_type = 3;
		}
  	}
	if(name_type) {
		for(j = 0; j < findnum; j++) {
			target_ip[j] = ip_addr[j];
		}
		*num = findnum;
	}
	return name_type;
}

static int  NcsGetSmbConf( char* tag, char* menu, char* str, int str_max )
{
	int	ret;

	ret = NcsGetXXXConf( SMB_CONF_PATH_NAME, tag, menu, str, str_max );
	return( ret );
}

static int  NcsGetNbtdConf( char* tag, char* menu, char* str, int str_max )
{
	int	ret;

	ret = NcsGetXXXConf( NBTD_CONF_PATH_NAME, tag, menu, str, str_max );
	return( ret );
}

static char * smb_strrchr(char *s, int c)
{
        register char *s0 = NULL;

        while (*s) {
                if (*s == c)
                        s0 = s;
                s++;
        }
        return(s0);
}

static u_long smb_randseed = 1;

void smb_srandom( u_long seed)
{
        smb_randseed = seed;
}

/*
 * Pseudo-random number generator for randomizing the profiling clock,
 * and whatever else we might use it for.  The result is uniform on
 * [0, 2^31 - 1].
 */
u_long smb_random()
{
        register long x, hi, lo, t;

        /*
         * Compute x[n + 1] = (7^5 * x[n]) mod (2^31 - 1).
         * From "Random number generators: good ones are hard to find",
         * Park and Miller, Communications of the ACM, vol. 31, no. 10,
         * October 1988, p. 1195.
         */
        x = smb_randseed;
        hi = x / 127773;
        lo = x % 127773;
        t = 16807 * lo - 2836 * hi;
        if (t <= 0)
                t += 0x7fffffff;
        smb_randseed = t;
        return (t);
}

static void *(*alloc_func)(size_t);
static void (*free_func)(void*);

void NcsSmbSetAllocatorFunc( void *(*_alloc_func)(size_t), void(*_free_func)(void *))
{
   SMBC_LOG(("NcsSmbSetAllocatorFunc(0x%x, 0x%x) enter", _alloc_func, _free_func));
   if (_alloc_func && _free_func) {
     alloc_func = _alloc_func;
     free_func = _free_func;
   } else {
     alloc_func = malloc;
     free_func = free;
   }
   SMBC_LOG(("NcsSmbSetAllocatorFunc exit"));
}

void *smb_malloc(size_t size)
{
	if(alloc_func) {
		return(alloc_func(size));
	} else {
		return(malloc(size));
	}
}

void smb_free(void *memp)
{
	if(free_func) {
		return(free_func(memp));
	} else {
		return(free(memp));
	}
}

static int smb_GetIFConf(stIFConf* dst, stIFConf* src)
{
	int s;
	caddr_t	buf = NULL;
	caddr_t	temp;
	
	struct ifconf	ifc;
	struct ifreq*	ifr;
	struct sockaddr_dl* sdl;
	int allocsize = 0;
	int i;
	
	if((dst == NULL) || (src == NULL)){
		return -1;
	}
	memset(dst, 0, sizeof(stIFConf));
	
	// dgram socket open
	s = socket(AF_INET, SOCK_DGRAM, 0);
	if(s < 0){
		printf("socket failed\n");
#ifndef SMBCD_USE
		smbc_mfprintf("smb_GetIFConf: socket failed\n");
#endif
		return -1;
	}
	
	do{
		allocsize += 1024;
		temp = (caddr_t)realloc(buf, allocsize);
		if(temp == NULL){
			free(buf);
			close(s);
			printf("realloc failed\n");
#ifndef SMBCD_USE
			smbc_mfprintf("smb_GetIFConf: realloc failed\n");
#endif
			return -1;
		}
		buf = temp;
		
		ifc.ifc_len = allocsize;
		ifc.ifc_buf = buf;
		
		if(ioctl(s, SIOCGIFCONF, &ifc) < 0){
			free(buf);
			close(s);
			printf("ioctl error\n");
#ifndef SMBCD_USE
			smbc_mfprintf("smb_GetIFConf: ioctl error\n");
#endif
			return -1;
		}
	} while(ifc.ifc_len >= allocsize);
	
	// printf("ifc_len = %d, allocsize = %d\n", ifc.ifc_len, allocsize);
	
	for(i=0; i<ifc.ifc_len; ){
		ifr = (struct ifreq* )((caddr_t)ifc.ifc_req + i);
		i += sizeof(ifr->ifr_name) + MAX(ifr->ifr_addr.sa_len, sizeof(struct sockaddr));
		
		if(ifr->ifr_addr.sa_family != AF_LINK){
			continue;
		}
		
		sdl = (struct sockaddr_dl* ) &ifr->ifr_addr;
		
		// printf("type = %d\n", sdl->sdl_type);
		/* keÿvmF */
		if(src->iftype != 0){
			if(src->iftype != sdl->sdl_type){
				/* v */
				continue;
			}
		}
		if(src->ifname[0] != '\0'){
			if(strcmp(src->ifname, ifr->ifr_name)){
				/* v */
				continue;
			}
		}
		if(src->mac_len > 0){
			if(memcmp(src->mac, LLADDR(sdl), sdl->sdl_alen)){
				/* v */
				continue;
			}
		}
		
		/* v */
		dst->iftype = sdl->sdl_type;		/* C^[tFCX^CṽRs[ */
		strcpy(dst->ifname, ifr->ifr_name);	/* C^[tFCX̃Rs[ */
		memcpy(dst->mac, LLADDR(sdl), sdl->sdl_alen);	/* MACAhX̃Rs[ */
		dst->mac_len = sdl->sdl_alen;					/* MACAhX̃Rs[ */
					
		free(buf);
		close(s);
		return 1;
	}
	
	// printf("not found\n");
	free(buf);
	close(s);
	
	return 0;
}

static int smb_GetIFBCAddr(u_char iftype, u_long* bcnaddr, const char* ifname)
{
	int s;
	struct ifreq ifr;
	stIFConf conf;
	stIFConf a;
	int res;
	
	if(bcnaddr == NULL){
		return -1;
	}
	
	memset(&ifr, 0, sizeof ifr);
	
	if(ifname == NULL){
		/* ^Cvl[ɕϊ */
		memset(&a, 0, sizeof a);
		a.iftype = iftype;
		res = smb_GetIFConf(&conf, &a);
		if(res <= 0){
			// printf("GetIFConf return %d\n", res);
			return -1;
		}
		strncpy(ifr.ifr_name, conf.ifname, IFNAMSIZ-1);
	}
	else{
		strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
	}
	// printf("ifname = %s\n", ifr.ifr_name);
	
	s = socket(AF_INET, SOCK_DGRAM, 0);
	if(s < 0){
		return -1;
	}
	
	if(ioctl(s, SIOCGIFBRDADDR, (caddr_t)&ifr) < 0){
		close(s);
		return -1;
	}
	close(s);
	
	*bcnaddr = *(u_long* )&((struct sockaddr_in* )&ifr.ifr_broadaddr)->sin_addr;
	
	return 0;
}

#ifndef SMBCD_USE
int NcsSmbMfprintfCreate ( key_t key )
{
	void *mf;

	mf = (void *)mfprintf_create( key, MPRINTF_SZ_4K);
	if(mf == NULL || (int)mf < 0) {
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	if( mfprintf_control(mf, MFPRINTF_TIMESTAMP, (char *)1) < 0 ) {
		return smbc_errcode(SMBERR_LIB, SMBERR_PERMANENT, SMBERR_IFERR);
	}
	smbc_mf = mf;
	return SMBSUCCESS;
}

int NcsSmbMfprintfDelete ( void )
{
	if(smbc_mf != NULL) {
		(void) mfprintf_control(smbc_mf, MFPRINTF_DELETE, (char *)1);
		smbc_mf = NULL;
	}
	return SMBSUCCESS;
}

#if 1
void smbc_mfprintf(char *format, ...)
{
        va_list ap;

	if(smbc_mf == NULL) {
		return;
	}
        va_start(ap, format);
	vmfprintf(smbc_mf, format, ap);
	va_end(ap);
	return;
}
#else
#define MAXNUM 32
static char           *digits = "0123456789abcdef";
void smbc_mfprintf(char *format, ...)
{
        va_list ap;
        register int    ch;
        register u_char *fmp;
        register int    *argp;
        register char   *str;
        register u_long number;
        register int    len;
        register char   pad;
        char            buf[MAXNUM];
	char		tmp_str[2];

	if(smbc_mf == NULL) {
		return;
	}

        va_start(ap,format);
        argp = (int *)ap;       /* only for stack based arguments */

	bzero(tmp_str, sizeof(tmp_str));
        for (fmp = format; ch = *fmp; fmp++) {
                if (ch != '%') {
			tmp_str[0] = ch;
                        mfputs(tmp_str, smbc_mf);
                        continue;
                }
                if ((ch = *++fmp) == 0)
                        break;
                len = 0;
                if (ch >= '0' && ch <= '9') {
                        pad = (ch == '0')? '0':' ';
                        for (; (ch = *fmp) >= '0' && ch <= '9'; fmp++)
                                len = ch - (int)'0';
                }
                switch (ch) {
                case 's':       /* string */
                        str = (char *)*argp++;
			mfputs(str, smbc_mf);
                        break;
                case 'd':       /* decimal */
                        str = buf;
                        if (number = (u_long)*argp++) {
                                if ((long)number < 0) {
                                        number = -number;
					tmp_str[0] = '-';
                                        mfputs(tmp_str, smbc_mf);
                                }
                                for (; number; number /= 10)
                                        *str++ = digits[number % 10];
                        }
                        else
                                *str++ = '0';
                        while (str - buf < len)
                                *str++ = pad;
                        while (--str >= buf) {
				tmp_str[0] = *str;
                                mfputs(tmp_str, smbc_mf);
			}
                        break;
                case 'x':       /* hex */
                        str = buf;
                        if (number = (u_long)*argp++)
                                for (; number; number >>= 4)
                                        *str++ = digits[number & 0xf];
                        else
                                *str++ = '0';
                        while (str - buf < len)
                                *str++ = pad;
                        while (--str >= buf) {
				tmp_str[0] = *str;
                                mfputs(tmp_str, smbc_mf);
			}
                        break;
                case 'c':
                        ch = (char)*argp++;
                        /* fall through */
                default:
			tmp_str[0] = ch;
                        mfputs(tmp_str, smbc_mf);
                        break;
                }
        }
        va_end(ap);
        return;
}
#endif
#endif


